{$INCLUDE cHeader.inc}
unit cDynDomain;

{                                                                              }
{                      Dynamic Domain updater v0.04 (L0)                       }
{                                                                              }
{        This unit is copyright  2000 by David Butler (david@e.co.za)         }
{                                                                              }
{                  This unit is part of Delphi Fundamentals.                   }
{                  It's original file name is cDynDomain.pas                   }
{                                                                              }
{                I invite you to use this unit, free of charge.                }
{        I invite you to distibute this unit, but it must be for free.         }
{             I also invite you to contribute to its development,              }
{             but do not distribute a modified copy of this file.              }
{       Send modifications, suggestions and bug reports to david@e.co.za       }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   2000/06/18  v0.01  Initial version (ods.org updater)                       }
{   2000/06/22  v0.02  Cleanup code.                                           }
{   2000/08/06  v0.03  Renamed to cDynDomain.                                  }
{                      Added support for eyep.net                              }
{   2000/09/24  v0.04  Major rewrite. Added abstraction ADynDomainService      }
{                      and two base classes TTextDynDomainService and          }
{                      ThttpDynDomainService to easily add new services.       }
{                      Added support for yi.org                                }
{                      Added support for hn.org                                }
{                                                                              }
interface

uses
  SysUtils,

  // Delphi Fundamentals
  cStreams,
  cHTTP;

const
  UnitVersion = '0.04';

{                                                                              }
{ TDynDomainUpdater                                                            }
{   Before connecting, set UserName and Password.                              }
{   Call AddDynamicDomain for each domain to use. Leave IP blank to tell the   }
{   ODS server to use the IP of the incomming connection (CONNIP).             }
{   Domains can also be added by setting the DomainIP or DomainActive          }
{   properties.                                                                }
{   Call ActivateDomains/DeactivateDomains to activate/deactivate all the      }
{   dynamic domains in one session with the server.                            }
{   Set the DomainActive property to activate/deactivate a single domain in a  }
{   session with the server.                                                   }
{   Return from a method assumes the method was successful, otherwise an       }
{   exception will be raised.                                                  }
{                                                                              }
{   Examples:                                                                  }
{     With TDynDomainUpdater.Create do                                         }
{       try                                                                    }
{         DomainActive ['dj.nailed.org'] := True;                              }
{       finally                                                                }
{         Free;                                                                }
{       end;                                                                   }
{                                                                              }
{     With TDynDomainUpdater.Create do                                         }
{       try                                                                    }
{         AddDomain ('dj.nailed.org', '');                                     }
{         AddDomain ('dummy.ods.org', '127.0.0.1');                            }
{         ActivateDomains;                                                     }
{       finally                                                                }
{         Free;                                                                }
{       end;                                                                   }
{                                                                              }
{     With TDynDomainUpdater.Create do                                            }
{       try                                                                    }
{         DomainIP ['dj.nailed.org'] := '';                                    }
{         DomainIP ['dummy.ods.org'] := '127.0.0.1';                           }
{         DeactivateDomains;                                                   }
{       finally                                                                }
{         Free;                                                                }
{       end;                                                                   }
{                                                                              }
const
  // Status codes
  DynDomain_Connecting    = -1;
  DynDomain_ConnectFailed = -2;
  DynDomain_Connected     = -3;
  DynDomain_ServerBusy    = -4;
  DynDomain_LoggingIn     = -5;
  DynDomain_LoggedIn      = -6;
  DynDomain_LoginFailed   = -7;
  DynDomain_Deactivating  = -8;
  DynDomain_Deactivated   = -9;
  DynDomain_Activating    = -10;
  DynDomain_Activated     = -11;
  DynDomain_UpdateFailed  = -12;
  DynDomain_Disconnected  = -13;

  // additional ods.org status codes
  ODS_RPL_ServerReady          = 100;
  ODS_RPL_ServerBusy           = 999;
  ODS_RPL_LoginSuccessful      = 225;
  ODS_RPL_AdminLoginSuccessful = 226;
  ODS_RPL_AddedOK              = 795;
  ODS_RPL_BadDomainName        = 797;
  ODS_RPL_DeletedOK            = 901;
  ODS_RPL_UnableToDelete       = 903;

  // additional eyep.net status codes
  EYEP_RPL_MissingField        = 1;
  EYEP_RPL_BadUsernamePassword = 2;
  EYEP_RPL_InvalidHostForUser  = 3;
  EYEP_RPL_UpdateOK            = 4;
  EYEP_RPL_InvalidIP           = 5;
  EYEP_RPL_TemporarilyDown     = 6;

type
  { ADynDomainService                                                          }
  ADynDomainService = class;
  TDynDomainServiceStatusEvent = Procedure (const Sender : ADynDomainService; const StatusCode : Integer; const StatusMessage : String) of object;
  ADynDomainService = class
    protected
    FOnStatus : TDynDomainServiceStatusEvent;

    Procedure SetStatus (const StatusCode : Integer; const StatusMsg : String);

    public
    UpdateOK  : Boolean;
    UpdateMsg : String;

    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String;
              const WildSubDomain, BackupMX : Boolean); virtual; abstract;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); virtual; abstract;
    Property  OnStatus : TDynDomainServiceStatusEvent read FOnStatus write FOnStatus;
  end;
  CDynDomainService = class of ADynDomainService;

  TTextDynDomainService = class (ADynDomainService)
    protected
    FConn : TTextParser;

    Procedure Connect (const Host : String; const Port : Integer); virtual;
    Procedure Disconnect; virtual;

    public
    Destructor Destroy; override;
  end;

  ThttpDynDomainService = class (ADynDomainService)
    protected
    FClient  : ThttpClient;
    FRequest : ThttpRequest;

    Procedure Connect (const Host : String; const Port : Integer = 80;
              const URI : String = ''; const Referer : String = '';
              const BasicAuth : String = ''; const Method : String = http_GET);
    Procedure Disconnect;
    Function  DoRequest : ThttpResponse;

    public
    Destructor Destroy; override;
  end;

  { TODSOrgService                                                             }
  TODSOrgService = class (TTextDynDomainService)
    protected

    Function GetReply : Integer;
    Function StatusCodeAsMessage (const StatusCode : Integer) : String;
    Procedure Login (const Username, Password : String);
    Procedure Disconnect; override;

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TEyepNetService                                                            }
  TEyepNetService = class (TTextDynDomainService)
    protected
    Function  StatusCodeAsMessage (const StatusCode : Integer) : String;
    Function  GetReply : Integer;

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TYIOrgService                                                              }
  TYIOrgService = class (ThttpDynDomainService)
    protected
    Function CheckResponse (const Domain, IP : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TDynDNSOrgService                                                          }
  TDynDNSOrgService = class (ThttpDynDomainService)
    protected
    Function CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { THNOrgService                                                              }
  THNOrgService = class (ThttpDynDomainService)
    protected
    Function CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TDHSOrgService                                                             }
  TDHSOrgService = class (ThttpDynDomainService)
    protected
    Function CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TNS1NetService                                                             }
  TNS1NetService = class (ThttpDynDomainService)
    protected
    Function CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TDynsCXService                                                             }
  TDynsCXService = class (ThttpDynDomainService)
    protected
    Function CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TDtDNSNetService                                                           }
  TDtDNSNetService = class (ThttpDynDomainService)
    protected
    Function CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TDynoDNSService                                                            }
  TDynoDNSService = class (ThttpDynDomainService)
    protected
    Function  CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
    Procedure ConnectDynoDNS (const Domain, Password : String);

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TCJBNetService                                                             }
  TCJBNetService = class (ThttpDynDomainService)
    protected
    Function  CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
    Procedure ConnectDynoDNS (const Domain, Password : String);

    public
    Procedure ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean); override;
    Procedure DeactivateDomain (const Domain, UserName, Password, OfflineURL : String); override;
  end;

  { TDynDomainUpdater                                                          }
  TDomainEntry = record
    ServiceType    : CDynDomainService;
    Domain         : String;
    IP             : String;
    MXAddress      : String;
    BackupMX       : Boolean;
    WildSubDomains : Boolean;
    OfflineURL     : String;
    Active         : Boolean;
  end;
  EDynDomainUpdater = class (Exception);
  TDynDomainUpdater = class; // forward
  TDynDomainOnDomainActivated = Procedure (const Sender : TDynDomainUpdater; const Domain : String);
  TDynDomainOnDomainDeactivated = Procedure (const Sender : TDynDomainUpdater; const Domain : String);
  TDynDomainOnStatus = Procedure (const Sender : TDynDomainUpdater; const StatusCode : Integer; const StatusMessage : String) of object;
  TDynDomainUpdater = class
    protected
    FUserName            : String;
    FPassword            : String;
    FDomains             : Array of TDomainEntry;
    FStatusCode          : Integer;
    FStatusMessage       : String;
    FOnStatus            : TDynDomainOnStatus;
    FOnDomainActivated   : TDynDomainOnDomainActivated;
    FOnDomainDeactivated : TDynDomainOnDomainDeactivated;

    Function  DeactivateDomain (const DomainIdx : Integer) : Boolean;
    Function  ActivateDomain (const DomainIdx : Integer) : Boolean;
    Procedure SetStatus (const StatusCode : Integer; const StatusMessage : String);
    Function  GetDomainActive (const Domain : String) : Boolean;
    Procedure SetDomainActive (const Domain : String; const Active : Boolean);
    Function  GetDomainIP (const Domain : String) : String;
    Procedure SetDomainIP (const Domain : String; const IP : String);
    Function  IndexOf (const Domain : String) : Integer;

    public
    Property  UserName : String read FUserName write FUserName;
    Property  Password : String read FPassword write FPassword;

    Procedure AddDynamicDomain (const ServiceType : CDynDomainService;
              const Domain : String; const IP : String = ''; const MXAddress : String = '';
              const OfflineURL : String = ''; const BackupMX : Boolean = False; const WildSubDomains : Boolean = False);
    Procedure RemoveDynamicDomain (const Domain : String);
    Procedure ActivateDomains;
    Procedure DeactivateDomains;
    Property  DomainActive [const Domain : String] : Boolean read GetDomainActive write SetDomainActive;
    Property  DomainIP [const Domain : String] : String read GetDomainIP write SetDomainIP;

    Property  OnStatus : TDynDomainOnStatus read FOnStatus write FOnStatus;
    Property  OnDomainActivated : TDynDomainOnDomainActivated read FOnDomainActivated write FOnDomainActivated;
    Property  OnDomainDeactivated : TDynDomainOnDomainDeactivated read FOnDomainDeactivated write FOnDomainDeactivated;
    Property  StatusCode : Integer read FStatusCode;
    Property  StatusMessage : String read FStatusMessage;
  end;

implementation

uses
  // Delphi Fundamentals
  cUtils,
  cSysUtils,
  cStrings,
  cInternetStandards;



Function ActivatedMessage (const DomainName : String; const Activated : Boolean) : String;
  Begin
    Result := DomainName + Cond (Activated, ' activated.', ' deactivated.');
  End;


{                                                                              }
{ ADynDomainService                                                            }
{                                                                              }
Procedure ADynDomainService.SetStatus (const StatusCode : Integer; const StatusMsg : String);
  Begin
    if Assigned (FOnStatus) then
      FOnStatus (self, StatusCode, StatusMsg);
  End;



{                                                                              }
{ TTextDynDomainService                                                        }
{                                                                              }
Destructor TTextDynDomainService.Destroy;
  Begin
    FreeAndNil (FConn);
    inherited Destroy;
  End;

Procedure TTextDynDomainService.Connect (const Host : String; const Port : Integer);
var S : String;
  Begin
    S := Host + ':' + IntToStr (Port);
    SetStatus (DynDomain_Connecting, 'Connecting to ' + S + '.');
    FConn := TTextParser.Create (TSocketStream.Create (Host, Port, 5 * 60 * 1000, True));
    SetStatus (DynDomain_Connected, 'Connected to ' + S + '.');
  End;

Procedure TTextDynDomainService.Disconnect;
  Begin
    FreeAndNil (FConn);
  End;



{                                                                              }
{ TODSOrgService                                                               }
{   www.ods.org                                                                }
{                                                                              }
const
  ODSOrgServer  = 'update.ods.org';
  ODSOrgPort    = 7070;

Function TODSOrgService.StatusCodeAsMessage (const StatusCode : Integer) : String;
  Begin
    Case StatusCode of
      ODS_RPL_ServerReady : Result := 'Server ready';
      ODS_RPL_ServerBusy           : Result := 'Server busy';
      ODS_RPL_LoginSuccessful      : Result := 'Login successful';
      ODS_RPL_AdminLoginSuccessful : Result := 'Admin Login successful';
      ODS_RPL_AddedOK              : Result := 'Added OK';
      ODS_RPL_DeletedOK            : Result := 'Delete OK';
      ODS_RPL_UnableToDelete       : Result := 'Unable to delete';
      ODS_RPL_BadDomainName        : Result := 'Invalid domain name';
      else Result := 'Code ' + IntToStr (StatusCode);
    end;
  End;

Function TODSOrgService.GetReply : Integer;
var L, Msg : String;
  Begin
    try
      L := FConn.Readln;
    except
      on E : Exception do raise EDynDomainUpdater.Create ('No reply from the server (' + E.Message + ').');
    end;
    try
      Msg := CopyAfter (L, ' ');
      Result := StrToInt (CopyBefore (L, ' '));
      SetStatus (Result, Msg);
    except
      Result := -1;
    end;
  End;

Procedure TODSOrgService.Login (const Username, Password : String);
var ReplyCode : Integer;
  Begin
    FConn.EOLType := [eol_CRLF, eol_EOF];
    if GetReply <> ODS_RPL_ServerReady then
      begin
        SetStatus (DynDomain_ServerBusy, 'Server not ready.');
        raise EDynDomainUpdater.Create ('Server not ready.');
      end;

    SetStatus (DynDomain_LoggingIn, 'Logging in.');
    FConn.Stream.Write ('LOGIN ' + UserName + ' ' + Password + #13);
    ReplyCode := GetReply;
    if (ReplyCode <> ODS_RPL_LoginSuccessful) and (ReplyCode <> ODS_RPL_AdminLoginSuccessful) then
      begin
        SetStatus (DynDomain_LoginFailed, 'Login failed.');
        raise EDynDomainUpdater.Create ('Login failed.');
      end;
    SetStatus (DynDomain_LoggedIn, 'Logged in.');
  End;

Procedure TODSOrgService.Disconnect;
  Begin
    try
      FConn.Stream.Write ('QUIT'#13);
    except end;
    inherited Disconnect;
  End;

Procedure TODSOrgService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var ReplyCode : Integer;
  Begin
    Connect (ODSOrgServer, ODSOrgPort);
    try
      Login (UserName, Password);
      try
        FConn.Stream.Write ('DELRR ' + Domain + ' A' + #13);
        ReplyCode := GetReply;
      except end;
      SetStatus (DynDomain_Activating, 'Activating ' + Domain + '.');

      FConn.Stream.Write ('ADDRR ' + Domain + ' A ' + Cond (IP = '', 'CONNIP', IP) + #13);
      ReplyCode := GetReply;
      UpdateOK := ReplyCode = ODS_RPL_AddedOK;
      if UpdateOK then
        UpdateMsg := ActivatedMessage (Domain, True) else
        begin
          UpdateMsg := Domain + ' update failed; ' + StatusCodeAsMessage (ReplyCode);
          raise EDynDomainUpdater.Create (UpdateMsg);
        end;
    finally
      Disconnect;
    end;
  End;

Procedure TODSOrgService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var ReplyCode : Integer;
  Begin
    Connect (ODSOrgServer, ODSOrgPort);
    try
      Login (UserName, Password);
      SetStatus (DynDomain_Deactivating, 'Deactivating ' + Domain + '.');
      FConn.Stream.Write ('DELRR ' + Domain + ' A' + #13);
      ReplyCode := GetReply;
      UpdateOK := (ReplyCode = ODS_RPL_DeletedOK) or (ReplyCode = ODS_RPL_UnableToDelete);
      if UpdateOK then
        UpdateMsg := ActivatedMessage (Domain, False) else
        begin
          UpdateMsg := Domain + ' update failed; ' + StatusCodeAsMessage (ReplyCode);
          raise EDynDomainUpdater.Create (UpdateMsg);
        end;
    finally
      Disconnect;
    end;
  End;



{                                                                              }
{ TEyepNetService                                                              }
{   www.eyep.net                                                               }
{                                                                              }
const
  EyepNetServer = 'eyep.net';
  EyepNetPort   = 80;

Function TEyepNetService.StatusCodeAsMessage (const StatusCode : Integer) : String;
  Begin
    Case StatusCode of
      EYEP_RPL_MissingField        : Result := 'Missing field';
      EYEP_RPL_BadUsernamePassword : Result := 'Authentication failed';
      EYEP_RPL_InvalidHostForUser  : Result := 'Invalid host for user';
      EYEP_RPL_UpdateOK            : Result := 'Update OK';
      EYEP_RPL_InvalidIP           : Result := 'Invalid IP';
      EYEP_RPL_TemporarilyDown     : Result := 'Server busy';
      else Result := 'Code ' + IntToStr (StatusCode);
    end;
  End;

Function TEyepNetService.GetReply : Integer;
var L : String;
  Begin
    try
      L := FConn.Readln;
    except
      on E : Exception do
        raise EDynDomainUpdater.Create ('No reply from the server (' + E.Message + ').');
    end;
    Result := StrToIntDef (CopyAfter (UpperCase (L), 'RETURN'), -1);
    if Result > 0 then
      SetStatus (Result, '');
  End;

Procedure TEyepNetService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var ReplyCode : Integer;
  Begin
    Connect (EyepNetServer, EyepNetPort);
    try
      FConn.EOLType := [eol_CR, eol_EOF];
      FConn.Stream.Write ('GET /update.php/' + UserName + '/' + Password + '/' +
                          Domain + Cond (IP <> '', '/' + IP, '') + c_CRLF);
      ReplyCode := GetReply;
      UpdateOK := ReplyCode = EYEP_RPL_UpdateOK;
      if UpdateOK then
        UpdateMsg := ActivatedMessage (Domain, True) else
        begin
          UpdateMsg := Domain + ' update failed; ' + StatusCodeAsMessage (ReplyCode);
          raise EDynDomainUpdater.Create (UpdateMsg);
        end;
    finally
      Disconnect;
    end;
  End;

Procedure TEyepNetService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var ReplyCode : Integer;
  Begin
    Connect (EyepNetServer, EyepNetPort);
    try
      FConn.EOLType := [eol_CR, eol_EOF];
      FConn.Stream.Write ('GET /update.php/' + UserName + '/' + Password + '/' +
                          Domain + '/0.0.0.0' + c_CRLF);
      ReplyCode := GetReply;
      UpdateOK := ReplyCode = EYEP_RPL_UpdateOK;
      if UpdateOK then
        UpdateMsg := ActivatedMessage (Domain, False) else
        begin
          UpdateMsg := Domain + ' update failed; ' + StatusCodeAsMessage (ReplyCode);
          raise EDynDomainUpdater.Create (UpdateMsg);
        end;
    finally
      Disconnect;
    end;
  End;



{                                                                              }
{ ThttpDynDomainService                                                        }
{                                                                              }
Destructor ThttpDynDomainService.Destroy;
  Begin
    FreeAndNil (FClient);
    FreeAndNil (FRequest);
    inherited Destroy;
  End;

Procedure ThttpDynDomainService.Connect (const Host : String; const Port : Integer; const URI : String; const Referer : String; const BasicAuth : String; const Method : String);
  Begin
    FClient := ThttpClient.CreateEx (Host, Port);
    FRequest := ThttpRequest.Create (Method, URI);
    FRequest.Host := Host;
    if Referer <> '' then
      FRequest.Referer := Referer;
    if BasicAuth <> '' then
      FRequest.BasicAuth := BasicAuth;
    FRequest.UserAgent := 'Mozilla/4.0 (compatible; Fundamentals DynDomain ' + UnitVersion + ')';
    FRequest.Header.AddField ('Pragma', 'no-cache');
  End;

Procedure ThttpDynDomainService.Disconnect;
  Begin
    FreeAndNil (FClient);
  End;

Function ThttpDynDomainService.DoRequest : ThttpResponse;
  Begin
    try
      try
        Result := FClient.Request (FRequest);
      finally
        FRequest := nil;
      end;
    finally
      try
        Disconnect;
      except end;
    end;
  End;


{                                                                              }
{ TYIOrgService                                                                }
{    www.yi.org                                                                }
{                                                                              }
const
  YIOrgServer   = 'www.yi.org';
  YIOrgPort     = 80;

{ <!-- MS3V STATUS:OK HOSTNAME:netcruiser ACT:1 IP:196.31.185.222 MX:na -->    }
{ <!-- MS3V STATUS:ERROR HOSTNAME:netcruiser ACT:0 IP:0.0.0.0 MX:na -->        }
Function TYIOrgService.CheckResponse (const Domain, IP : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
var C, S, RIP : String;
  Begin
    C := Response.Content;
    S := CopyBetween (C, '<!-- MS3V ', ' -->');
    RIP := CopyBetween (S, ' IP:', ' ');
    UpdateOK := (Response.StatusCode = http_OK) and (PosNext ('STATUS:OK', S) > 0);
    if UpdateOK then
      UpdateMsg := ActivatedMessage (Domain, True) else
      begin
        if Response.StatusCode <> http_OK then
          UpdateMsg := httpStatusMessage (Response.StatusCode) + '.' else
          begin
            S := CopyBetween (C, 'An error has occured. <B>', '<');
            if S <> '' then
              UpdateMsg := Domain + ' update failed' + Cond (S <> '', ';'#13 + S + #13 + 'Currently ' + RIP, '.');
            raise EDynDomainUpdater.Create (UpdateMsg);
          end;
      end;
  End;

Procedure TYIOrgService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var Response : ThttpResponse;
    S, RIP   : String;
  Begin
    Connect (YIOrgServer, YIOrgPort, '/bin/dyndns.fcgi?ipaddr=' + IP,
             'http://www.yi.org/',
             UserName + ':' + Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, IP, Response, True);
    finally
      FreeAndNil (Response);
    end;
  End;

Procedure TYIOrgService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var Response : ThttpResponse;
    S, RIP   : String;
  Begin
    Connect (YIOrgServer, YIOrgPort, '/bin/dyndns.fcgi?ipaddr=0.0.0.0',
             'http://www.yi.org/',
             UserName + ':' + Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, '', Response, False);
    finally
      FreeAndNil (Response);
    end;
  End;


{                                                                              }
{ TDynDNSOrgService                                                            }
{                                                                              }
{ from http://support.dyndns.org/dyndns/clients/devel/query.shtml :            }
{   (Host: header, an Authorization: header, and a User-Agent: header)         }
{   GET /nic/update?hostname=yourhost.ourdomain.ext&myip=ipaddress&            }
{       wildcard=OFF&mx=mail.exchanger.ext&backmx=NO HTTP/1.1                  }
{                                                                              }
{ good Update good and successful, IP updated                                  }
{ nochg No changes, update considered abusive                                  }
{                                                                              }
const
  DynDNSOrgServer = 'members.dyndns.org';
  DynDNSOrgPort   = 80;

Function TDynDNSOrgService.CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
var S : String;
  Begin
    S := Response.Content;
    UpdateOK := (Response.StatusCode = http_OK) and
                  (Match ('good', S) or
                   Match ('Update good', S) or
                   Match ('IP updated', S) or
                   Match ('nochg', S) or
                   Match ('No changes', S) or
                   Match ('update considered abusive', S));
    Result := UpdateOK;
    if UpdateOK then
      begin
        if Match ('update considered abusive', S) then
          UpdateMsg := 'Domain updated, considered abusive by server.' else
          UpdateMsg := ActivatedMessage (Domain, Activation) +
                    Cond (Match ('nochg', S) or Match ('No changes', S),
                        ' No changes.', '');
      end else
      begin
        if Response.StatusCode <> http_OK then
          UpdateMsg := httpStatusMessage (Response.StatusCode) + '.' else
        if Match ('notfqdn', S) then
          UpdateMsg := 'A Fully-Qualified Domain Name was not provided.' else
        if Match ('nohost', S) then
          UpdateMsg := Domain + ' does not exist on the server.' else
        if Match ('!yours', S) then
          UpdateMsg := 'Invalid username.' else
        if Match ('abuse', S) then
          UpdateMsg := 'The hostname is blocked for abuse;'#13'Contact www.dyndns.org' else
        if Match ('dnserr', S) then
          UpdateMsg := 'www.dyndns.org has encountered a DNS error.' else
          UpdateMsg := Domain + ' update failed; ' + Trim (S) + '.';
        raise EDynDomainUpdater.Create (UpdateMsg);
      end;
  End;

Procedure TDynDNSOrgService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var Response : ThttpResponse;
  Begin
    Connect (DynDNSOrgServer, DynDNSOrgPort, '/nic/update?' +
             'hostname=' + Domain +
             Cond (IP <> '', '&myip=' + IP, '') +
             Cond (WildSubDomain, '&wildcard=ON', '') +
             Cond (MXAddress <> '', '&mx=' + MXAddress + Cond (BackupMX, '&backmx=YES', ''), ''),
             '', UserName + ':' + Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, True);
    finally
      FreeAndNil (Response);
    end;
  End;

Procedure TDynDNSOrgService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var Response : ThttpResponse;
  Begin
    Connect (DynDNSOrgServer, DynDNSOrgPort, '/nic/update?' +
             'hostname=' + Domain +
             '&myip=0.0.0.0', '',
             UserName + ':' + Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, False);
    finally
      FreeAndNil (Response);
    end;
  End;


{                                                                              }
{ THNOrgService                                                                }
{   www.hn.org                                                                 }
{                                                                              }
const
  HNOrgServer   = 'www.hn.org';
  HNOrgPort     = 80;

{ <!-- DDNS_Response_Code=101 -->                                              }
Function THNOrgService.CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
var S : String;
  Begin
    S := CopyBetween (LowerCase (Response.Content), '<!-- ddns_response_code=', ' -->');
    UpdateOK := (Response.StatusCode = http_OK) and (S = '101');
    if UpdateOK then
      UpdateMsg := Domain + Cond (Activation, ' activated.', ' deactivated.') else
      begin
        if Response.StatusCode <> http_OK then
          UpdateMsg := httpStatusMessage (Response.StatusCode) + '.' else
        if S = '201' then
          UpdateMsg := 'Update failed; Previous update occured less'#13'than 5 minutes ago.' else
        if S = '202' then
          UpdateMsg := 'Update failed; Server error' else
        if S = '203' then
          UpdateMsg := 'Account frozen by hn.org admin' else
        if S = '204' then
          UpdateMsg := 'Account currently locked, try again later.' else
          UpdateMsg := 'Update failed (' + S + ')';
        raise EDynDomainUpdater.Create (UpdateMsg);
      end;
  End;

Procedure THNOrgService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var Response : ThttpResponse;
  Begin
    Connect (HNOrgServer, HNOrgPort, '/vanity/update?' +
                 'ver=1&ip=' + IP,
                 '', UserName + ':' + Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, True);
    finally
      FreeAndNil (Response);
    end;
  End;

Procedure THNOrgService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var Response : ThttpResponse;
  Begin
    Connect (HNOrgServer, HNOrgPort, '/vanity/update?' +
                 'ver=1&ip=0.0.0.0',
                 '', UserName + ':' + Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, False);
    finally
      FreeAndNil (Response);
    end;
  End;



{                                                                              }
{ TDHSOrgService                                                               }
{   www.dhs.org                                                                }
{                                                                              }
const
  DHSOrgServer = 'members.dhs.org';
  DHSOrgPort   = 80;

Function TDHSOrgService.CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
var S : String;
    E : String;
  Begin
    S := Response.Content;
    UpdateOK := (Response.StatusCode = http_OK) and
                (Activation and (PosNext ('done', S) > 0)) or
                (not Activation and
                     ((PosNext ('host offline', S) > 0) or
                      (PosNext ('already offline', S) > 0))
                     );
    if UpdateOK then
      UpdateMsg := Domain + Cond (Activation, ' activated.', ' deactivated.') else
      begin
        if Response.StatusCode <> http_OK then
          UpdateMsg := httpStatusMessage (Response.StatusCode) + '.' else
          UpdateMsg := 'Update failed; ' +
                       Trim (CopyBetween (S, 'An error occured while processing your request: ', ['<', #$A, #$D, #9]));
        raise EDynDomainUpdater.Create (UpdateMsg);
      end;
  End;

Procedure TDHSOrgService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var Response  : ThttpResponse;
    Dom, Host : String;
  Begin
    Dom := CopyAfter (Domain, '.');
    Host := CopyBefore (Domain, '.');
    Connect (DHSOrgServer, DHSOrgPort, '/nic/hosts',
                 'http://members.dhs.org/nic/hosts',
                 UserName + ':' + Password, http_POST);
    FRequest.ContentType := ct_URLEncoded;
    FRequest.Content := 'hostscmd=edit&hostscmdstage=2&type=4' +
                        '&updatetype=Online' +
                        '&ip=' + IP +
                        '&mx=' + MXAddress +
                        '&offline_url=&cloak=&cloak_title=&submit=Update' +
                        '&domain=' + Dom + '&hostname=' + Host;
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, True);
    finally
      FreeAndNil (Response);
    end;
  End;

Procedure TDHSOrgService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var Response  : ThttpResponse;
    Dom, Host : String;
  Begin
    Dom := CopyAfter (Domain, '.');
    Host := CopyBefore (Domain, '.');
    Connect (DHSOrgServer, DHSOrgPort, '/nic/hosts',
                 'http://members.dhs.org/nic/hosts',
                 UserName + ':' + Password, http_POST);
    FRequest.ContentType := ct_URLEncoded;
    FRequest.Content := 'hostscmd=edit&hostscmdstage=2&type=4' +
                        '&updatetype=Offline' +
                        '&ip=0.0.0.0' +
                        '&mx=' +
                        '&offline_url=' + Cond (OfflineURL = '', 'http://127.0.0.1/', OfflineURL) +
                        '&cloak=&cloak_title=&submit=Update' +
                        '&domain=' + Dom + '&hostname=' + Host;
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, False);
    finally
      FreeAndNil (Response);
    end;
  End;



{                                                                              }
{ TNS1NetService                                                               }
{   www.ns1.net                                                                }
{                                                                              }
{ NEW FORMAT =============================================================     }
{ No auto detection!
http://ns1.net//dynmanager.php?do_clientupdate=HI&username=$username&password=$password&hostname=$hostname&domain=$domain&type=$type&record=$record
  =============================================================}

const
  NS1NetServer = 'ns1.net';
  NS1NetPort   = 80;

Function TNS1NetService.CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
var S, L, T : String;
  Begin
    S := Response.Content;
    T := Trim (CopyBetween (S, 'Currently Points to ', [' ', '<']));
    UpdateOK := Response.StatusCode = http_OK;
    if UpdateOK then
      UpdateMsg := Domain + Cond (Activation, ' activated' + Cond (T <> '', ';'#13'Pointing to ' + T, '') + '.',
                   ' deactivated.') else
      begin
        UpdateMsg := httpStatusMessage (Response.StatusCode) + '.';
        raise EDynDomainUpdater.Create (UpdateMsg);
      end;
  End;

Procedure TNS1NetService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var Response  : ThttpResponse;
  Begin
    Connect (NS1NetServer, NS1NetPort, '/dynmanager.php',
                 'http://ns1.net/dynmanager.php', '', http_POST);
    FRequest.ContentType := ct_URLEncoded;
    FRequest.Content :=
                 'do_clientupdate=HI' +
                 '&username=' + UserName +
                 '&password=' + Password +
                 '&hostname=' +
                 '&domain=' + Domain +
                 '&type=A' +
                 '&record=' + IP;
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, True);
    finally
      FreeAndNil (Response);
    end;
  End;

Procedure TNS1NetService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var Response  : ThttpResponse;
  Begin
    Connect (NS1NetServer, NS1NetPort, '/dynmanager.php',
                 'http://ns1.net/dynmanager.php', '', http_POST);
    FRequest.ContentType := ct_URLEncoded;
    FRequest.Content :=
                 'do_clientupdate=HI' +
                 '&username=' + UserName +
                 '&password=' + Password +
                 '&hostname=' +
                 '&domain=' + Domain +
                 '&type=A' +
                 '&record=0.0.0.0';
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, False);
    finally
      FreeAndNil (Response);
    end;
  End;


{                                                                              }
{ TDynsCXService                                                               }
{   www.dyns.cx                                                                }
{   http://www.dyns.cx/postscript.php?username=joe&password=foo&               }
{                      host=worldofjoe  [&ip=1.2.3.4]                          }
{                                                                              }
const
  DynsCXServer = 'www.dyns.cx';
  DynsCXPort   = 80;

Function TDynsCXService.CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
var S : String;
    E : String;
  Begin
    S := Response.Content;
    UpdateOK := (Response.StatusCode = http_OK) and
                (PosNext ('200 ', S) > 0);   // 200 Host worldofjoe updated (111.111.111.111)
    if UpdateOK then
      begin
        E := CopyBetween (S, '(', ')');
        UpdateMsg := Domain + Cond (Activation, ' activated' + Cond (E <> '', ';'#13 + E, '.'),
                     ' deactivated.');
      end else
      begin
        if Response.StatusCode <> http_OK then
          UpdateMsg := httpStatusMessage (Response.StatusCode) + '.' else
          begin
            E := Trim (CopyAfter (S, ' '));
            UpdateMsg := 'Update failed' + Cond (E <> '', '; ' + E, '.');
            raise EDynDomainUpdater.Create (UpdateMsg);
          end;
      end;
  End;

Procedure TDynsCXService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var Response  : ThttpResponse;
    H         : String;
  Begin
    H := CopyBefore (Domain, '.');
    Connect (DynsCXServer, DynsCXPort, '/postscript.php?' +
                 'username=' + UserName +
                 '&password=' + Password +
                 '&host=' + H +
                 Cond (IP = '', '', '&ip=' + IP));
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, True);
    finally
      FreeAndNil (Response);
    end;
  End;

Procedure TDynsCXService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var Response  : ThttpResponse;
    H         : String;
  Begin
    H := CopyBefore (Domain, '.');
    Connect (DynsCXServer, DynsCXPort, '/postscript.php?' +
                 'username=' + UserName +
                 '&password=' + Password +
                 '&host=' + H +
                 '&ip=0.0.0.0');
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, False);
    finally
      FreeAndNil (Response);
    end;
  End;



{                                                                              }
{ TDtDNSNetService                                                             }
{   www.dtdns.net                                                              }
{   Example: autodns.cfm?id=test&pw=test  [&ip=24.3.122.217]                   }
{                                                                              }
{ Server Returning non standard headers!                                       }
const
  DtDNSNetServer = 'www.dtdns.net';
  DtDNSNetPort   = 80;

Function TDtDNSNetService.CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
var S : String;
    E : String;
  Begin
    S := Response.Content;
    E := CopyAfter (S, 'now points to ');
    UpdateOK := (Response.StatusCode = http_OK) and (E <> '');
    if UpdateOK then
      UpdateMsg := Domain + Cond (Activation, ' activated; ' + E, ' deactivated.') else
      begin
        if Response.StatusCode <> http_OK then
          UpdateMsg := httpStatusMessage (Response.StatusCode) + '.' else
          UpdateMsg := 'Update failed; ' + Trim (S);
        raise EDynDomainUpdater.Create (UpdateMsg);
      end;
  End;

Procedure TDtDNSNetService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var Response  : ThttpResponse;
  Begin
    Connect (DtDNSNetServer, DtDNSNetPort, '/autodns.cfm?' +
                 'id=' + UserName +
                 '&pw=' + Password +
                 Cond (IP = '', '', '&ip=' + IP));
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, True);
    finally
      FreeAndNil (Response);
    end;
  End;

Procedure TDtDNSNetService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var Response  : ThttpResponse;
  Begin
    Connect (DtDNSNetServer, DtDNSNetPort, '/autodns.cfm?' +
                 'id=' + UserName +
                 '&pw=' + Password +
                 '&ip=0.0.0.0');
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, False);
    finally
      FreeAndNil (Response);
    end;
  End;



{ DynoDNS Service                                                              }
{ host.dynodns.net:password "http://www.dynodns.net/pr/updatens.cgi"           }
{		match = re.search(r"resolve\ to\ (?P<ipaddress>)<br>"                      }
const
  DynoDNSServer = 'www.dynodns.net';
  DynoDNSPort   = 80;

Function TDynoDNSService.CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
var S : String;
    E : String;
  Begin
    S := Response.Content;
    UpdateOK := (Response.StatusCode = http_OK) and
                (PosNext ('resolve to', S) > 0);
    if UpdateOK then
      begin
        E := CopyBetween (S, 'resolve to ', '<br>');
        UpdateMsg := Domain + Cond (Activation, ' activated' + Cond (E <> '', ';'#13 + E, '.'),
                     ' deactivated.');
      end else
      begin
        if Response.StatusCode <> http_OK then
          UpdateMsg := httpStatusMessage (Response.StatusCode) + '.' else
          begin
            E := '';
            UpdateMsg := 'Update failed' + Cond (E <> '', '; ' + E, '.');
            raise EDynDomainUpdater.Create (UpdateMsg);
          end;
      end;
  End;

Procedure TDynoDNSService.ConnectDynoDNS (const Domain, Password : String);
  Begin
    Connect (DynoDNSServer, DynoDNSPort, '/pr/updatens.cgi',
                 Domain + ':' + Password, http_GET);
  End;

Procedure TDynoDNSService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var Response  : ThttpResponse;
  Begin
    ConnectDynoDNS (Domain, Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, True);
    finally
      FreeAndNil (Response);
    end;
  End;

{ Deactivate not implemented on server ????? }
Procedure TDynoDNSService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var Response  : ThttpResponse;
    H         : String;
  Begin
    H := CopyBefore (Domain, '.');
    ConnectDynoDNS (Domain, Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, False);
    finally
      FreeAndNil (Response);
    end;
  End;


{ TCJBNetService                                                               }
{<input name=username maxlength=32>
<input name=password type=password maxlength=16>
<input type=submit value="Update IP">
Referer: http://www.cjb.net/cgi-bin/dynip.cgi }
const
  CJBNetServer = 'www.cjb.net';
  DJBNetPort   = 80;

Function TCJBNetService.CheckResponse (const Domain : String; const Response : ThttpResponse; const Activation : Boolean) : Boolean;
var S : String;
    E : String;
  Begin
    S := Response.Content;
    UpdateOK := (Response.StatusCode = http_OK) and
                (PosNext ('resolve to', S) > 0);
    if UpdateOK then
      begin
        E := CopyBetween (S, 'resolve to ', '<br>');
        UpdateMsg := Domain + Cond (Activation, ' activated' + Cond (E <> '', ';'#13 + E, '.'),
                     ' deactivated.');
      end else
      begin
        if Response.StatusCode <> http_OK then
          UpdateMsg := httpStatusMessage (Response.StatusCode) + '.' else
          begin
            E := '';
            UpdateMsg := 'Update failed' + Cond (E <> '', '; ' + E, '.');
            raise EDynDomainUpdater.Create (UpdateMsg);
          end;
      end;
  End;

Procedure TCJBNetService.ConnectDynoDNS (const Domain, Password : String);
  Begin
    Connect (DynoDNSServer, DynoDNSPort, '/pr/updatens.cgi',
                 Domain + ':' + Password, http_GET);
  End;

Procedure TCJBNetService.ActivateDomain (const Domain, IP, UserName, Password, MXAddress, OfflineURL : String; const WildSubDomain, BackupMX : Boolean);
var Response  : ThttpResponse;
  Begin
    ConnectDynoDNS (Domain, Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, True);
    finally
      FreeAndNil (Response);
    end;
  End;

{ Deactivate not implemented on server ????? }
Procedure TCJBNetService.DeactivateDomain (const Domain, UserName, Password, OfflineURL : String);
var Response  : ThttpResponse;
    H         : String;
  Begin
    H := CopyBefore (Domain, '.');
    ConnectDynoDNS (Domain, Password);
    Response := DoRequest;
    try
      CheckResponse (Domain, Response, False);
    finally
      FreeAndNil (Response);
    end;
  End;




{                                                                              }
{ TDynDomainUpdater                                                            }
{                                                                              }
Procedure TDynDomainUpdater.AddDynamicDomain (const ServiceType : CDynDomainService; const Domain : String; const IP : String; const MXAddress, OfflineURL : String; const BackupMX : Boolean; const WildSubDomains : Boolean);
var L : Integer;
  Begin
    L := Length (FDomains);
    SetLength (FDomains, L + 1);
    FDomains [L].ServiceType := ServiceType;
    FDomains [L].Domain := Domain;
    FDomains [L].IP := IP;
    FDomains [L].MXAddress := MXAddress;
    FDomains [L].OfflineURL := OfflineURL;
    FDomains [L].BackupMX := BackupMX;
    FDomains [L].WildSubDomains := WildSubDomains;
  End;

Procedure TDynDomainUpdater.RemoveDynamicDomain (const Domain : String);
var I, J, L : Integer;
  Begin
    I := IndexOf (Domain);
    if I >= 0 then
      begin
        L := Length (FDomains);
        For J := I to L - 2 do
          FDomains [J] := FDomains [J + 1];
        SetLength (FDomains, L - 1);
      end;
  End;

Function TDynDomainUpdater.IndexOf (const Domain : String) : Integer;
var I : Integer;
  Begin
    For I := 0 to Length (FDomains) - 1 do
      if FDomains [I].Domain = Domain then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function TDynDomainUpdater.GetDomainIP (const Domain : String) : String;
var I : Integer;
  Begin
    I := IndexOf (Domain);
    if I = -1 then
      raise EDynDomainUpdater.Create ('Domain not found.');
    Result := FDomains [I].IP;
  End;

Function TDynDomainUpdater.GetDomainActive (const Domain : String) : Boolean;
var I : Integer;
  Begin
    I := IndexOf (Domain);
    if I = -1 then
      raise EDynDomainUpdater.Create ('Domain not found.');
    Result := FDomains [I].Active;
  End;

Procedure TDynDomainUpdater.SetDomainIP (const Domain : String; const IP : String);
var I : Integer;
  Begin
    I := IndexOf (Domain);
    if I = -1 then
      AddDynamicDomain (TEyepNetService, Domain, IP) else
      FDomains [I].IP := IP;
  End;

Procedure TDynDomainUpdater.SetStatus (const StatusCode : Integer; const StatusMessage : String);
  Begin
    FStatusCode := StatusCode;
    FStatusMessage := StatusMessage;
    if Assigned (FOnStatus) then
      FOnStatus (self, StatusCode, StatusMessage);
  End;

Function TDynDomainUpdater.ActivateDomain (const DomainIdx : Integer) : Boolean;
var ReplyCode : Integer;
    State     : String;
    Upd       : ADynDomainService;
  Begin
    With FDomains [DomainIdx] do
      begin
        State := 'Activating ' + Domain + '.';
        SetStatus (DynDomain_Activating, State);
        Upd := ServiceType.Create;
        try
          Upd.ActivateDomain (Domain, IP, UserName, Password, MXAddress, OfflineURL, WildSubDomains, BackupMX);
          if not Upd.UpdateOK then
            begin
              SetStatus (DynDomain_UpdateFailed, Upd.UpdateMsg);
              raise EDynDomainUpdater.Create (Upd.UpdateMsg);
            end;
          Active := True;
          SetStatus (DynDomain_Activated, Cond (Upd.UpdateMsg <> '', Upd.UpdateMsg,
                     Domain + ' activated.'));
        finally
          FreeAndNil (Upd);
        end;
        if Assigned (FOnDomainActivated) then
          FOnDomainActivated (self, Domain);
      end;
  End;

Function TDynDomainUpdater.DeactivateDomain (const DomainIdx : Integer) : Boolean;
var ReplyCode : Integer;
    Upd       : ADynDomainService;
  Begin
    With FDomains [DomainIdx] do
      begin
        SetStatus (DynDomain_Deactivating, 'Deactivating ' + Domain + '.');
        Upd := ServiceType.Create;
        try
          Upd.DeactivateDomain (Domain, UserName, Password, OfflineURL);
          Active := False;
          SetStatus (DynDomain_Deactivated, Cond (Upd.UpdateMsg <> '', Upd.UpdateMsg,
                     Domain + ' deactivated.'));
        finally
          FreeAndNil (Upd);
        end;
        if Assigned (FOnDomainDeactivated) then
          FOnDomainDeactivated (self, Domain);
      end;
  End;

Procedure TDynDomainUpdater.SetDomainActive (const Domain : String; const Active : Boolean);
var I : Integer;
  Begin
    I := IndexOf (Domain);
    if I = -1 then
      begin
        AddDynamicDomain (TEyepNetService, Domain, '');
        I := Length (FDomains) - 1;
      end;

    if Active then
      ActivateDomain (I) else
      DeactivateDomain (I);
  End;

Procedure TDynDomainUpdater.ActivateDomains;
var I : Integer;
    Msg : String;
  Begin
    if Length (FDomains) = 0 then
      raise EDynDomainUpdater.Create ('No domains set.');

    Msg := '';
    For I := 0 to Length (FDomains) - 1 do
      With FDomains [I] do
        if not Active then
          try
            ActivateDomain (I);
          except
            on E : Exception do
              Msg := Msg + Cond (Msg <> '', #13#10, '') + Domain + ' (' + E.Message + ')';
          end;
    if Msg <> '' then
      raise EDynDomainUpdater.Create ('Domains not activated:' + #13#10 + Msg);
  End;

Procedure TDynDomainUpdater.DeactivateDomains;
var I : Integer;
    Msg : String;
  Begin
    Msg := '';
    For I := 0 to Length (FDomains) - 1 do
      With FDomains [I] do
        if not Active then
          try
            DeactivateDomain (I);
          except
            on E : Exception do
              Msg := Msg + Cond (Msg <> '', #13#10, '') + Domain + ' (' + E.Message + ')';
          end;
    if Msg <> '' then
      raise EDynDomainUpdater.Create ('Some or all domains not deactivated:' + #13#10 + Msg);
  End;

end.

{ TO ADD ===================================================================== }
{ MYIP.ORG }
{ Login ------------ }
GET http://www.myip.org/cgi-bin/Login.py HTTP/1.1
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, */*
Referer: http://www.myip.org/cgi-bin/AddUser.py
Accept-Language: en-gb
Accept-Encoding: gzip, deflate
User-Agent: Mozilla/4.0 (compatible; MSIE 5.01; Windows NT)
Host: www.myip.org
Proxy-Connection: Keep-Alive
Cookie: dynid=DAVIDECOZA; dynpwd=fonebone

{ Update -------------- }
<form method=post action=/cgi-bin/Update.py>
<input type=hidden name=id value=DAVIDECOZA>
<input type=hidden name=pwd value=fonebone>
Hostname:
<select name=hostname>
<option>all</option>
</select>
<br>
Location: <input type=name name=ip value=196.31.187.42>
<br>
Record Type: <select name=recordtype><option>A<option>NS<option>MX</select>
<br>
<input type=submit value=Update>
</form>
</blockquote>
</td>
<td>
Add Host:
<blockquote>
<form method=post action=/cgi-bin/AddHost.py>
<input type=hidden name=id value=DAVIDECOZA>
<input type=hidden name=pwd value=fonebone>
Hostname: <input type=name name=host>
<br>
Domain: 
<select name=domain>
<option>myip.org
</select>
<br>
<input type=submit value=Add>
</form>
</blockquote>
</td></tr>
<tr><td>
Delete Host:
<blockquote>
<form method=post action=/cgi-bin/DelHost.py>
<input type=hidden name=id value=DAVIDECOZA>
<input type=hidden name=pwd value=fonebone>
<select name=hostname>
</select>
<br>
<input type=submit value=Delete>
</form>
</blockquote>
</td>
<td>
Change Password:
<blockquote>
<form method=post action=/cgi-bin/PWD.py>
<input type=hidden name=id value=DAVIDECOZA>
Password: <input type=password name=pwd value="">
<br>New Password: <input type=password name=n_pwd>
<br>New Password (confirm): <input type=password name=n_pwd2>
<br><input type=submit value=Change>
</form>
</blockquote>
</td></tr>
<tr><td align=center valign=middle>
Delete User: 
<form method=post action=/cgi-bin/DelUser.py>
<input type=hidden name=id value=DAVIDECOZA>
<input type=hidden name=pwd value=fonebone>
<input type=submit value=Delete>
</form>
</td>
<td align=center colspan=2>
<form method=post action=/cgi-bin/Login.py>
<input type=hidden name=id value=DAVIDECOZA>
<input type=hidden name=pwd value=fonebone>
<input type=hidden name=set-cookies value=1>
<input type=submit value="Set Cookies">
</form>




{ TO ADD ===================================================================== }
{ NO-IP.COM }
POST http://www.no-ip.com/members/dns/?sid=2589dfc81ccb9e83b1e3597ad484b9e2&Page=mod HTTP/1.1
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, */*
Referer: http://www.no-ip.com/members/dns/?sid=2589dfc81ccb9e83b1e3597ad484b9e2&Page=mod
Accept-Language: en-gb
Content-Type: application/x-www-form-urlencoded
Accept-Encoding: gzip, deflate
User-Agent: Mozilla/4.0 (compatible; MSIE 5.01; Windows NT)
Host: www.no-ip.com
Content-Length: 96
Proxy-Connection: Keep-Alive
Pragma: no-cache

hosts=netcruiser.no-ip.com&IP=196.31.187.42&ip_type=dyn&host_type=on&mod=submitted&Submit=Submit



Response:
<p class=info>Host <B></B> has been modified!<br><br>Rember updates to dynamic hosts take around 5 minutes.</p></td></tr></table>




