{ Conditional defines:                                                         }
{    DEBUG                                                                     }
{    DEBUGLOG - Call log event with extra debugging info                       }
{$I cHeader.inc}
unit cHTTP;

interface

uses
  // System units
  SysUtils,
  Classes,
  scktcomp,

  // L0 units
  cInternetStandards,
  cStreams,
  cTypes,
  cDataStructs;



{                                                                              }
{ HTTP Protocol v0.08 (L1)                                                     }
{                                                                              }
{   Implementation of the HTTP Protocol for clients and servers.               }
{                                                                              }
{ This unit is copyrighted  2000 by David Butler (david@e.co.za)              }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send modifications, suggestions and bug        }
{ reports to david@e.co.za                                                     }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   2000/03/11  v0.01  Initial version.                                        }
{   2000/03/26  v0.02  Exception handling for ThttpClientServerThread.         }
{   2000/04/28  v0.03  Moved HTTP code from cInternetStandards.                }
{   2000/05/07  v0.04  Added ThttpMessage.                                     }
{   2000/07/05  v0.05  Back into compilable state after some changes to L0.    }
{   2000/08/27  v0.06  Added progress events.                                  }
{   2000/08/28  v0.07  Revised unit.                                           }
{   2000/09/24  v0.08  Support form "chunked" transfer-encoding in THTTPClient }
{                                                                              }
{ From RFC2616 (Hypertext Transfer Protocol -- HTTP/1.1)                       }
{ From RFC2396 (Uniform Resource Identifiers (URI): Generic Syntax)            }
{                                                                              }

const
  { cHTTP unit version number }
  UnitVersion = '0.08';

const
  { HTTP Protocol version numbers }
  http_MajorVersion = 1;
  http_MinorVersion = 1;

const
  { HTTP Methods }
  http_GET  = 'GET';
  http_POST = 'POST';
  http_PUT  = 'PUT';
  http_HEAD = 'HEAD';

const
  { HTTP Status code classes }
  http_Informational        = 1;     // 1xx Informational
  http_Successful           = 2;     // 2xx Successful
  http_Redirection          = 3;     // 3xx Redirection
  http_ClientError          = 4;     // 4xx Client Error
  http_ServerError          = 5;     // 5xx Server Error

  { HTTP Status codes }
  http_Continue             = 100;
  http_SwitchingProtocols   = 101;
  http_OK                   = 200;
  http_Created              = 201;
  http_Accepted             = 202;
  http_NonAuthorInformation = 203;
  http_NoContent            = 204;
  http_ResetContent         = 205;
  http_PartialContent       = 206;
  http_MultipleChoices      = 300;
  http_MovedPermanently     = 301;
  http_Found                = 302;
  http_SeeOther             = 303;
  http_NotModified          = 304;
  http_UseProxy             = 305;
  http_TemporaryRedirect    = 307;
  http_BadRequest           = 400;
  http_Unauthorized         = 401;
  http_PaymentRequired      = 402;
  http_Forbidden            = 403;
  http_NotFound             = 404;
  http_MethodNotFound       = 405;
  http_NotAcceptable        = 406;
  http_ProxyAuthRequired    = 407;
  http_RequestTimeout       = 408;
  http_Conflict             = 409;
  http_Gone                 = 410;
  http_LengthRequired       = 411;
  http_PreconditionFailed   = 412;
  http_EntityTooLarge       = 413;
  http_URITooLong           = 414;
  http_UnsupportedMediaType = 415;
  http_RangeNotSatisfiable  = 416;
  http_ExpectationFailed    = 417;
  http_InternalServerError  = 500;
  http_NotImplemented       = 501;
  http_BadGateway           = 502;
  http_ServiceUnavailable   = 503;
  http_GatewayTimeOut       = 504;
  http_VersionNotSupported  = 505;

Function httpStatusMessage (const StatusCode : Integer) : String;



{ Ehttp                                                                        }
type
  Ehttp = class (Exception)
    httpStatusCode : Integer;
    Constructor Create (const StatusCode : Integer; const Msg : String);
  end;


{                                                                              }
{ ThttpMessage                                                                 }
{   Base class for client and server messages.                                 }
{                                                                              }
type
  ThttpMessage = class
    protected
    FMajorVersion  : Integer;
    FMinorVersion  : Integer;
    FHeader        : THeader;
    FIncludeHeader : Boolean;
    FIncludeDate   : Boolean;

    Procedure SetHeader (const Header : THeader); virtual;
    Function  GetHeader : THeader; virtual;
    Function  GetAsString : String; virtual;    Procedure SetAsString (const S : String); virtual; abstract;

    public
    Constructor Create (const MajorVersion : Integer = http_MajorVersion;
                        const MinorVersion : Integer = http_MinorVersion;
                        const Header : THeader = nil);
    Destructor Destroy; override;

    Procedure PrepareToSend; virtual;
    Property  MajorVersion : Integer read FMajorVersion;
    Property  MinorVersion : Integer read FMinorVersion;
    Function  IsPreHTTP11Message : Boolean;
    Property  Header : THeader read GetHeader write SetHeader;
    Property  AsString : String read GetAsString write SetAsString;
    Property  IncludeHeader : Boolean read FIncludeHeader write FIncludeHeader;
    Property  IncludeDate : Boolean read FIncludeDate write FIncludeDate;
  end;



{                                                                              }
{ ThttpRequest                                                                 }
{                                                                              }
type
  EhttpRequest = class (Exception);
  ThttpRequest = class (ThttpMessage)
    protected
    FMethod  : String;
    FURI     : String;
    FContent : String;

    Function  GetAsString : String; override;
    Function  GetRanges : RangeArray;
    Function  GetHost : String;
    Procedure SetHost (const Host : String);
    Function  GetReferer : String;
    Procedure SetReferer (const Referer : String);
    Function  GetAuthorization : String;
    Procedure SetAuthorization (const Authorization : String);
    Function  GetBasicAuth : String;
    Procedure SetBasicAuth (const BasicAuth : String);
    Function  GetUserAgent : String;
    Procedure SetUserAgent (const UserAgent : String);
    Function  GetContentType : String;
    Procedure SetContentType (const ContentType : String);

    public
    Constructor Create (const Method, URI : String;
                const MajorVersion : Integer = http_MajorVersion;
                const MinorVersion : Integer = http_MinorVersion;
                const Header : THeader = nil;
                const Content : String = '');
    Procedure PrepareToSend; override;
    Property  Method : String read FMethod;
    Property  URI : String read FURI;
    Property  Ranges : RangeArray read GetRanges;
    Property  Host : String read GetHost write SetHost;
    Property  Content : String read FContent write FContent;
    Property  ContentType : String read GetContentType write SetContentType;
    Property  Referer : String read GetReferer write SetReferer;
    Property  Authorization : String read GetAuthorization write SetAuthorization;
    Property  BasicAuth : String read GetBasicAuth write SetBasicAuth;
    Property  UserAgent : String read GetUserAgent write SetUserAgent;
  end;



{                                                                              }
{ ThttpResponse                                                                }
{                                                                              }
type
  ThttpResponse = class (ThttpMessage)
    protected
    FStatusCode     : Integer;
    FStatusMessage  : String;
    FContent        : String;
    FContentStream  : TExStream;
    FContentRanges  : RangeArray;
    FIncludeContent : Boolean;
    FContentType    : String;
    FServerName     : String;

    Procedure SetContent (const Content : String);
    Procedure SetContentStream (const ContentStream : TExStream);
    Procedure SetStatusCode (const StatusCode : Integer);
    Function  GetAsString : String; override;
    Function  GetStatusMessage : String;
    Function  GetContentLength : Int64;
    Function  CalcContentLength : Int64;
    Function  GetHeaderAsString : String;

    public
    Constructor Create (const StatusCode : Integer = http_OK;
                const MajorVersion : Integer = http_MajorVersion;
                const MinorVersion : Integer = http_MinorVersion;
                const StatusMessage : String = ''); overload;
    Constructor Create (const ResponseLine : String); overload;
    Destructor Destroy; override;

    Procedure PrepareToSend; override;
    Property  StatusCode : Integer read FStatusCode write SetStatusCode;
    Property  StatusMessage : String read GetStatusMessage write FStatusMessage;
    Property  IncludeContent : Boolean read FIncludeContent write FIncludeContent;
    Property  Content : String read FContent write SetContent;
    Property  ContentStream : TExStream read FContentStream write SetContentStream;
    Property  ContentRanges : RangeArray read FContentRanges write FContentRanges;
    Property  ContentType : String read FContentType write FContentType;
    Property  ServerName : String read FServerName write FServerName;
    Property  AsString : String read GetAsString write SetAsString;
    Property  ContentLength : Int64 read GetContentLength;
    Property  HeaderAsString : String read GetHeaderAsString;
  end;



{                                                                              }
{ ThttpServer                                                                  }
{                                                                              }
type
  ThttpServer = class;
  ThttpServerLogType = (slt_Socket,
              slt_Server,
              slt_Referer,
              slt_Request
              {$IFDEF DEBUGLOG}
              , slt_Debug
              {$ENDIF}
              );
  TOnhttpServerLog = Procedure (const Sender : TObject; const LogType : ThttpServerLogType; const LogMsg : String) of object;
  ThttpServerClientThread = class;
  TOnhttpServerRequest = Function (const Sender : ThttpServerClientThread; const Request : THttpRequest) : ThttpResponse of object;
  ThttpServerClientNotifyEvent = Procedure (const Sender : ThttpServer; const Client : ThttpServerClientThread) of object;
  EhttpServer = class (Exception);
  ThttpServer = class (TComponent)
    protected
    FActive           : Boolean;
    FListenPort       : Integer;
    FServerSocket     : TServerSocket;
    FOnLog            : TOnhttpServerLog;
    FOnPreRequest     : TOnhttpServerRequest;
    FOnPostRequest    : TOnhttpServerRequest;
    FOnRequest        : TOnhttpServerRequest;
    FTimeOut          : Integer;
    FActiveClients    : Array of ThttpServerClientThread;
    FOnClientActive   : ThttpServerClientNotifyEvent;
    FOnClientInactive : ThttpServerClientNotifyEvent;
    FOnClientProgress : ThttpServerClientNotifyEvent;

    Procedure OnGetThread (Sender : TObject; ClientSocket : TServerClientWinSocket;
              var SocketThread : TServerClientThread);
    Procedure SetActive (const Active : Boolean);
    Procedure SetListenPort (const ListenPort : Integer);
    Procedure Log (const LogType : ThttpServerLogType; const LogMsg : String);
    Procedure RaiseError (const Msg : String);
    Procedure OnServerClientActive (const Sender : ThttpServerClientThread);
    Procedure OnServerClientInactive (const Sender : ThttpServerClientThread);
    Procedure OnServerClientProgress (const Sender : ThttpServerClientThread);

    public
    Constructor Create (AOwner : TComponent); override;
    Destructor Destroy; override;

    Property Active : Boolean read FActive write SetActive;
    Property ListenPort : Integer read FListenPort write SetListenPort;
    Property OnLog : TOnhttpServerLog read FOnLog write FOnLog;
    Property OnPreRequest : TOnhttpServerRequest read FOnPreRequest write FOnPreRequest;
    Property OnPostRequest : TOnhttpServerRequest read FOnPostRequest write FOnPostRequest;
    Property OnRequest : TOnhttpServerRequest read FOnRequest write FOnRequest;
    Property TimeOut : Integer read FTimeOut write FTimeOut;
    Property OnClientActive : ThttpServerClientNotifyEvent read FOnClientActive write FOnClientActive;
    Property OnClientInactive : ThttpServerClientNotifyEvent read FOnClientInactive write FOnClientInactive;
    Property OnClientProgress : ThttpServerClientNotifyEvent read FOnClientProgress write FOnClientProgress;
  end;

  { ThttpServerClientThread                                                    }
  ThttpServerClientThreadNotifyEvent = Procedure (const Sender : ThttpServerClientThread) of object;
  ThttpServerClientThread = class (TServerClientThread)
    protected
    FServer     : ThttpServer;
    FClient     : TExStream;
    FClientP    : TTextParser;
    FOnLog      : TOnhttpServerLog;
    FTimeOut    : Integer;
    FLogType    : ThttpServerLogType;
    FLogMsg     : String;
    FOnActive   : ThttpServerClientThreadNotifyEvent;
    FOnInactive : ThttpServerClientThreadNotifyEvent;
    FOnProgress : ThttpServerClientThreadNotifyEvent;

    FRemoteAddress    : String;
    FRemoteHost       : String;
    FProgressPosition : Int64;
    FProgressTotal    : Int64;
    FRequest          : ThttpRequest;
    FResponse         : ThttpResponse;
    FTransferStart    : LongWord;

    Procedure NotifyLog;
    Procedure Log (const LogType : ThttpServerLogType; const LogMsg : String);
    Procedure ExecCGIScript (const CGIAppPath : String);
    Procedure NotifyActive;
    Procedure NotifyInactive;
    Procedure NotifyProgress;
    Procedure NotifyPreRequest;
    Procedure NotifyPostRequest;
    Procedure NotifyRequest;
    Procedure OnClientWriteProgress (const Sender : TExStream; const Position, Total : Int64; var Abort : Boolean);
    Function  GetProgressPercent : Integer;
    Function  GetTransferRate : String;

    public
    Constructor Create (const Server : ThttpServer;
                const ClientSocket : TServerClientWinSocket);
    Destructor Destroy; override;
    Procedure ClientExecute; override;
    Procedure DoTerminate; override;

    Property OnLog : TOnhttpServerLog read FOnLog write FOnLog;
    Property TimeOut : Integer read FTimeOut write FTimeOut;
    Property Client : TExStream read FClient;
    Property RemoteAddress : String read FRemoteAddress;
    Property RemoteHost : String read FRemoteHost;
    Property Request : ThttpRequest read FRequest;
    Property Response : ThttpResponse read FResponse;
    Property OnActive : ThttpServerClientThreadNotifyEvent read FOnActive write FOnActive;
    Property OnInactive : ThttpServerClientThreadNotifyEvent read FOnInactive write FOnInactive;
    Property OnProgress : ThttpServerClientThreadNotifyEvent read FOnProgress write FOnProgress;
    Property ProgressPosition : Int64 read FProgressPosition;
    Property ProgressTotal : Int64 read FProgressTotal;
    Property ProgressPercent : Integer read GetProgressPercent;
    Property TransferRate : String read GetTransferRate;
  end;

Function  httpDecodeURI (const URI : String) : String;
Procedure httpDecodeRequestLine (const R : String;
                                var MajorVersion, MinorVersion : Integer;
                                var Method, URI : String);
Procedure httpDecodeURL (const URL : String; var Protocol, Address, Path : String);
Function  httpDecodeRangeField (const Range : String) : RangeArray;


{                                                                              }
{ ThttpClient                                                                  }
{                                                                              }
type
  ThttpClientLogType = (clt_Socket,
              clt_Client
              {$IFDEF DEBUGLOG}
              , clt_Debug
              {$ENDIF}
              );
  ThttpClient = class;
  TOnhttpClientLog = Procedure (const Sender : ThttpClient; const LogType : ThttpClientLogType; const LogMsg : String) of object;
  ThttpClient = class
  protected
    FClient       : TSocketStream;
    FHost         : String;
    FPort         : Integer;
    FTimeOut      : Integer;
    FActive       : Boolean;
    FOnLog        : TOnhttpClientLog;

    Procedure Log (const LogType : ThttpClientLogType; const LogMsg : String);
    Function  GetRemoteHost : String;
    Procedure SetRemoteHost (const Host : String);
    Function  GetRemoteAddr : String;
    Procedure SetRemoteAddr (const Address : String);
    Function  GetRemotePort : Integer;
    Procedure SetRemotePort (const Port : Integer);
    Procedure SetActive (const Active : Boolean);

  public
    Procedure Start;
    Procedure Stop;

    Constructor Create;
    Constructor CreateEx (const RemoteHost : String; const RemotePort : Integer = 80);
    Destructor Destroy; override;

    Property Active : Boolean read FActive write SetActive;

    Property RemoteHost : String read GetRemoteHost write SetRemoteHost;
    Property RemoteAddr : String read GetRemoteAddr write SetRemoteAddr;
    Property RemotePort : Integer read GetRemotePort write SetRemotePort;
    Property OnLog : TOnhttpClientLog read FOnLog write FOnLog;

    Function Request (const Req : ThttpRequest) : ThttpResponse;
  end;

  EhttpClientThread = class (Exception);
  ThttpClientThread = class (TThread)
    protected
    FClient   : ThttpClient;
    FRequest  : ThttpRequest;
    FResponse : ThttpResponse;
    FError    : String;

    Procedure Execute; override;

    public
    Constructor Create (const Host : String; const Port : Integer);
    Destructor Destroy; override;

    Function Request (const Req : ThttpRequest) : ThttpResponse;
    Property Error : String read FError;
    Property Client : ThttpClient read FClient;
  end;



implementation



uses
  Windows,
  Math,

  // L0 units
  cUtils,
  cMaths,
  cStrings,
  cDateTime,   // High-precision timer
  cHTML;       // htmlDocument

const
  CRLF = ASCII_CR + ASCII_LF;
  SP   = ASCII_SP;

  cHTTP_ServerName = 'Fundamentals-http/' + UnitVersion;

  MaxLineLength = 8192;  // Max length of a single CRLF delimited line in a request/response



{ HTTP Status Codes Encoding                                                   }
Function httpStatusMessage (const StatusCode : Integer) : String;
type
  ThttpStatusCode = record
    Code   : Integer;
    Status : String;
  end;
const
  httpStatusCodes = 41;
  httpStatusCode : Array [1..httpStatusCodes] of ThttpStatusCode = (
      (Code: http_Continue;              Status: 'Continue'),
      (Code: http_SwitchingProtocols;    Status: 'Switching Protocols'),
      (Code: http_OK;                    Status: 'OK'),
      (Code: http_Created;               Status: 'Created'),
      (Code: http_Accepted;              Status: 'Accepted'),
      (Code: http_NonAuthorInformation;  Status: 'Non-Authoritative Information'),
      (Code: http_NoContent;             Status: 'No Content'),
      (Code: http_ResetContent;          Status: 'Reset Content'),
      (Code: http_PartialContent;        Status: 'Partial Content'),
      (Code: http_MultipleChoices;       Status: 'Multiple Choices'),
      (Code: http_MovedPermanently;      Status: 'Moved Permanently'),
      (Code: http_Found;                 Status: 'Found'),
      (Code: http_SeeOther;              Status: 'See Other'),
      (Code: http_NotModified;           Status: 'Not modified'),
      (Code: http_UseProxy;              Status: 'Use Proxy'),
      (Code: 306;                        Status: 'Reserved'),
      (Code: http_TemporaryRedirect;     Status: 'Temporary Redirect'),
      (Code: http_BadRequest;            Status: 'Bad Request'),
      (Code: http_Unauthorized;          Status: 'Unauthorized'),
      (Code: http_PaymentRequired;       Status: 'Payment Required'),
      (Code: http_Forbidden;             Status: 'Forbidden'),
      (Code: http_NotFound;              Status: 'Not Found'),
      (Code: http_MethodNotFound;        Status: 'Method Not Found'),
      (Code: http_NotAcceptable;         Status: 'Not Acceptable'),
      (Code: http_ProxyAuthRequired;     Status: 'Proxy Authentication Required'),
      (Code: http_RequestTimeout;        Status: 'Request Timeout'),
      (Code: http_Conflict;              Status: 'Conflict'),
      (Code: http_Gone;                  Status: 'Gone'),
      (Code: http_LengthRequired;        Status: 'Length Required'),
      (Code: http_PreconditionFailed;    Status: 'Precondition Failed'),
      (Code: http_EntityTooLarge;        Status: 'Request Entity Too Large'),
      (Code: http_URITooLong;            Status: 'Request-URI Too Long'),
      (Code: http_UnsupportedMediaType;  Status: 'Unsupported Media Type'),
      (Code: http_RangeNotSatisfiable;   Status: 'Requested Range Not Satisfiable'),
      (Code: http_ExpectationFailed;     Status: 'Expectation Failed'),
      (Code: http_InternalServerError;   Status: 'Internal Server Error'),
      (Code: http_NotImplemented;        Status: 'Not Implemented'),
      (Code: http_BadGateway;            Status: 'Bad Gateway'),
      (Code: http_ServiceUnavailable;    Status: 'Service Unavailable'),
      (Code: http_GatewayTimeout;        Status: 'Gateway Timeout'),
      (Code: http_VersionNotSupported;   Status: 'HTTP Version Not Supported'));
var I : Integer;
  Begin
    For I := 1 to httpStatusCodes do
      if httpStatusCode [I].Code = StatusCode then
        begin
          Result := httpStatusCode [I].Status;
          break;
        end;
  End;

Function httpStatus (const StatusCode : Integer; const StatusMessage : String = '') : String;
  Begin
    Result := IntToStr (StatusCode) + ' ';
    if StatusMessage <> '' then
      Result := Result + StatusMessage else
      Result := Result + httpStatusMessage (StatusCode);
  End;

Constructor Ehttp.Create (const StatusCode : Integer; const Msg : String);
  Begin
    inherited Create (httpStatus (StatusCode, Msg));
    httpStatusCode := StatusCode;
  End;



{ HTTP Version Number Encoding                                                 }
Function httpVersionNumber (const MajorVersion : Integer = http_MajorVersion;
                            const MinorVersion : Integer = http_MinorVersion) : String;
  Begin
    Result := 'HTTP/' + IntToStr (MajorVersion) + '.' + IntToStr (MinorVersion);
  End;



{                                                                              }
{ ThttpMessage                                                                 }
{                                                                              }
Constructor ThttpMessage.Create (const MajorVersion : Integer; const MinorVersion : Integer; const Header : THeader);
  Begin
    inherited Create;
    FMajorVersion := MajorVersion;
    FMinorVersion := MinorVersion;
    FHeader := Header;
    FIncludeHeader := True;
    FIncludeDate := True;
  End;

Destructor ThttpMessage.Destroy;
  Begin
    FreeAndNil (FHeader);
    inherited Destroy;
  End;

Function ThttpMessage.IsPreHTTP11Message : Boolean;
  Begin
    Result := (FMajorVersion < 1) or
              ((FMajorVersion = 1) and (FMinorVersion < 1));
  End;

Procedure ThttpMessage.SetHeader (const Header : THeader);
  Begin
    FreeAndNil (FHeader);
    FHeader := Header;
  End;

Function ThttpMessage.GetHeader : THeader;
  Begin
    if not Assigned (FHeader) then
      FHeader := THeader.Create;
    Result := FHeader;
  End;

Procedure ThttpMessage.PrepareToSend;
  Begin
    if FIncludeHeader and not Assigned (FHeader) then
      FHeader := THeader.Create;
    if Assigned (FHeader) then
      if FIncludeDate then
        FHeader ['Date'] := TDateField.CreateNow;
  End;

Function ThttpMessage.GetAsString : String;
  Begin
    if FIncludeHeader and Assigned (FHeader) then
      Result := FHeader.AsString else
      Result := '';
  End;



{                                                                              }
{ ThttpRequest                                                                 }
{                                                                              }
Constructor ThttpRequest.Create (const Method, URI : String; const MajorVersion : Integer; const MinorVersion : Integer; const Header : THeader; const Content : String);
  Begin
    inherited Create (MajorVersion, MinorVersion, Header);
    FMethod := Method;
    FURI := URI;
    FContent := Content;
  End;

Function ThttpRequest.GetAsString : String;
  Begin
    Assert (FURI <> '', 'ThttpRequest.GetAsString: No URI specified.');
    Assert (FMethod <> '', 'ThttpRequest.GetAsString: No Method specified.');
    Result := FMethod + SP + FURI + SP + httpVersionNumber (FMajorVersion, FMinorVersion) + CRLF +
              inherited GetAsString +
              FContent;
  End;

Function ThttpRequest.GetHost : String;
  Begin
    Result := Header.FieldBody ['Host'];
  End;

Procedure ThttpRequest.SetHost (const Host : String);
  Begin
    Header.FieldBody ['Host'] := Host;
  End;

Function ThttpRequest.GetReferer : String;
  Begin
    Result := Header.FieldBody ['Referer'];
  End;

Procedure ThttpRequest.SetReferer (const Referer : String);
  Begin
    Header.FieldBody ['Referer'] := Referer;
  End;

Function ThttpRequest.GetAuthorization : String;
  Begin
    Result := Header.FieldBody ['Authorization'];
  End;

Procedure ThttpRequest.SetAuthorization (const Authorization : String);
  Begin
    Header.FieldBody ['Authorization'] := Authorization;
  End;

Function ThttpRequest.GetUserAgent : String;
  Begin
    Result := Header.FieldBody ['User-Agent'];
  End;

Procedure ThttpRequest.SetUserAgent (const UserAgent : String);
  Begin
    Header.FieldBody ['User-Agent'] := UserAgent;
  End;

Function ThttpRequest.GetContentType : String;
  Begin
    Result := Header.FieldBody ['Content-Type'];
  End;

Procedure ThttpRequest.SetContentType (const ContentType : String);
  Begin
    Header.FieldBody ['Content-Type'] := ContentType;
  End;

Function ThttpRequest.GetBasicAuth : String;
  Begin
    Result := GetAuthorization;
    if LowerCase (CopyLeft (Result, 6)) = 'basic ' then  // TODO: support "realm=..."
      Result := MIMEBase64Decode (CopyFrom (Result, 7)) else
      Result := '';
  End;

Procedure ThttpRequest.SetBasicAuth (const BasicAuth : String);
  Begin
    SetAuthorization ('Basic ' + MIMEBase64Encode (BasicAuth));
  End;

Procedure ThttpRequest.PrepareToSend;
  Begin
    inherited PrepareToSend;
    if not IsPreHTTP11Message and (Host = '') then
      raise EhttpRequest.Create ('Host must be set for HTTP/1.1 request.');
    if FContent <> '' then
      Header.FieldBody ['Content-Length'] := IntToStr (Length (FContent));
  End;

Function ThttpRequest.GetRanges : RangeArray;
var S : String;
  Begin
    if Assigned (FHeader) then
      begin
        S := FHeader.FieldBody ['Range'];
        if S <> '' then
          Result := httpDecodeRangeField (S);
      end;
  End;


{ From RFC2396 (Uniform Resource Identifiers (URI): Generic Syntax)            }
Function httpDecodeURI (const URI : String) : String;
var I, J : Integer;
    R    : Boolean;
  Begin
    J := 0;
    Repeat
      I := PosNext ('%', URI, J);
      R := (I > 0) and (I < Length (URI) - 1);
      if R then
        begin
          Result := Result + CopyRange (URI, J + 1, I - 1);
          try
            Result := Result + Char (HexToInt (Copy (URI, I + 1, 2)));
          except
            Result := Result + Copy (URI, I, 3);
          end;
          J := I + 2;
        end;
    Until not R;
    Result := Result + CopyFrom (URI, J + 1);
  End;

Function httpDecodeFileLocation (const Path : String) : String;
  Begin
    Result := Replace ('/', '\', TrimLeft (Path, ['/']));
    if (Length (Result) >= 2) and (Result [2] = '|') then
      Result [2] := ':';
  End;

Procedure httpDecodeURL (const URL : String; var Protocol, Address, Path : String);
var S : String;
    I : Integer;
  Begin
    Address := '';
    Path := '';
    S := httpDecodeURI (URL);
    I := PosNext (':', S);
    if I = 0 then // no protocol
      Protocol := '' else
      begin
        Protocol := LowerCase (CopyLeft (S, I - 1));
        S := CopyFrom (S, I + 1);
      end;

    if Protocol = 'file' then
      Path := httpDecodeFILELocation (S) else
    if Protocol = 'mail' then
      Address := S else
    if (Protocol = 'ftp') or (Protocol = 'http') then
      if Match ('//', S) then
        begin
          Address := CopyBefore (S, '/', True, 3);
          Path := CopyFrom (S, '/', False, 3);
        end else
        Path := S
    else
      Path := S;
  End;

{ eg bytes=500-600,601-999,-50                                                 }
{    bytes=100-                                                                }
Function httpDecodeRangeField (const Range : String) : RangeArray;
var S : String;
    T : StringArray;
    I : Integer;
    L, H : Int64;
  Begin
    S := LowerCase (CopyBefore (Range, '=', False));
    if S <> 'bytes' then
      exit;
    S := CopyAfter (Range, '=');
    T := Split (S, ',');
    For I := 0 to Length (T) - 1 do
      begin
        S := Trim (T [I]);
        L := StrToInt64Def (CopyBefore (S, '-', False), -1);
        H := StrToInt64Def (CopyAfter (S, '-', False), -1);
        Append (ObjectArray (Result), TRange.Create (L, H));
      end;
  End;

Procedure DecodeBasicAuthorization (const Authorization : String; var UserId, Password : String);
var S : String;
  Begin
    S := MIMEBase64Decode (Authorization);
    UserId := CopyBefore (S, ':');
    Password := CopyAfter (S, ':');
  End;

Function EncodeBasicAuthorization (const UserId, Password : String) : String;
  Begin
//    Result := EncodeBase
  End;

{ 401 WWW-Authenticate: Basic realm="..."                                      }
{     Set-Cookie: xxxxx                                                        }
{    Authorization,           // Section 14.8                                  }
{To receive authorization, the client sends the userid and password,
   separated by a single colon (":") character, within a base64  encoded
   string in the credentials.
          basic-credentials = "Basic" SP basic-cookie
          basic-cookie   = <base64 [7] encoding of user-pass,
                           except not limited to 76 char/line>
          user-pass   = userid ":" password }


{ From RFC2616 (Hypertext Transfer Protocol -- HTTP/1.1) 5.1                   }
{   Methods: GET, OPTIONS, HEAD, POST, PUT, DELETE, TRACE, CONNECT             }
{   Returns MajorVersion, MinorVersion = 0 if request not in a valid format.   }
Procedure httpDecodeRequestLine (const R : String;
         var MajorVersion, MinorVersion : Integer;
         var Method, URI : String);
var S, T : String;
    I    : Integer;
  Begin
    // Check version number
    S := Trim (R, SPACE);
    MajorVersion := 0;
    MinorVersion := 0;
    I := PosPrev (SPACE, S);
    if I > 0 then
      begin
        T := UpperCase (CopyFrom (S, I + 1));
        if Match ('HTTP/', T, 1) then
          begin
            S := Trim (CopyLeft (S, I - 1), SPACE);
            T := CopyFrom (T, 6);
            I := PosNext ('.', T);
            if I > 0 then
              try
                MajorVersion := StrToInt (CopyLeft (T, I - 1));
                MinorVersion := StrToInt (CopyFrom (T, I + 1));
              except end;
          end;
      end;

    // Method & URI
    Method := UpperCase (CopyBefore (S, SPACE, True));
    URI := httpDecodeURI (Trim (CopyAfter (S, SPACE), SPACE));
  End;

{ Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF              }
Function httpStatusLine (const StatusCode : Integer; const MajorVersion : Integer = http_MajorVersion;
                            const MinorVersion : Integer = http_MinorVersion;
                            const StatusMessage : String = '') : String;
  Begin
    Result := httpVersionNumber (MajorVersion, MinorVersion) +
              ' ' +
              httpStatus (StatusCode, StatusMessage) +
              CRLF;
  End;

{ ThttpResponse                                                                }
Procedure httpDecodeResponseLine (const R : String;
         var MajorVersion, MinorVersion : Integer;
         var ResponseCode : Integer;
         var ResponseMessage : String);
var S : String;
  Begin
    MajorVersion := 0;
    MinorVersion := 0;
    ResponseCode := 0;
    ResponseMessage := '';

    S := Trim (LowerCase (R), SPACE);

    if not Match ('http/', S, 1) then
      exit;
    MajorVersion := StrToIntDef (CopyBetween (S, '/', '.'), 0);
    MinorVersion := StrToIntDef (CopyBetween (S, '.', SPACE), 0);

    S := CopyAfter (S, SPACE, False);
    ResponseCode := StrToIntDef (CopyBefore (S, SPACE), 0);

    ResponseMessage := CopyAfter (S, SPACE, False);
  End;

Constructor ThttpResponse.Create (const StatusCode : Integer; const MajorVersion : Integer; const MinorVersion : Integer; const StatusMessage : String);
  Begin
    inherited Create (MajorVersion, MinorVersion);
    SetStatusCode (StatusCode);
    FStatusMessage := StatusMessage;
  End;

Constructor ThttpResponse.Create (const ResponseLine : String);
var MajV, MinV, RespC : Integer;
    RespM : String;
  Begin
    httpDecodeResponseLine (ResponseLine, MajV, MinV, RespC, RespM);
    inherited Create (MajV, MinV);
    SetStatusCode (RespC);
    FStatusMessage := RespM;
  End;

Destructor ThttpResponse.Destroy;
  Begin
    FreeAndNil (FContentStream);
    inherited Destroy;
  End;

Procedure ThttpResponse.PrepareToSend;
  Begin
    inherited PrepareToSend;
    FHeader ['Server'] := THeaderField.CreateEx ('Server',
            Cond (FServerName <> '', FServerName + ' ', '') + cHTTP_ServerName);
    if ContentType <> '' then
      FHeader.FieldBody ['Content-Type'] := ContentType;
    if (StatusCode div 100 <> http_Informational) and
       (StatusCode <> http_NoContent) and
       (StatusCode <> http_NotModified) then
      FHeader.FieldBody ['Content-Length'] := IntToStr (CalcContentLength);
  End;

Procedure ThttpResponse.SetContent (const Content : String);
  Begin
    FContent := Content;
    FIncludeContent := True;
  End;

Procedure ThttpResponse.SetContentStream (const ContentStream : TExStream);
  Begin
    FreeAndNil (FContentStream);
    FContentStream := ContentStream;
    FIncludeContent := Assigned (ContentStream);
  End;

Procedure ActualRange (const R : TRange; const Size : Int64; var Position, Count : Int64);
  Begin
    if R.Low = -1 then
      begin
        Count := Min (Size, R.High);
        Position := Size - Count;
      end else
    if R.High = -1 then
      begin
        Count := Max (0, Size - R.Low);
        Position := R.Low;
      end else
    if R.Low < Size then
      begin
        Position := R.Low;
        Count := Min (R.High, Size - 1) - R.Low + 1;
      end else
      begin
        Position := 0;
        Count := 0;
      end;
  End;

Function ThttpResponse.GetContentLength : Int64;
  Begin
    if Header.HasField ('Content-Length') then
      Result := Header.Field ['Content-Length'].BodyAsInteger else
      Result := CalcContentLength;
  End;

Function ThttpResponse.CalcContentLength : Int64;
var I : Integer;
    S, P, C : Int64;
  Begin
    if Assigned (FContentStream) then
      begin
        S := FContentStream.Size;
        if Length (FContentRanges) > 0 then
          begin
            Result := 0;
            For I := 0 to Length (FContentRanges) - 1 do
              begin
                ActualRange (FContentRanges [I], S, P, C);
                Result := Result + C;
              end;
          end else
          Result := S;
      end else
      Result := Length (FContent);
  End;

Procedure ThttpResponse.SetStatusCode (const StatusCode : Integer);
  Begin
    FStatusCode := StatusCode;
    FIncludeDate := StatusCode div 100 <> http_Informational;
  End;

Function ThttpResponse.GetStatusMessage : String;
  Begin
    if FStatusMessage <> '' then
      Result := FStatusMessage else
      Result := httpStatusMessage (FStatusCode);
  End;

Function ThttpResponse.GetHeaderAsString : String;
  Begin
    Result := httpStatusLine (StatusCode) + FHeader.AsString;
  End;

Function ThttpResponse.GetAsString : String;
  Begin
    Result := GetHeaderAsString;
    if IncludeContent then
      Result := Result + Content;
  End;



{                                                                              }
{ ThttpServer                                                                  }
{                                                                              }
Constructor ThttpServer.Create (AOwner : TComponent);
  Begin
    inherited Create (AOwner);
    FListenPort := 80;
    FTimeOut := 60 * 1000; // One minute
  End;

Destructor ThttpServer.Destroy;
  Begin
    FreeAndNil (FServerSocket);
    inherited Destroy;
  End;

Procedure ThttpServer.Log (const LogType : ThttpServerLogType; const LogMsg : String);
  Begin
    if Assigned (FOnLog) then
      FOnLog (self, LogType, LogMsg);
  End;

Procedure ThttpServer.RaiseError (const Msg : String);
  Begin
    raise EhttpServer.Create (Msg);
  End;

Procedure ThttpServer.OnGetThread (Sender : TObject; ClientSocket: TServerClientWinSocket; var SocketThread: TServerClientThread);
var T : ThttpServerClientThread;
    I : Integer;
  Begin
    // Create new thread
    T := ThttpServerClientThread.Create (self, ClientSocket);
    T.OnLog := FOnLog;
    T.TimeOut := FTimeOut;
    T.OnActive := OnServerClientActive;
    T.OnInactive := OnServerClientInactive;
    T.OnProgress := OnServerClientProgress;
    SocketThread := T;

    // Thread cache management
    I := FServerSocket.Socket.ActiveThreads;
    if I > FServerSocket.ThreadCacheSize then
      FServerSocket.ThreadCacheSize := I;

    {$IFDEF DEBUGLOG}
    Log (slt_Debug, 'New thread created (' + IntToStr (I) + ' threads active).');
    {$ENDIF}
  End;

Procedure ThttpServer.SetActive (const Active : Boolean);
  Begin
    if FActive = Active then
      exit;

    if Active then
      begin
        try
          FServerSocket := TServerSocket.Create (self);
          FServerSocket.ServerType := stThreadBlocking;
          FServerSocket.Port := FListenPort;
          FServerSocket.OnGetThread := OnGetThread;
          FServerSocket.Active := True;
        except
          on E : Exception do RaiseError ('Error creating server socket:'#13 + ExceptionToStr (E));
        end;
        FActive := True;
        Log (slt_Server, 'Server active. Listening on port ' + IntToStr (FListenPort) + '.');
      end else
      begin
        try
          FServerSocket.Active := False;
        except
          on E : Exception do Log (slt_Server, 'Error deactivating server socket (' + ExceptionToStr (E) + ')');
        end;
        FreeAndNil (FServerSocket);
        FActive := False;
        Log (slt_Server, 'Server shutdown.');
      end;
  End;

Procedure ThttpServer.SetListenPort (const ListenPort : Integer);
var R : Boolean;
  Begin
    if ListenPort = FListenPort then
      exit;
    Assert (ListenPort > 0, 'ThttpServer.SetListenPort: Invalid ListenPort');

    try
      R := Active;
      try
        if R then
          SetActive (False);
        FListenPort := ListenPort;
      finally
        if R then
          SetActive (True);
      end;
    except
      on E : Exception do RaiseError ('Could not set server port to ' + IntToStr (ListenPort) + ': ' + ExceptionToStr (E));
    end;
    Log (slt_Server, 'Listen port set to ' + IntToStr (ListenPort));
  End;

Procedure ThttpServer.OnServerClientProgress (const Sender : ThttpServerClientThread);
  Begin
    if Assigned (FOnClientProgress) then
      FOnClientProgress (self, Sender);
  End;

Procedure ThttpServer.OnServerClientActive (const Sender : ThttpServerClientThread);
  Begin
    Append (ObjectArray (FActiveClients), Sender);
    if Assigned (FOnClientActive) then
      FOnClientActive (self, Sender);
  End;

Procedure ThttpServer.OnServerClientInactive (const Sender : ThttpServerClientThread);
var I : Integer;
  Begin
    I := PosNext (Sender, ObjectArray (FActiveClients));
    if I >= 0 then
      Remove (ObjectArray (FActiveClients), I, 1, False);
    if Assigned (FOnClientInactive) then
      FOnClientInactive (self, Sender);
  End;

{ ThttpServerClientThread                                                      }
Constructor ThttpServerClientThread.Create (const Server : ThttpServer; const ClientSocket : TServerClientWinSocket);
  Begin
    inherited Create (False, ClientSocket);
    FServer := Server;
    KeepInCache := True;
  End;

Destructor ThttpServerClientThread.Destroy;
  Begin
    FreeAndNil (FRequest);
    inherited Destroy;
  End;

Procedure ThttpServerClientThread.NotifyActive;
  Begin
    if Assigned (FOnActive) then
      FOnActive (self);
  End;

Procedure ThttpServerClientThread.NotifyInactive;
  Begin
    if Assigned (FOnInactive) then
      FOnInactive (self);
  End;

Procedure ThttpServerClientThread.NotifyProgress;
  Begin
    if Assigned (FOnProgress) then
      FOnProgress (self);
  End;

Procedure ThttpServerClientThread.NotifyPreRequest;
  Begin
    if Assigned (FServer.FOnPreRequest) then
      FServer.FOnPreRequest (self, FRequest);
  End;

Procedure ThttpServerClientThread.NotifyPostRequest;
  Begin
    if Assigned (FServer.FOnPostRequest) then
      FServer.FOnPostRequest (self, FRequest);
  End;

Procedure ThttpServerClientThread.NotifyRequest;
  Begin
    if Assigned (FServer.FOnPreRequest) then
      Synchronize (NotifyPreRequest);
    if Assigned (FServer.FOnRequest) then
      FResponse := FServer.FOnRequest (self, FRequest);
    if Assigned (FServer.FOnPostRequest) then
      Synchronize (NotifyPostRequest);
  End;

Procedure ThttpServerClientThread.OnClientWriteProgress (const Sender : TExStream; const Position, Total : Int64; var Abort : Boolean);
  Begin
    FProgressPosition := Position;
    FProgressTotal := Total;

    if Terminated then
      Abort := True else
      if Assigned (FOnProgress) then
        Synchronize (NotifyProgress);
  End;

Procedure ThttpServerClientThread.ClientExecute;
var StartTime : Int64;

  Function ReadLine (var S : String) : Boolean;
    Begin
      S := '';
      Result := False;

      if Terminated then
        exit;

      try
        S := FClientP.Readln;
        Result := True;
      except
        on E : Exception do
          Log (slt_Server, 'Read error (' + ExceptionToStr (E) + ')');
      end;
    End;

  Procedure SendResponse (const Response : ThttpResponse);
  var S : String;
      F : Integer;
      C, P : Int64;
      L : Int64;
      RA : RangeArray;
    Begin
      try try
        if Response.StatusCode = http_OK then
          begin
            RA := Request.Ranges;
            if Length (RA) > 0 then
              begin
                Response.StatusCode := http_PartialContent;
                Response.ContentRanges := RA;
              end;

            if Request.Method = http_HEAD then
              Response.IncludeContent := False;
          end;

        Response.PrepareToSend;
        {$IFDEF DEBUGLOG}
        Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' SEND: PREPARED');
        {$ENDIF}
        if Assigned (Response.FContentStream) then
          begin
            FClient.Write (Response.HeaderAsString);
            {$IFDEF DEBUGLOG}
            Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' SEND: SENT HEADER');
            {$ENDIF}
            if Response.FIncludeContent then
              begin
                if Length (Response.FContentRanges) > 0 then
                  begin
                    L := Response.FContentStream.Size;
                    For F := 0 to Length (Response.FContentRanges) - 1 do
                      begin
                        ActualRange (Response.FContentRanges [F], L, P, C);
                        if C > 0 then
                          begin
                            FTransferStart := GetTickCount;
                            Response.FContentStream.Position := P;
                            FClient.Write (Response.FContentStream, C, 2048);
                          end;
                      end;
                  end
                else
                  begin
                    FTransferStart := GetTickCount;
                    FClient.Write (Response.FContentStream, 2048);
                    {$IFDEF DEBUGLOG}
                    Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' SEND: SENT CONTENT');
                    {$ENDIF}
                  end;
              end;
          end else
          begin
            S := Response.GetAsString;
            FClient.Write (S);
          end;
      except
        on E : Exception do
          begin
            Log (slt_Server, 'Error sending HTTP response code ' + IntToStr (Response.StatusCode) + ' (' + ExceptionToStr (E) + ')');
            ClientSocket.Close;
          end;
      end;
      finally
        Response.Free;
      end;
    End;

var S, Method, URI : String;
    MajVer, MinVer : Integer;
    R              : Boolean;
    Hdr            : THeader;
    PersistentConn : Boolean;

  Begin
    StartTime := StartTimer;
    {$IFDEF DEBUGLOG}
    Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' START:' + IntToStr (Integer (self)));
    {$ENDIF}
    try
      {$IFDEF DEBUGLOG}
      Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' START1:');
      {$ENDIF}
      FRemoteAddress := ClientSocket.RemoteAddress;
      FRemoteHost := ClientSocket.RemoteHost;

      {$IFDEF DEBUGLOG}
      Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' START2:');
      {$ENDIF}
      FClient := TSocketStream.CreateEx (ClientSocket, FTimeOut);
      FClient.OnWriteProgress := OnClientWriteProgress;
      {$IFDEF DEBUGLOG}
      Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' START3:');
      {$ENDIF}
    except
      on E : Exception do Log (slt_Socket, 'Client gone (' + ExceptionToStr (E) + ')');
    end;

    FClientP := TTextParser.Create (FClient);
    {$IFDEF DEBUGLOG}
    Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' START4:');
    {$ENDIF}
    if Assigned (FOnActive) then
      Synchronize (NotifyActive);
    {$IFDEF DEBUGLOG}
    Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' START5:');
    {$ENDIF}

    try try
      FClientP.EOLType := [eol_CRLF];
      FClientP.MaxLineLength := MaxLineLength;
      PersistentConn := True;
      // Request service loop
      While not Terminated and not FClient.EOF and PersistentConn do
        begin
          FProgressPosition := 0;
          FProgressTotal := 0;

          {$IFDEF DEBUGLOG}
          Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' AWAITING REQ:');
          {$ENDIF}
          StartTime := StartTimer;
          {$IFDEF DEBUGLOG}
          Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' ------------- TIMER RESET -------------');
          {$ENDIF}

          // Get request line
          Repeat
            R := ReadLine (S);
          Until (S <> '') or not R or Terminated;
          {$IFDEF DEBUGLOG}
          Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' GOT REQ: ' + S);
          {$ENDIF}
          if not R or Terminated then
            break;

          httpDecodeRequestLine (S, MajVer, MinVer, Method, URI);

          // Get request header (if HTTP/1 compliant request)
          if MajVer >= 1 then
            begin
              Hdr := THeader.Create;
              Repeat
                R := ReadLine (S);
                if S <> '' then
                  try
                    {$IFDEF DEBUGLOG}
                    Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' HDR:' + S);
                    {$ENDIF}
                    Hdr.AddField (S);
                  except end;
              Until (S = '') or not R or Terminated;
              if Terminated then
                break;

              S := LowerCase (Hdr.FieldBody ['Connection']);
              if S = 'keep-alive' then
                PersistentConn := True else
              if S = 'close' then
                PersistentConn := False else
                PersistentConn := not ((MajVer = 1) and (MinVer = 0));

              S := Hdr.FieldBody ['Referer'];
              if S <> '' then
                Log (slt_Referer, FRemoteHost + ';' + S);
            end else
            PersistentConn := False;

          FreeAndNil (FRequest);
          FRequest := ThttpRequest.Create (Method, URI, MajVer, MinVer, Hdr);

          // Get response
          if Assigned (FServer.FOnRequest) then
            NotifyRequest else
            FResponse := nil;
          if not Assigned (FResponse) then
            FResponse := ThttpResponse.Create (http_NotFound);

          {$IFDEF DEBUGLOG}
          Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' RESPONSE PREPARED:');
          {$ENDIF}

          // Log request details
          Log (slt_Request, FRemoteHost + ';' + Method + ';' + URI + ';' + IntToStr (FResponse.StatusCode) + ';' + IntToStr (FResponse.ContentLength));
          if Terminated then
            break;

          // Respond
          SendResponse (FResponse);
          {$IFDEF DEBUGLOG}
          Log (slt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' SENT RESPONSE:');
          {$ENDIF}
        end;
    except
      on E : Ehttp do
        begin
          Log (slt_Server, 'HTTP Error (' + ExceptionToStr (E) + ')');
          SendResponse (ThttpResponse.Create (E.httpStatusCode));
        end;
      on E : Exception do
        begin
          Log (slt_Server, 'Exception (' + ExceptionToStr (E) + ')');
          FResponse := ThttpResponse.Create (http_OK);
          FResponse.Content := htmlDocument ('Internal error', E.ClassName + ': ' + E.Message);
          try
            SendResponse (FResponse);
          except end;
        end;
    end;
    finally
      {$IFDEF DEBUGLOG}
      Log (slt_Debug, '#Fin:' + IntToStr (Integer (self)) + ' Time: ' + IntToStr (MillisecondsElapsed (StartTime)) + 'ms');
      {$ENDIF}
      if Assigned (FOnInactive) then
        Synchronize (NotifyInactive);
      try
//        ClientSocket.Close;
      except
        on E : Exception do Log (slt_Socket, 'Error closing socket: ' + ExceptionToStr (E));
      end;
      FreeAndNil (FClientP);
    end;
  End;

Procedure ThttpServerClientThread.DoTerminate;
  Begin
    inherited DoTerminate;
  End;

Procedure ThttpServerClientThread.NotifyLog;
  Begin
    if Assigned (FOnLog) then
      FOnLog (self, FLogType, FLogMsg);
  End;

Function ThttpServerClientThread.GetProgressPercent : Integer;
  Begin
    if FProgressTotal = 0 then
      Result := 0 else
      Result := Trunc (FProgressPosition / FProgressTotal * 100.0);
  End;

Function ThttpServerClientThread.GetTransferRate : String;
  Begin
    Result := cStrings.TransferRate (FProgressPosition, GetTickCount - FTransferStart);
  End;

Procedure ThttpServerClientThread.Log (const LogType : ThttpServerLogType; const LogMsg : String);
  Begin
    if Assigned (FOnLog) then
      begin
        FLogType := LogType;
        FLogMsg := LogMsg;
        Synchronize (NotifyLog);
      end;
  End;

Procedure ThttpServerClientThread.ExecCGIScript (const CGIAppPath : String);
  Begin
//    SetEnvironmentVariable ('PATH_INFO', );
//    Add('PATH_TRANSLATED', AuxS);
    SetEnvironmentVariable ('REMOTE_HOST', PChar (FRemoteHost));
    SetEnvironmentVariable ('REMOTE_ADDR', PChar (FRemoteAddress));
    SetEnvironmentVariable ('GATEWAY_INTERFACE', 'CGI/1.1');
    SetEnvironmentVariable ('SCRIPT_NAME', PChar (ExtractFileName (CGIAppPath)));
    SetEnvironmentVariable ('REQUEST_METHOD', PChar (FRequest.Method));
    SetEnvironmentVariable ('HTTP_ACCEPT', PChar (FRequest.Header.FieldBody ['Accept']));
//    SetEnvironmentVariable ('HTTP_ACCEPT_CHARSET', PChar (FRequest.Header.FieldBody ['Accept-CharSet')));
//    SetEnvironmentVariable ('HTTP_ACCEPT_ENCODING', d.RequestRequestHeader.AcceptEncoding);
//    SetEnvironmentVariable ('HTTP_ACCEPT_LANGUAGE', d.RequestRequestHeader.AcceptLanguage);
//    SetEnvironmentVariable ('HTTP_FROM', d.RequestRequestHeader.From);
    SetEnvironmentVariable ('HTTP_HOST', PChar (FRequest.Header.FieldBody ['Host']));
    SetEnvironmentVariable ('HTTP_REFERER', PChar (FRequest.Header.FieldBody ['Referer']));
    SetEnvironmentVariable ('HTTP_USER_AGENT', PChar (FRequest.Header.FieldBody ['User-Agent']));
//    SetEnvironmentVariable ('QUERY_STRING', d.URIQuery);
//    SetEnvironmentVariable ('SERVER_SOFTWARE', CServerName);
//    SetEnvironmentVariable ('SERVER_NAME', 'RIT Research Labs');
    SetEnvironmentVariable ('SERVER_PROTOCOL', PChar (httpVersionNumber));
    SetEnvironmentVariable ('SERVER_PORT', PChar (IntToStr (FServer.FListenPort)));
//    SetEnvironmentVariable ('CONTENT_TYPE', d.RequestEntityHeader.ContentType);
//    SetEnvironmentVariable ('CONTENT_LENGTH', d.RequestEntityHeader.ContentLength);
//    SetEnvironmentVariable ('USER_NAME', d.AuthUser);
//    SetEnvironmentVariable ('USER_PASSWORD', d.AuthPassword);
//    SetEnvironmentVariable ('AUTH_TYPE', d.AuthType);

{  FAUTH_TYPE:=GetEnvVariable('AUTH_TYPE');
  FCONTENT_TYPE:=GetEnvVariable('CONTENT_TYPE');
  FCONTENT_LENGTH:=GetEnvVariable('CONTENT_LENGTH');
  FGATEWAY_INTERFACE:=GetEnvVariable('GATEWAY_INTERFACE');
  FPATH_INFO:=GetEnvVariable('PATH_INFO');
  FPATH_TRANSLATED:=GetEnvVariable('PATH_TRANSLATED');
  FREMOTE_ADDR:=GetEnvVariable('REMOTE_ADDR');
  FREMOTE_HOST:=GetEnvVariable('REMOTE_HOST');
  FREMOTE_IDENT:=GetEnvVariable('REMOTE_IDENT');
  FREMOTE_USER:=GetEnvVariable('REMOTE_USER');
  FREQUEST_METHOD:=GetEnvVariable('REQUEST_METHOD');
  FSCRIPT_NAME:=GetEnvVariable('SCRIPT_NAME');
  FSERVER_NAME:=GetEnvVariable('SERVER_NAME');
  FSERVER_PORT:=GetEnvVariable('SERVER_PORT');
  FSERVER_PROTOCOL:=GetEnvVariable('SERVER_PROTOCOL');
  FSERVER_SOFTWARE:=GetEnvVariable('SERVER_SOFTWARE');
  FHTTP_ACCEPT:=GetEnvVariable('HTTP_ACCEPT');
  FHTTP_USER_AGENT:=GetEnvVariable('HTTP_USER_AGENT');
  FHTTP_REFERER:=GetEnvVariable('HTTP_REFERER');
  FHTTP_COOKIE:=GetEnvVariable('HTTP_COOKIE'); }
(*  // Two cases - Get and Post method
  if CompareText(FREQUEST_METHOD,'GET')=0 then begin
    // GET method
    FQUERY_STRING:=GetEnvVariable('QUERY_STRING');
    FCONTENT_LENGTH := IntToStr(Length(FQUERY_STRING));
  end
  else if CompareText(FREQUEST_METHOD,'POST')=0 then begin
    // POST method
    TempLength := StrToIntDef(FCONTENT_LENGTH, 0);
    if TempLength <> 0 then begin
      // We have to read the exact amount of characters as there may be more
      // data on the standard input
      for i :=1 to TempLength do begin
        Read(AChar);
        FQUERY_STRING:=FQUERY_STRING + AChar;
      end;
      FQUERY_STRING := FQUERY_STRING + #0;
    end
    else FQUERY_STRING:='';
  end
  else begin
    // UNKNOWN METHOD
    FQUERY_STRING := '';
  end;
end; *)


  End;

{                                                                              }
{ ThttpClient                                                                  }
{                                                                              }
Procedure ThttpClient.Start;
  Begin
    FClient := TSocketStream.Create (FHost, FPort, FTimeOut, True);
  End;

Procedure ThttpClient.Stop;
  Begin
    FreeAndNil (FClient);
  End;

Constructor ThttpClient.Create;
  Begin
    inherited Create;
    FTimeOut := 60000;
  End;

Constructor ThttpClient.CreateEx (const RemoteHost : String; const RemotePort : Integer);
  Begin
    inherited Create;
    FTimeOut := 60000;
    SetRemoteHost (RemoteHost);
    SetRemotePort (RemotePort);
  End;

Destructor ThttpClient.Destroy;
  Begin
    FreeAndNil (FClient);
    inherited Destroy;
  End;

Procedure ThttpClient.Log (const LogType : ThttpClientLogType; const LogMsg : String);
  Begin
    if Assigned (FOnLog) then
      FOnLog (self, LogType, LogMsg);
  End;

Function ThttpClient.GetRemoteAddr : String;
  Begin
    Result := FHost;
  End;

Procedure ThttpClient.SetRemoteAddr (const Address : String);
  Begin
    FHost := Address;
  End;

Function ThttpClient.GetRemoteHost : String;
  Begin
    Result := FHost;
  End;

Procedure ThttpClient.SetRemoteHost (const Host : String);
  Begin
    FHost := Host;
  End;

Function ThttpClient.GetRemotePort : Integer;
  Begin
    Result := FPort;
  End;

Procedure ThttpClient.SetRemotePort (const Port : Integer);
  Begin
    FPort := Port;
  End;

Procedure ThttpClient.SetActive (const Active : Boolean);
  Begin
    if FActive = Active then
      exit;

    if Active then
      Start else
      Stop;
    FActive := Active;
  End;

Function ThttpClient.Request (const Req : ThttpRequest) : ThttpResponse;
var S, T, U : String;
    SP : TTextParser;
    Response : ThttpResponse;
    L : Integer;
    H : AHeaderField;
    {$IFDEF DEBUGLOG}
    StartTime : Int64;
    {$ENDIF}

  Function ReadChunkSize : Integer;
    Begin
      Result := StrToIntDef ('$' + Trim (CopyBefore (SP.Readln, ';', True)), 0);
    End;

  Begin
    try
      {$IFDEF DEBUGLOG}
      StartTime := StartTimer;
      Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST:');
      {$ENDIF}

      if not FActive then
        Active := True;

      {$IFDEF DEBUGLOG}
      Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST: ACTIVE:');
      {$ENDIF}

      S := RemoteHost;
      if S <> '' then
        Req.Host := S else
        Req.Host := RemoteAddr;
      Req.PrepareToSend;

      {$IFDEF DEBUGLOG}
      Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST: PREPARED:');
      {$ENDIF}

      SP := TTextParser.Create (FClient);
      try
        SP.EOLType := [eol_CRLF, eol_EOF];
        SP.MaxLineLength := MaxLineLength;
        FClient.Write (Req.AsString);
        {$IFDEF DEBUGLOG}
        Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST: WRITTEN:');
        {$ENDIF}
        S := SP.Readln;
        {$IFDEF DEBUGLOG}
        Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST: RESPONSE: ' + S);
        {$ENDIF}
        Response := ThttpResponse.Create (S);
        try
          While not FClient.EOF and (S <> '') do
            begin
              S := SP.Readln;
              {$IFDEF DEBUGLOG}
              Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST: HDR:' + S);
              {$ENDIF}
              if S <> '' then
                try
                  Response.Header.AddField (S);
                except end;
            end;
          {$IFDEF DEBUGLOG}
          Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST: RESPONDED:');
          {$ENDIF}
          H := Response.Header ['Content-Length'];
          if Assigned (H) then
            begin
              L := H.BodyAsInteger;
              if L > 0 then
                Response.Content := FClient.Read (L);
              {$IFDEF DEBUGLOG}
              Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST: GOT CONTENT: LEN=' + IntToStr (Length (Response.Content)));
              {$ENDIF}
            end else
            begin
              H := Response.Header ['Transfer-Encoding'];
              if Assigned (H) and (CopyBefore (LowerCase (H.Body), ';') = 'chunked') then
                begin
                  L := ReadChunkSize;
                  T := '';
                  While L > 0 do
                    begin
                      S := FClient.Read (L);
                      T := T + S;
                      SP.Readln;
                      L := ReadChunkSize;
                    end;

                  Response.Content := T;
                  Repeat
                    S := SP.Readln;
                    {$IFDEF DEBUGLOG}
                    if S <> '' then
                      Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST: FOOTER:' + S);
                    {$ENDIF}
                    if S <> '' then
                      Response.Header.AddField (S);
                  Until FClient.EOF or (S = '');
                  H.Body := '';
                  Response.Header.AddField ('Content-Length', IntToStr (Length (Response.FContent)));
                end else
                begin
                  T := '';
                  Repeat
                    S := FClient.ReadAvailable;
                    if S = '' then
                      try
                        S := FClient.Read (1);
                      except
                        S := '';
                      end;
                    T := T + S;
                  Until S = '';
                  Response.Content := T;
                end;
            end;
        except
          FreeAndNil (Response);
          raise;
        end;
        Result := Response;
      finally
        SP.ReleaseStream;
        FreeAndNil (SP);
      end;
    finally
      Req.Free;
      {$IFDEF DEBUGLOG}
      Log (clt_Debug, IntToStr (MillisecondsElapsed (StartTime)) + ' REQUEST: FIN:');
      {$ENDIF}
    end;
  End;

Constructor ThttpClientThread.Create (const Host : String; const Port : Integer);
  Begin
    inherited Create (True);
    FClient := ThttpClient.Create;
    FClient.RemoteHost := Host;
    FClient.RemotePort := Port;
    FreeOnTerminate := False;
  End;

Destructor ThttpClientThread.Destroy;
  Begin
    Terminate;
    if Suspended then
      begin
        Resume;
        WaitFor;
      end;
    inherited Destroy;
  End;

Procedure ThttpClientThread.Execute;
  Begin
    While not Terminated do
      begin
        try
          FResponse := FClient.Request (FRequest);
          FError := '';
        except
          on E : Exception do FError := ExceptionToStr (E);
        end;
        if not Terminated then
          Suspend;
      end;
  End;

Function ThttpClientThread.Request (const Req : ThttpRequest) : ThttpResponse;
  Begin
    if not Suspended then
      raise EhttpClientThread.Create ('Thread already busy.');
    FRequest := Req;
    Resume;
    Repeat
      Sleep (0);
    Until Terminated or Suspended;
    if not Terminated then
      Result := FResponse else
      raise EhttpClientThread.Create ('Thread terminated.');
  End;


end.








{ ============================================================================ }
{ * TODO:                                                                      }
{                                                                              }
{         * Throttling }
{         * Fewer messages to main loop }
{         * Public: response field 14.35 (special methods) }
{         * Upgrade: response field }
{         * CGI }
{         * Authentication }


{ 401 WWW-Authenticate: Basic realm="..."                                      }
{     Set-Cookie: xxxxx                                                        }
{    Authorization,           // Section 14.8                                  }
{To receive authorization, the client sends the userid and password,
   separated by a single colon (":") character, within a base64  encoded
   string in the credentials.
          basic-credentials = "Basic" SP basic-cookie
          basic-cookie   = <base64 [7] encoding of user-pass,
                           except not limited to 76 char/line>
          user-pass   = userid ":" password }
{    IfModifiedSince,         // Section 14.24
    IfMatch,                 // Section 14.25
    IfNoneMatch,             // Section 14.26
    IfRange,                 // Section 14.27
    IfUnmodifiedSince,       // Section 14.28
    Range,                   // Section 14.36
    Referer,                 // Section 14.37
    UserAgent: string;       // Section 14.42

  TEntityHeader = class
    Allow,                   // Section 14.7
    ContentBase,             // Section 14.11
    ContentEncoding,         // Section 14.12
    ContentLanguage,         // Section 14.13
    ContentLength,           // Section 14.14
    ContentLocation,         // Section 14.15
    ContentMD5,              // Section 14.16
    ContentRange,            // Section 14.17
    ContentType,             // Section 14.18
    ETag,                    // Section 14.20
    Expires,                 // Section 14.21
    LastModified,            // Section 14.29 }
//          Sleep(2000); // delay 2 seconds if authentification failed



    FAccept                     := 'image/gif, image/x-xbitmap, ' +
                                   'image/jpeg, image/pjpeg, */*';
    FAgent                      := 'Mozilla/3.0 (compatible)';

    if FNoCache then
        SendCommand('Pragma: no-cache');
    if FModifiedSince <> 0 then
        SendCommand('If-Modified-Since: ' +
                    RFC1123_Date(FModifiedSince) + ' GMT');
    if (FProxy <> '') and (FProxyUsername <> '') then
        SendCommand('Proxy-Authorization: Basic ' +
                    EncodeStr(encBase64, FProxyUsername + ':' + FProxyPassword));
    if (FContentRangeBegin <> '') or (FContentRangeEnd <> '') then begin            {JMR!! Added this line!!!}
        SendCommand('Range: bytes=' + FContentRangeBegin + '-' + FContentRangeEnd); {JMR!! Added this line!!!}

{ ============================================================================ }
{ * TODO:                                                                      }
{                                                                              }

{ Syntax of an URL: protocol://[user[:password]@]server[:port]/path         }
{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
{ Retrieve a single value by name out of an URL encoded data stream         }
{ In the stream, every space is replaced by a '+'. The '%' character is     }
{ an escape character. The next two are 2 digits hexadecimal codes ascii    }
{ code value. The stream is consitueted by name=value couples separated     }
{ by a single '&' character. The special characters are coded by the '%'    }
{ followed by hex-ascii character code.                                     }
function ExtractURLEncodedValue(


{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
function EncodeLine(
    Encoding : THttpEncoding;
    SrcData  : PChar;
    Size     : Integer) : String;
var
    Offset : Integer;
    Pos1   : Integer;
    Pos2   : Integer;
    I      : Integer;
begin
    SetLength(Result, Size * 4 div 3 + 4);
    FillChar(Result[1], Size * 4 div 3 + 2, #0);

    if Encoding = encUUEncode then begin
        Result[1] := Char(((Size - 1) and $3f) + $21);
        Size      := ((Size + 2) div 3) * 3;
    end;
    Offset := 2;
    Pos1   := 0;
    Pos2   := 0;
    case Encoding of
        encUUEncode:        Pos2 := 2;
        encBase64, encMime: Pos2 := 1;
    end;
    Result[Pos2] := #0;

    while Pos1 < Size do begin
        if Offset > 0 then begin
            Result[Pos2] := Char(ord(Result[Pos2]) or
                                 ((ord(SrcData[Pos1]) and
                                  ($3f shl Offset)) shr Offset));
            Offset := Offset - 6;
            Inc(Pos2);
            Result[Pos2] := #0;
        end
        else if Offset < 0 then begin
            Offset := Abs(Offset);
            Result[Pos2] := Char(ord(Result[Pos2]) or
                                 ((ord(SrcData[Pos1]) and
                                  ($3f shr Offset)) shl Offset));
            Offset := 8 - Offset;
            Inc(Pos1);
        end
        else begin
            Result[Pos2] := Char(ord(Result[Pos2]) or
                                 ((ord(SrcData[Pos1]) and $3f)));
            Inc(Pos2);
            Inc(Pos1);
            Result[Pos2] := #0;
            Offset    := 2;
        end;
    end;

    case Encoding of
    encUUEncode:
        begin
            if Offset = 2 then
                Dec(Pos2);
            for i := 2 to Pos2 do
                Result[i] := bin2uue[ord(Result[i])+1];
        end;
    encBase64, encMime:
        begin
            if Offset = 2 then
                Dec(Pos2);
            for i := 1 to Pos2 do
                Result[i] := bin2b64[ord(Result[i])+1];
            while (Pos2 and 3) <> 0  do begin
                Inc(Pos2);
                Result[Pos2] := '=';
            end;
        end;
    end;
    SetLength(Result, Pos2);
end;

{                                                                              }
{ * TODO:                                                                      }
{ ============================================================================ }

