{$INCLUDE cHeader.inc}
unit cIRC;

{                                                                              }
{ IRC Client v0.05 (L0)                                                        }
{                                                                              }
{   Implements an IRC client that can connect to multiple servers              }
{   simultaneously. Everything runs event driven in the main thread.           }
{                                                                              }
{                                                                              }
{        This unit is copyright  2000 by David Butler (david@e.co.za)         }
{                                                                              }
{                  This unit is part of Delphi Fundamentals.                   }
{                     Its original file name is cIRC.pas                       }
{                                                                              }
{                I invite you to use this unit, free of charge.                }
{        I invite you to distibute this unit, but it must be for free.         }
{             I also invite you to contribute to its development,              }
{             but do not distribute a modified copy of this file.              }
{       Send modifications, suggestions and bug reports to david@e.co.za       }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   21/06/2000  v0.01  Initial version.                                        }
{   25/06/2000  v0.02  Further development.                                    }
{   24/07/2000  v0.03  Channel lists added.                                    }
{   29/07/2000  v0.04  Created IrcConsts.inc.                                  }
{                      Now use FPiette's WSocket instead of Delphi's sockets.  }
{   30/07/2000  v0.05  More complete implementations (added channel modes,     }
{                      channel topics, invites).                               }
{                      Refactored unit.                                        }
{                      Removed TMultiIRCClient.                                }
{                                                                              }
interface

uses
  extctrls, // TTimer
  Graphics, // TColor
  Classes,
  SysUtils,

  // Francois Piette's ICS
  WSocket,

  // Delphi Fundamentals
  cUtils,
  cStrings,
  cDataStructs;



{                                                                              }
{ TIrcClient                                                                   }
{   Handles a single client connection.                                        }
{                                                                              }
type
  TIrcClient = class;
  TChannel = class
    protected
    FClient     : TIrcClient;
    FChannel    : String;
    FTopic      : String;
    FUsers      : StringArray;
    FJoined     : Boolean;
    FKicked     : Boolean;
    FKickReason : String;

    Procedure SetTopic (const Topic : String);
    Procedure SetKicked (const Reason : String);
    Procedure SetJoined;
    Procedure SetNotJoined;

    public
    Constructor Create (const Client : TIrcClient; const Channel : String);

    Function UserIndexByNick (const Nick : String) : Integer;
    Property Client : TIrcClient read FClient;
    Property Channel : String read FChannel;
    Property Topic : String read FTopic write SetTopic;
    Property Users : StringArray read FUsers;
    Property Joined : Boolean read FJoined;
    Property Kicked : Boolean read FKicked;
    Property KickReason : String read FKickReason;
  end;
  TIrcClientNotifyEvent = Procedure (const Sender : TIrcClient) of object;
  TIrcClientChannelNotifyEvent = Procedure (const Sender : TIrcClient;
    const Channel : TChannel) of object;
  TIrcClientMessageEvent = Procedure (const Sender : TIrcClient;
    const Name, User, Host : String;
    const NumericCmd : Integer; const Cmd : String;
    const Parameters : StringArray) of object;
  TIrcClientCTCPMessageEvent = Procedure (const Sender : TIrcClient;
    const Destination, User, Cmd : String;
    const Parameters : StringArray) of object;
  TIrcClientPrivMsgEvent = Procedure (const Sender : TIrcClient;
    const Destination, User, Msg : String) of object;
  TIrcClientNickInUseEvent = Procedure (const Sender : TIrcClient; var NewNick : String) of object;
  TIrcClientUserInfoEvent = Procedure (const Sender : TIrcClient; const LoggedIn : Boolean;
    const Nick, User, Host, RealName : String) of object;
  TIrcClientChannelInviteEvent = Procedure (const Sender : TIrcClient;
    const Channel, InvitedBy : String) of object;
  TIrcClientChannelKickEvent = Procedure (const Sender : TIrcClient;
    const Channel : TChannel; const KickedBy, Reason : String) of object;
  TIrcClientStatusEvent = Procedure (const Sender : TIrcClient;
    const Status : String) of object;
  EIrcClient = class (Exception);
  TChannelArray = Array of TChannel;
  TIrcClient = class
    protected
    FHost      : String;
    FPort      : Integer;
    FUserName  : String;
    FPassword  : String;
    FNick      : String;
    FRealName  : String;

    FOnStatus : TIrcClientStatusEvent;

    // Socket
    FClient : TWSocket;
    FInBuf  : String;

    // Logging in
    FActive           : Boolean;
    FLoggedIn         : Boolean;
    FLoggingIn        : Boolean;
    FOnLoggingIn      : TIrcClientNotifyEvent;
    FOnLoggedIn       : TIrcClientNotifyEvent;
    FOnLoggedOut      : TIrcClientNotifyEvent;
    FLoginTimer       : TTimer;
    FLoginTimeOut     : Integer;
    FLoginRetries     : Integer;
    FConnectRetryWait : Integer;
    FLoginRetryWait   : Integer;
    FLoginAttemptNr   : Integer;
    FAutoReLogin      : Boolean;

    // Received Messages
    FOnMessage   : TIrcClientMessageEvent;
    FOnNickInUse : TIrcClientNickInUseEvent;
    FOnPrivMsg   : TIrcClientPrivMsgEvent;
    FOnNoticeMsg : TIrcClientPrivMsgEvent;
    FOnCTCPMsg   : TIrcClientCTCPMessageEvent;
    FOnCTCPReply : TIrcClientCTCPMessageEvent;

    // Message of the day
    FMOTD    : String;
    FGotMOTD : Boolean;

    // Channel list
    FChannelList            : StringArray;
    FNewChanList            : StringArray;
    FOnChannelList          : TIrcClientNotifyEvent;
    FRefreshChanListPending : Boolean;
    FAwaitingChannelList    : Boolean;

    // Joined channels
    FJoinedChannels          : TChannelArray;
    FOnChannelJoin           : TIrcClientChannelNotifyEvent;
    FOnChannelPart           : TIrcClientChannelNotifyEvent;
    FOnChannelKick           : TIrcClientChannelKickEvent;
    FOnChannelUserListChange : TIrcClientChannelNotifyEvent;
    FOnInvite                : TIrcClientChannelInviteEvent;
    FOnChannelTopic          : TIrcClientChannelNotifyEvent;

    // Buddy checker
    FBuddyList             : StringArray;
    FBuddiesOnline         : StringArray;
    FBuddiesTmp            : StringArray;
    FBuddyTimer            : TTimer;
    FIsOnlineCount         : Integer;
    FOnBuddiesOnlineChange : TIrcClientNotifyEvent;

    // User information
    FOnUserInfo : TIrcClientUserInfoEvent;

    Procedure SetNick (const Nick : String);
    Procedure SetHost (const Host : String);
    Procedure SetPort (const Port : Integer);
    Procedure SetUsername (const Username : String);
    Procedure SetPassword (const Password : String);
    Procedure SetRealName (const RealName : String);

    // Socket
    Procedure OnClientRead (Sender: TObject; Error: word);

    // Loggin in
    Procedure SetActive (const Active : Boolean);
    Procedure OnClientChangeState (Sender: TObject; OldState, NewState : TSocketState);
    Procedure OnClientError (Sender: TObject);
    Procedure OnClientDnsLookupDone (Sender: TObject; Error: Word);
    Function  IsConnected : Boolean;
    Procedure SetLoggedIn;
    Procedure SetLoggedOut;
    Procedure AttemptLogin;
    Procedure LogIn;
    Procedure LogOut (const QuitMessage : String);
    Procedure OnLoginTimeOut (Sender : TObject);
    Procedure OnLoginRetry (Sender : TObject);
    Procedure OnReLogin (Sender : TObject);

    // Channels
    Function  GetChannelIndex (const Channel : String) : Integer;
    Function  GetChannelList : StringArray;
    Function  CreateChannel (const Channel : String) : TChannel; virtual;
    Function  GetJoinedChannels : StringArray;

    // Buddy checker
    Procedure CheckBuddyList;
    Procedure OnBuddyTimer (Sender : TObject);
    Procedure SetBuddyList (const BuddyList : StringArray);

    public
    Constructor Create;
    Destructor Destroy; override;

    Property  Host : String read FHost write SetHost;
    Property  Port : Integer read FPort write SetPort;
    Property  UserName : String read FUserName write SetUserName;
    Property  Password : String read FPassword write SetPassword;
    Property  Nick : String read FNick write SetNick;
    Property  RealName : String read FRealName write SetRealName;
    Property  OnNickInUse : TIrcClientNickInUseEvent read FOnNickInUse write FOnNickInUse;

    Property  OnStatus : TIrcClientStatusEvent read FOnStatus write FOnStatus;

    // LoginRetries number of attempts to login (initially or after a successful login and AutoReLogin set).
    // LoginRetries = -1 indicate the client should try indefinitely.
    // LoginRetryWait is time to wait to re-login on disconnection after a successful login
    // ConnectRetyWait is time to wait to reconnect after a unsucessful connection
    // if AutoReLogin is set, the connection will be re-attempted on disconnection after a successful login.
    // LogInTimeout is the time to wait for a welcome line after logging in.
    Property  Active : Boolean read FActive write SetActive;
    Property  LoginTimeout : Integer read FLoginTimeOut write FLoginTimeOut;
    Property  LoginRetries : Integer read FLoginRetries write FLoginRetries;
    Property  LoginRetryWait : Integer read FLoginRetryWait write FLoginRetryWait;
    Property  ConnectRetryWait : Integer read FConnectRetryWait write FConnectRetryWait;
    Property  AutoReLogin : Boolean read FAutoReLogin write FAutoReLogin;
    Property  LoggingIn : Boolean read FLoggingIn;
    Property  LoginAttemptNr : Integer read FLoginAttemptNr;
    Property  LoggedIn : Boolean read FLoggedIn;
    Property  OnLoggingIn : TIrcClientNotifyEvent read FOnLoggingIn write FOnLoggingIn;
    Property  OnLoggedIn : TIrcClientNotifyEvent read FOnLoggedIn write FOnLoggedIn;
    Property  OnLoggedOut : TIrcClientNotifyEvent read FOnLoggedOut write FOnLoggedOut;
    Property  MessageOfTheDay : String read FMOTD;

    // OnMessage gets called for every message received
    // OnPrivMsg gets called for every Priv message received
    // OnCTCPMsg gets called for every CTCP message received
    // ie OnCTPCPMsg gets preceded by calls to OnPrivMsg and OnMessage for the same message
    Property  OnMessage : TIrcClientMessageEvent read FOnMessage write FOnMessage;
    Property  OnPrivMsg : TIrcClientPrivMsgEvent read FOnPrivMsg write FOnPrivMsg;
    Property  OnNoticeMsg : TIrcClientPrivMsgEvent read FOnNoticeMsg write FOnNoticeMsg;
    Property  OnCTCPMsg : TIrcClientCTCPMessageEvent read FOnCTCPMsg write FOnCTCPMsg;
    Property  OnCTCPReply : TIrcClientCTCPMessageEvent read FOnCTCPReply write FOnCTCPReply;

    // Send Message
    Procedure SendMessage (const Msg : String);
    Procedure PrivMsg (const Destination, Msg : String);
    Procedure Notice (const Destination, Msg : String);
    Procedure CTCPMsg (const Destination, Cmd : String; const Parameters : StringArray);
    Procedure CTCPReply (const Destination, Cmd : String; const Parameters : StringArray);

    // Away. Set AwayMsg = '' to cancel away message.
    Procedure SetAway (const AwayMsg : String);

    // Channel list
    // RefreshChannelList can be called at any time.
    Procedure RefreshChannelList;
    Property  ChannelList : StringArray read GetChannelList;
    Property  OnChannelList : TIrcClientNotifyEvent read FOnChannelList write FOnChannelList;
    Property  AwaitingChannelList : Boolean read FAwaitingChannelList;

    // Channels
    // Join and Part can be called at any time.
    // OnChannelJoin and OnChannelPart signals when a channel is actually joined/parted.
    // Channels are re/joined on login.
    Function  GetChannel (const Channel : String) : TChannel;
    Function  Join (const Channel : String) : TChannel;
    Procedure Part (const Channel : String);
    Procedure PartAll;
    Property  ChannelNames : StringArray read GetJoinedChannels;
    Property  Channels : TChannelArray read FJoinedChannels;
    Property  OnChannelJoin : TIrcClientChannelNotifyEvent read FOnChannelJoin write FOnChannelJoin;
    Property  OnChannelPart : TIrcClientChannelNotifyEvent read FOnChannelPart write FOnChannelPart;
    Property  OnChannelKick : TIrcClientChannelKickEvent read FOnChannelKick write FOnChannelKick;
    Property  OnChannelUserListChange : TIrcClientChannelNotifyEvent read FOnChannelUserListChange write FOnChannelUserListChange;
    Property  OnChannelInvite : TIrcClientChannelInviteEvent read FOnInvite write FOnInvite;
    Property  OnChannelTopic : TIrcClientChannelNotifyEvent read FOnChannelTopic write FOnChannelTopic;

    // Buddies using IsOn packets
    Procedure AddBuddy (const Nick : String);
    Procedure RemoveBuddy (const Nick : String);
    Property  BuddyList : StringArray read FBuddyList write SetBuddyList;
    Property  BuddiesOnline : StringArray read FBuddiesOnline;
    Property  OnBuddiesOnlineChange : TIrcClientNotifyEvent read FOnBuddiesOnlineChange write FOnBuddiesOnlineChange;

    // User information
    // Responses to WhoIs and WhoWas triggers the OnUserInfo event
    Procedure WhoIs (const Nick : String);
    Procedure WhoWas (const Nick : String);
    Property  OnUserInfo : TIrcClientUserInfoEvent read FOnUserInfo write FOnUserInfo;

    // Channel control
    // Can only be used when channel is joined.
    Procedure ChannelMode (const ChannelOrUser, Mode, Parameter : String);
    Procedure SetChannelOperator (const Channel, User : String; const Operator : Boolean);
    Procedure SetChannelPrivate (const Channel : String; const Priv : Boolean);
    Procedure SetChannelInviteOnly (const Channel : String; const InviteOnly : Boolean);
    Procedure SetChannelSecret (const Channel : String; const Secret : Boolean);
    Procedure SetChannelTopicOperatorOnly (const Channel : String; const TopicOperatorOnly : Boolean);
    Procedure SetChannelNoOutsideMessages (const Channel : String; const NoOutsideMessages : Boolean);
    Procedure SetChannelUserLimit (const Channel : String; const UserLimit : Integer);
    Procedure SetChannelBan (const Channel : String; const BanMask : String; const Ban : Boolean);
    Procedure SetChannelModerated (const Channel : String; const Moderated : Boolean);
    Procedure SetChannelUserSpeak (const Channel, User : String; const MaySpeak : Boolean);
    Procedure SetChannelPassword (const Channel, Password : String);
    Procedure SetUserInvisible (const Channel : String; const Invisible : Boolean);
    Procedure SetUserServerNotices (const Channel : String; const ReceiveServerNotices : Boolean);
    Procedure SetUserWallops (const Channel : String; const ReceiveWallops : Boolean);
    Procedure Kick (const Channel, Nick : String);
    Procedure UpdateChannelTopic (const Channel : String);
    Procedure SetChannelTopic (const Channel : String; const Topic : String);
    Procedure InviteUser (const Channel, User : String);
  end;







{$I IrcConsts.inc}

type
  EIrc = class (Exception);



{ Nicks                                                                        }
const
  MaxNickLength    = 9;
  cs_NickNameStart    = cs_Alpha;
  cs_NickName         = cs_AlphaNumeric + ['-', '[', ']', '\', '`', '^', '{', '}', '_'];

Function  RandomNick : String;
Function  CompareNicks (const Nick1, Nick2 : String) : Boolean;
Function  LowerNick (const Nick : String) : String;



{ Channels                                                                     }
const
  MaxChannelLength    = 200;
  cs_ChannelNameStart = ['#', '&'];
  cs_ChannelName      = cs_Everything - [',', ' ', #7, #13, #10, #0];

type
  TIrcChannelType = (ct_Global, ct_Local);

Procedure DecomposeChannelListEntry (const Entry : String; var Channel, Topic : String; var Users : Integer);
Function  ChannelType (const Channel : String) : TIrcChannelType;
Function  IsValidChannelName (const Channel : String) : Boolean;



{ Miscellaneous                                                                }
Function  IRCColourToTColor (const IRCCol : Integer) : TColor;
Function  ExtractNotifyMessage (const NumericCmd : Integer; const Cmd : String; const Parameters : StringArray) : String;
Function  IsCTCPMessage (const Msg : String) : Boolean;
Procedure DecomposeCTCPMessage (const Msg : String; var Command : String; var Params : StringArray);
Function  EncodeCTCPMessage (const Command : String; const Params : StringArray) : String;



implementation

uses
  Math,
  cMaths,
  cDateTime;



{ Nicks                                                                        }
Function RandomNick : String;
  Begin
    Result := LowerCase (RandomPseudoword (5 + RandomUniform (MaxNickLength - 4)));
  End;

Function LowerNick (const Nick : String) : String;
  Begin
    Result := ReplaceChar (['[', ']', '\'], ['{', '}', '|'], LowerCase (Nick));
  End;

Function CompareNicks (const Nick1, Nick2 : String) : Boolean;
  Begin
    Result := LowerNick (Nick1) = LowerNick (Nick2);
  End;


{ Channels                                                                     }
Function ChannelType (const Channel : String) : TIrcChannelType;
var R : Boolean;
  Begin
    R := Length (Channel) > 0;
    if R then
      Case Channel [1] of
        '#' : Result := ct_Global;
        '&' : Result := ct_Local;
        else R := False;
      end;
    if not R then
      raise EIrc.Create ('Invalid channel name.');
  End;

Function IsValidChannelName (const Channel : String) : Boolean;
var L : Integer;
  Begin
    L := Length (Channel);
    Result := (L > 0) and (Channel [1] in cs_ChannelNameStart) and
              Match (cs_ChannelName, Channel, 2, L - 1) and (L <= MaxChannelLength);
  End;



{ Miscellaneous                                                                }
Function ExtractNotifyMessage (const NumericCmd : Integer; const Cmd : String; const Parameters : StringArray) : String;

  Function Par2 : String;
    Begin
      Result := Join (Copy (Parameters, 1, Length (Parameters) - 1), ' ');
    End;

  Begin
    Case NumericCmd of
      -1 : if Cmd = 'NOTICE' then
             Result := Par2 else
             Result := '';
      RPL_X_WELCOME, RPL_X_YOURHOST, RPL_X_CREATED, RPL_X_MYINFO,
      RPL_LUSERCLIENT, RPL_LUSEROP, RPL_LUSERUNKNOWN,
      RPL_LUSERCHANNELS, RPL_LUSERME,
      RPL_MOTDSTART, RPL_MOTD, RPL_ENDOFMOTD :
        Result := Par2;
    end;
  End;

Function IsOnlineIndicator (const MsgNumeric : Integer) : Boolean;
  Begin
    Case MsgNumeric of
      RPL_X_WELCOME, RPL_X_YOURHOST, RPL_X_CREATED, RPL_X_MYINFO,
      RPL_LUSERCLIENT, RPL_LUSEROP, RPL_LUSERUNKNOWN,
      RPL_LUSERCHANNELS, RPL_LUSERME,
      RPL_MOTDSTART, RPL_MOTD, RPL_ENDOFMOTD : Result := True;
      else Result := False;
    end;
  End;


{ IRC Colours                                                                  }
{                                                                              }
{ 0  white                                                                     }
{ 1  black                                                                     }
{ 2  blue     (navy)                                                           }
{ 3  green                                                                     }
{ 4  red                                                                       }
{ 5  brown    (maroon)                                                         }
{ 6  purple                                                                    }
{ 7  orange   (olive)                                                          }
{ 8  yellow                                                                    }
{ 9  lt.green (lime)                                                           }
{ 10 teal     (a kinda green/blue cyan)                                        }
{ 11 lt.cyan  (cyan ?) (aqua)                                                  }
{ 12 lt.blue  (royal)                                                          }
{ 13 pink     (light purple) (fuchsia)                                         }
{ 14 grey                                                                      }
{ 15 lt.grey  (silver)                                                         }
Function IRCColourToTColor (const IRCCol : Integer) : TColor;
const ColorTable : Array [0..15] of TColor = (clWhite, clBlack, clBlue, clGreen,
      clRed, clMaroon, clPurple, clOlive, clYellow, clLime, clTeal,
      $00A0A000, $00A00000, clFuchsia, clGray, clSilver);
  Begin
    if IRCCol > 15 then
      Result := clBlack else
      Result := ColorTable [IRCCol];
  End;



{                                                                              }
{ Messages                                                                     }
{                                                                              }

{                                                                              }
{   <message>  ::= [':' <prefix> <SPACE> ] <command> <params> <crlf>           }
{   <prefix>   ::= <servername> | <nick> [ '!' <user> ] [ '@' <host> ]         }
{   <command>  ::= <letter>+  | <number> <number> <number>                     }
{   <SPACE>    ::= ' '+                                                        }
{   <params>   ::= <SPACE> [ ':' <trailing> | <middle> <params> ]              }
{                                                                              }
{   <middle>   ::= <Any *non-empty* sequence of octets not including SPACE     }
{                  or NUL or CR or LF, the first of which may not be ':'>      }
{   <trailing> ::= <Any, possibly *empty*, sequence of octets not including    }
{                  NUL or CR or LF>                                            }
{                                                                              }
{   <crlf>     ::= CR LF                                                       }
{                                                                              }
{ DecomposeIRCMessage                                                          }
{   Returns Prefix, Command and Parameters.                                    }
{   Command is always returned in upper case.                                  }
{ DecomposeIRCPrefix                                                           }
{   Returns Nick (or server name), User and Host.                              }
{   Host is always returned in lower case.                                     }
{                                                                              }

{ DecomposeIRCMessage                                                          }
Procedure DecomposeIRCMessage (const Msg : String; var Prefix, Command : String;
          var Params : StringArray);
var F, G, L : Integer;
  Begin
    Prefix := '';
    Command := '';
    Params := nil;
    if Msg = '' then
      exit;

    if Msg [1] = ':' then
      begin
        F := PosNext (' ', Msg);
        if F = 0 then
          begin
            Prefix := CopyFrom (Msg, 2);
            exit;
          end;
        Prefix := CopyRange (Msg, 2, F - 1);
      end else
      F := 0;

    F := PosNext (cs_Everything - [' '], Msg, F);
    G := PosNext (' ', Msg, F);
    if G = 0 then
      begin
        Command := UpperCase (CopyRange (Msg, F, Length (Msg)));
        exit;
      end;
    Command := UpperCase (CopyRange (Msg, F, G - 1));

    Repeat
      G := PosNext (cs_Everything - [' '], Msg, G);
      if G = 0 then
        exit;

      if Msg [G] = ':' then
        begin
          Append (Params, CopyFrom (Msg, G + 1));
          exit;
        end;

      F := PosNext (' ', Msg, G);
      if F = 0 then
        begin
          Append (Params, CopyFrom (Msg, G));
          exit;
        end;

      Append (Params, CopyRange (Msg, G, F - 1));
      G := F;
    Until False;
  End;

{ DecomposeIRCPrefix                                                           }
Procedure DecomposeIRCPrefix (const Prefix : String; var Name, User, Host : String);
var I, J : Integer;
  Begin
    I := PosNext (['!', '@'], Prefix);
    if I = 0 then
      begin
        Name := Prefix;
        User := '';
        Host := '';
      end else
      begin
        Name := CopyLeft (Prefix, I - 1);
        if Prefix [I] = '!' then
          begin
            J := PosNext ('@', Prefix, I);
            if J = 0 then
              begin
                User := CopyFrom (Prefix, I + 1);
                Host := '';
              end else
              begin
                User := CopyRange (Prefix, I + 1, J - 1);
                Host := LowerCase (CopyFrom (Prefix, J + 1));
              end;
          end else
          begin
            J := PosNext ('!', Prefix, I);
            if J = 0 then
              begin
                User := '';
                Host := LowerCase (CopyFrom (Prefix, I + 1));
              end else
              begin
                User := CopyFrom (Prefix, J + 1);
                Host := LowerCase (CopyRange (Prefix, I + 1, J - 1));
              end;
          end;
      end;
  End;



{ Decomposes and unquotes an CTCP message                                      }
Function IsCTCPMessage (const Msg : String) : Boolean;
var L : Integer;
  Begin
    L := Length (Msg);
    Result := (L >= 2) and (Msg [1] = #1) and (Msg [L] = #1);
  End;

Procedure DecomposeCTCPMessage (const Msg : String; var Command : String; var Params : StringArray);
var F : Integer;
    S : String;
  Begin
    S := UnescapeText (TrimQuotes (Msg), '\', ['\', 'a', '1', 'n', 'r', '0', '@'], ['\', #1, #1, #10, #13, #0, ' '], False);
    S := UnescapeText (S, #16, [#16, '0', 'n', 'r'], [#16, #0, #10, #13], True);
    F := PosNext (' ', S);
    if F = 0 then
      begin
        Command := S;
        Params := nil;
      end else
      begin
        Command := CopyLeft (S, F - 1);
        Params := Split (CopyRange (S, F + 1, Length (S) - 1), [' ']);
      end;
    Command := UpperCase (Command);
  End;

Function EncodeCTCPMessage (const Command : String; const Params : StringArray) : String;
var I : Integer;
    S : String;
  Begin
    Result := #1 + UpperCase (Command);
    For I := 0 to Length (Params) - 1 do
      begin
        S := EscapeText (Params [I], ['\', #1, ' ', #10, #13, #0], '\', ['\', 'a', '@', 'n', 'r', '0']);
        S := EscapeText (S, [#16], #16, [#16]);
        Result := Result + ' ' + S;
      end;
    Result := Result + #1;
  End;


{ Channel list entry                                                           }
Procedure DecomposeChannelListEntry (const Entry : String; var Channel, Topic : String; var Users : Integer);
var F, G : Integer;
  Begin
    Channel := '';
    Topic := '';
    Users := -1;
    F := PosNext (' ', Entry);
    if F >= 0 then
      begin
        Channel := CopyLeft (Entry, F - 1);
        G := PosNext (' ', Entry, F);
        if G >= 0 then
          begin
            Users := StrToIntDef (CopyRange (Entry, F + 1, G - 1), -1);
            Topic := CopyFrom (Entry, G + 1);
          end;
      end;
  End;



{                                                                              }
{ TIrcClient                                                                   }
{                                                                              }

{ TChannel                                                                     }
Constructor TChannel.Create (const Client : TIrcClient; const Channel : String);
  Begin
    inherited Create;
    FClient := Client;
    FChannel := Channel;
  End;

Function TChannel.UserIndexByNick (const Nick : String) : Integer;
var I : Integer;
  Begin
    For I := 0 to Length (FUsers) - 1 do
      if CompareNicks (Nick, TrimLeft (FUsers [I], ['+', '@'])) then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Procedure TChannel.SetTopic (const Topic : String);
  Begin
    FClient.SetChannelTopic (FChannel, Topic);
  End;

Procedure TChannel.SetKicked (const Reason : String);
  Begin
    FKicked := True;
    FKickReason := Reason;
    FJoined := False;
  End;

Procedure TChannel.SetJoined;
  Begin
    FKicked := False;
    FJoined := True;
  End;

Procedure TChannel.SetNotJoined;
  Begin
    FJoined := False;
  End;

{ TIrcClient                                                                   }
Constructor TIrcClient.Create;
  Begin
    inherited Create;
    FLoginTimeOut := 2 * 60 * 1000;
    FLoginRetries := 1;
    FLoginRetryWait := 15 * 1000;
    FConnectRetryWait := 2 * 60 * 1000;
    FAutoReLogin := True;

    FClient := TWSocket.Create (nil);
    FClient.OnDnsLookupDone := OnClientDnsLookupDone;
    FClient.OnChangeState := OnClientChangeState;
    FClient.OnError := OnClientError;
    FClient.OnDataAvailable := OnClientRead;
  End;

Destructor TIrcClient.Destroy;
  Begin
    FreeAndNil (FLoginTimer);
    FreeAndNil (FBuddyTimer);
    FreeAndNil (FClient);
    FreeObjectArray (FJoinedChannels);
    inherited Destroy;
  End;

Procedure TIrcClient.SetHost (const Host : String);
  Begin
    if FActive then
      raise EIrcClient.Create ('Can''t change host when active.');

    FHost := Host;
  End;

Procedure TIrcClient.SetPort (const Port : Integer);
  Begin
    if FActive then
      raise EIrcClient.Create ('Can''t change host when active.');

    FPort := Port;
  End;

Procedure TIrcClient.SetUsername (const Username : String);
  Begin
    if FActive then
      raise EIrcClient.Create ('Can''t change username when active.');

    FUsername := Username;
  End;

Procedure TIrcClient.SetPassword (const Password : String);
  Begin
    if FActive then
      raise EIrcClient.Create ('Can''t change password when active.');

    FPassword := Password;
  End;

Procedure TIrcClient.SetRealName (const RealName : String);
  Begin
    if FActive then
      raise EIrcClient.Create ('Can''t change name when active.');

    FRealName := RealName;
  End;

Procedure TIrcClient.SetNick (const Nick : String);
  Begin
    SendMessage ('NICK :' + Nick);
    FNick := Nick;
  End;



{ Logging in                                                                   }
Function TIrcClient.IsConnected : Boolean;
  Begin
    Result := Assigned (FClient) and (FClient.State = wsConnected);
  End;

Procedure TIrcClient.OnLoginRetry (Sender : TObject);
  Begin
    FreeAndNil (FLoginTimer);
    if FActive and ((FLoginRetries = -1) or (FLoginAttemptNr < FLoginRetries)) then
      AttemptLogin;
  End;

Procedure TIrcClient.OnReLogin (Sender : TObject);
  Begin
    FreeAndNil (FLoginTimer);
    if FActive then
      Login;
  End;

Procedure TIrcClient.SetLoggedOut;
var ReLogon : Boolean;
    Retry   : Boolean;
  Begin
    ReLogon := IsConnected and FAutoReLogin;
    Retry := FLoggingIn and ((FLoginRetries = -1) or (FLoginAttemptNr < FLoginRetries));

    FLoggedIn := False;
    FLoggingIn := False;
    FreeAndNil (FLoginTimer);
    FreeAndNil (FBuddyTimer);

    if Assigned (FOnLoggedOut) then
      FOnLoggedOut (self);

    if Assigned (FClient) then
      FClient.Close;

    if FActive and (ReLogon or Retry) then
      begin
        FLoginTimer := TTimer.Create (nil);
        FLoginTimer.Interval := Max (Cond (ReLogon, FLoginRetryWait, FConnectRetryWait), 1000);
        if ReLogon then
          FLoginTimer.OnTimer := OnReLogin else
          FLoginTimer.OnTimer := OnLoginRetry;
        if Assigned (FOnStatus) then
          FOnStatus (self, 'Retry pending at ' + FormatDateTime ('hh:nn:ss', AddMilliseconds (Now, FLoginTimer.Interval)) + '.');
        FLoginTimer.Enabled := True;
      end;
  End;

Procedure TIrcClient.OnClientError (Sender: TObject);
  Begin
    SetLoggedOut;
  End;

Procedure TIrcClient.OnClientChangeState (Sender: TObject; OldState, NewState : TSocketState);
var Msg : String;
  Begin
    if Assigned (FOnStatus) then
      begin
        Case NewState of
          wsConnecting :
            With TWSocket (Sender) do
              Msg := 'Connecting to ' + Addr + ':' + Port + '.';
          wsConnected : Msg := 'Connected.';
          else Msg := '';
        end;
        if Msg <> '' then
          FOnStatus (self, Msg);
      end;

    if (OldState = wsConnected) and (NewState <> wsConnected) then
      SetLoggedOut else
    if (OldState <> wsConnected) and (NewState = wsConnected) then
      begin
        FreeAndNil (FLoginTimer);

        // Log in
        if FPassword = '' then
          FPassword := RandomNick;
        SendMessage ('PASS :' + FPassword);

        if FNick = '' then
          FNick := RandomNick;
        SetNick (FNick);

        SendMessage ('USER ' + FUserName + ' ' + LocalHostName + ' ' + FHost + ' :' + FRealName);

        // Login time-out
        FLoginTimer := TTimer.Create (nil);
        FLoginTimer.Interval := FLoginTimeOut;
        FLoginTimer.OnTimer := OnLoginTimeOut;
        FLoginTimer.Enabled := True;

        if Assigned (FOnStatus) then
          FOnStatus (self, 'Logging in.');
      end;
  End;

Procedure TIrcClient.OnLoginTimeOut (Sender : TObject);
  Begin
    LogOut ('Timed-out waiting for a login response.');
  End;

Procedure TIrcClient.SetLoggedIn;
var I : Integer;
  Begin
    FreeAndNil (FLoginTimer);
    FLoggingIn := False;
    FLoggedIn := True;

    if Assigned (FOnLoggedIn) then
      FOnLoggedIn (self);

    // Buddy checker
    FBuddyTimer := TTimer.Create (nil);
    FBuddyTimer.Interval := 2 * 60 * 1000;
    FBuddyTimer.OnTimer := OnBuddyTimer;
    OnBuddyTimer (FBuddyTimer);
    FBuddyTimer.Enabled := True;

    // Channels
    For I := 0 to Length (FJoinedChannels) - 1 do
      Join (FJoinedChannels [I].Channel);

    // Channel list
    if FRefreshChanListPending then
      RefreshChannelList;
  End;

Procedure TIrcClient.OnClientDnsLookupDone (Sender: TObject; Error: Word);
  Begin
    if Error <> 0 then
      begin
        if Assigned (FOnStatus) then
          FOnStatus (self, 'DNS lookup failed.');
        SetLoggedOut;
      end else
      begin
        FClient.Addr := TWSocket (Sender).DnsResult;
        FClient.Port := IntToStr (FPort);
        if Assigned (FOnStatus) then
          FOnStatus (self, 'Resolved host to ' + FClient.Addr + '.');
        FClient.Connect;
      end;
  End;

Procedure TIrcClient.AttemptLogin;
  Begin
    Inc (FLoginAttemptNr);
    FLoggingIn := True;

    if Assigned (FOnLoggingIn) then
      FOnLoggingIn (self);

    FClient.DnsLookup (FHost);
  End;

Procedure TIrcClient.LogIn;
  Begin
    if FLoggingIn or FLoggedIn then
      exit;
    FLoginAttemptNr := 0;
    AttemptLogin;
  End;

Procedure TIrcClient.LogOut (const QuitMessage : String);
  Begin
    SendMessage ('QUIT' + Cond (QuitMessage <> '', ' :' + QuitMessage, ''));
    SetLoggedOut;
  End;

Procedure TIrcClient.SetActive (const Active : Boolean);
  Begin
    if Active = FActive then
      exit;

    FActive := Active;
    if Active then
      LogIn else
      LogOut ('');
  End;



{ Send Message                                                                 }
Procedure TIrcClient.SendMessage (const Msg : String);
  Begin
    if IsConnected then
      FClient.SendStr (Msg + c_CRLF);
  End;

Procedure TIrcClient.PrivMsg (const Destination, Msg : String);
  Begin
    SendMessage ('PRIVMSG ' + Destination + ' :' + Msg);
  End;

Procedure TIrcClient.Notice (const Destination, Msg : String);
  Begin
    SendMessage ('NOTICE ' + Destination + ' :' + Msg);
  End;

Procedure TIrcClient.CTCPMsg (const Destination, Cmd : String; const Parameters : StringArray);
  Begin
    PrivMsg (Destination, EncodeCTCPMessage (Cmd, Parameters));
  End;

Procedure TIrcClient.CTCPReply (const Destination, Cmd : String; const Parameters : StringArray);
  Begin
    Notice (Destination, EncodeCTCPMessage (Cmd, Parameters));
  End;



{ Message processing event                                                     }
Procedure TIrcClient.OnClientRead (Sender: TObject; Error: word);

  Procedure DoSetChanTopic (const Channel, Topic : String);
  var C : TChannel;
    Begin
      C := GetChannel (Channel);
      if Assigned (C) then
        C.Topic := Topic;
      if Assigned (FOnChannelTopic) then
        FOnChannelTopic (self, C);
    End;

var Msg, Pre, Cmd : String;
    Name, User, Host : String;
    CmdN : Integer;
    Par : StringArray;
    F, G : Integer;
    I, J : Integer;
    S : String;
    L : StringArray;
    R : Boolean;
    C : TChannel;
  Begin
    Repeat
      SetLength (S, 8192);
      I := TWSocket (Sender).Receive (@S [1], 8192);
      if I > 0 then
        begin
          SetLength (S, I);
          FInBuf := FInBuf + S;
        end;
    Until I <= 0;

    Repeat
      F := PosNext (ASCII_LF, FInBuf);
      if F > 0 then
        begin
          Msg := CopyLeft (FInBuf, F - 1);
          FInBuf := CopyFrom (FInBuf, F + 1);

          if Msg = '' then
            continue;
          if Msg [Length (Msg)] = #13 then
            SetLength (Msg, Length (Msg) - 1);
          Msg := cStrings.TrimRight (Msg, [' ']);
          if Msg = '' then
            continue;

          DecomposeIRCMessage (Msg, Pre, Cmd, Par);
          CmdN := StrToIntDef (Cmd, -1);
          DecomposeIRCPrefix (Pre, Name, User, Host);

          if FLoggingIn and IsOnlineIndicator (CmdN) then
            begin
              FNick := Par [0];
              SetLoggedIn;
            end;

          if Assigned (FOnMessage) then
            FOnMessage (self, Name, User, Host, CmdN, Cmd, Par);

          Case CmdN of
            -1 :
              if Cmd = 'PING' then
                SendMessage ('PONG' + Cond (Length (Par) > 0, ' :' + Par [0], '')) else
              if Cmd = 'PRIVMSG' then
                begin
                  if Assigned (FOnPrivMsg) and (Length (Par) >= 2) then
                    FOnPrivMsg (self, Par [0], Name, Par [1]);
                  if Assigned (FOnCTCPMsg) and ISCTCPMessage (Par [1]) then
                    begin
                      S := Par [0];
                      DecomposeCTCPMessage (Par [1], Cmd, Par);
                      FOnCTCPMsg (self, S, Name, Cmd, Par);
                    end;
                end else
              if Cmd = 'NOTICE' then
                begin
                  if Assigned (FOnNoticeMsg) and (Length (Par) >= 2) then
                    FOnNoticeMsg (self, Par [0], Name, Par [1]);
                  if Assigned (FOnCTCPReply) and ISCTCPMessage (Par [1]) then
                    begin
                      S := Par [0];
                      DecomposeCTCPMessage (Par [1], Cmd, Par);
                      FOnCTCPReply (self, S, Name, Cmd, Par);
                    end;
                end else
              if Cmd = 'PART' then
                begin
                  I := GetChannelIndex (Par [0]);
                  if I >= 0 then
                    C := FJoinedChannels [I] else
                    C := nil;
                  if CompareNicks (Name, FNick) then
                    begin
                      if Assigned (C) then
                        begin
                          C.SetNotJoined;
                          if Assigned (FOnChannelPart) then
                            FOnChannelPart (self, C);
                        end;
                    end else
                    if Assigned (C) then
                      begin
                        I := C.UserIndexByNick (Name);
                        if I >= 0 then
                          Remove (C.FUsers, I, 1);
                        if Assigned (FOnChannelUserListChange) then
                          FOnChannelUserListChange (self, C);
                      end;
                end else
              if Cmd = 'KICK' then
                begin
                  C := GetChannel (Par [0]);
                  if Assigned (C) then
                    begin
                      I := C.UserIndexByNick (Name);
                      if I >= 0 then
                        Remove (C.FUsers, I, 1);
                      if Assigned (FOnChannelUserListChange) then
                        FOnChannelUserListChange (self, C);
                    end;
                  if CompareNicks (Par [1], FNick) then
                    begin
                      if Assigned (C) then
                        C.SetKicked (Par [2]);
                      if Assigned (FOnChannelKick) then
                        FOnChannelKick (self, C, Name, Par [2]);
                      if Assigned (FOnChannelPart) then
                        FOnChannelPart (self, C);
                    end;
                end else
              if Cmd = 'JOIN' then
                begin
                  C := GetChannel (Par [0]);
                  if CompareNicks (Name, FNick) then
                    begin
                      if not Assigned (C) then
                        begin
                          C := CreateChannel (Par [0]);
                          Append (ObjectArray (FJoinedChannels), C);
                        end;
                      C.SetJoined;
                      if Assigned (FOnChannelJoin) then
                        FOnChannelJoin (self, C);
                    end else
                    if Assigned (C) then
                      begin
                        Append (C.FUsers, Name);
                        if Assigned (FOnChannelUserListChange) then
                          FOnChannelUserListChange (self, C);
                      end;
                end else
              if Cmd = 'QUIT' then
                begin
                  For I := 0 to Length (FJoinedChannels) - 1 do
                    begin
                      J := FJoinedChannels [I].UserIndexByNick (Name);
                      if J >= 0 then
                        begin
                          Remove (FJoinedChannels [I].FUsers, J, 1);
                          if Assigned (FOnChannelUserListChange) then
                            FOnChannelUserListChange (self, FJoinedChannels [I]);
                        end;
                    end;
                end else
              if Cmd = 'ERROR' then
                begin
                end else
              if Cmd = 'MODE' then
                begin
                end else
              if Cmd = 'INVITE' then
                begin
                  if CompareNicks (Par [0], FNick) then
                    if Assigned (FOnInvite) then
                      FOnInvite (self, Par [1], Name);
                end else
              if Cmd = 'TOPIC' then
                DoSetChanTopic (Par [0], Par [1]);
            RPL_NOTOPIC :
              DoSetChanTopic (Par [0], '');
            RPL_TOPIC :
              DoSetChanTopic (Par [0], Par [1]);
            RPL_ISON :
              begin
                For G := 1 to Length (Par) - 1 do
                  if Trim (Par [G]) <> '' then
                    Append (FBuddiesTmp, Split (Par [G], ' '));
                Dec (FIsOnlineCount);
                if FIsOnlineCount = 0 then
                  begin
                    R := not IsEqual (FBuddiesOnline, FBuddiesTmp);
                    FBuddiesOnline := FBuddiesTmp;
                    if R then
                      if Assigned (FOnBuddiesOnlineChange) then
                        FOnBuddiesOnlineChange (self);
                  end;
              end;
            RPL_NAMREPLY :
              begin
                I := Cond (Par [1] = '=', Integer (1), Integer (0));
                C := GetChannel (Par [1 + I]);
                if Assigned (C) then
                  Append (C.FUsers, Split (Trim (Par [2 + I])));
              end;
            RPL_ENDOFNAMES :
              begin
                C := GetChannel (Par [1]);
                if Assigned (C) then
                  begin
                    Sort (C.FUsers);
                    if Assigned (FOnChannelUserListChange) then
                      FOnChannelUserListChange (self, C);
                  end;
              end;
            ERR_NICKNAMEINUSE,
            ERR_NICKCOLLISION :
              begin
                S := RandomNick;
                if Assigned (FOnNickInUse) then
                  FOnNickInUse (self, S);
                SetNick (S);
              end;
            RPL_MOTDSTART :
              begin
                FMOTD := '';
                FGotMOTD := False;
              end;
            RPL_MOTD :
              if Length (Par) > 0 then
                FMOTD := FMOTD + Par [Length (Par) - 1] + #13#10;
            RPL_ENDOFMOTD :
              FGotMOTD := True;
            RPL_LISTSTART :
              SetLength (FNewChanList, 0);
            RPL_LIST :
              Append (FNewChanList, cStrings.Join (Copy (Par, 1, Length (Par) - 1), ' '));
            RPL_LISTEND :
              begin
                Sort (FNewChanList);
                FChannelList := FNewChanList;
                FAwaitingChannelList := False;
                if Assigned (FOnChannelList) then
                  FOnChannelList (self);
              end;
            RPL_WHOISUSER,
            RPL_WHOWASUSER :
              if Assigned (FOnUserInfo) then
                if Length (Par) >= 6 then
                  FOnUserInfo (self, (CmdN = RPL_WHOISUSER), Par [1], Par [2], Par [3], Par [5]);
            // Join errors
            ERR_BANNEDFROMCHAN : ;
            ERR_BADCHANNELKEY : ;
            ERR_NOSUCHCHANNEL : ;   // Leave, Mode
            ERR_INVITEONLYCHAN : ;
            ERR_CHANNELISFULL : ;
            ERR_TOOMANYCHANNELS : ;
            // Leave errors
            ERR_NOTONCHANNEL : ;   // Mode, Invite
            // Mode replies
            RPL_CHANNELMODEIS : ;
            ERR_CHANOPRIVSNEEDED : ; // Invite
            ERR_NOSUCHNICK : ; // Invite
            ERR_KEYSET : ;
            RPL_BANLIST : ;
            RPL_ENDOFBANLIST : ;
            ERR_UNKNOWNMODE : ;
            ERR_USERSDONTMATCH : ;
            RPL_UMODEIS : ;
            ERR_UMODEUNKNOWNFLAG : ;
            // Invite
            ERR_USERONCHANNEL : ;
            RPL_INVITING : ;
            RPL_AWAY : ; // PrivMsg
          end;
        end;
    Until F = 0;
  End;



{ Channel list                                                                 }
Procedure TIrcClient.RefreshChannelList;
  Begin
    if FAwaitingChannelList then
      exit;

    if FLoggedIn then
      begin
        SendMessage ('LIST');
        FRefreshChanListPending := False;
      end else
      FRefreshChanListPending := True;
    FAwaitingChannelList := True;
  End;

Function TIrcClient.GetChannelList : StringArray;
  Begin
    if Length (FChannelList) = 0 then
      Result := FNewChanList else
      Result := FChannelList;
  End;


{ Channels                                                                     }
Function TIrcClient.CreateChannel (const Channel : String) : TChannel;
  Begin
    Result := TChannel.Create (self, Channel);
  End;

Function TIrcClient.GetJoinedChannels : StringArray;
var I : Integer;
  Begin
    SetLength (Result, 0);
    For I := 0 to Length (FJoinedChannels) - 1 do
      Append (Result, FJoinedChannels [I].Channel);
  End;

Function TIrcClient.GetChannelIndex (const Channel : String) : Integer;
var I : Integer;
    S : String;
  Begin
    S := LowerCase (Channel);
    For I := 0 to Length (FJoinedChannels) - 1 do
      if LowerCase (FJoinedChannels [I].Channel) = S then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function TIrcClient.GetChannel (const Channel : String) : TChannel;
var I : Integer;
  Begin
    I := GetChannelIndex (Channel);
    if I >= 0 then
      Result := FJoinedChannels [I] else
      Result := nil;
  End;

Function TIrcClient.Join (const Channel : String) : TChannel;
  Begin
    Result := GetChannel (Channel);
    if not Assigned (Result) then
      begin
        Result := CreateChannel (Channel);
        Append (ObjectArray (FJoinedChannels), Result);
      end else
      SetLength (Result.FUsers, 0);
    if FLoggedIn then
      SendMessage ('JOIN :' + Channel);
  End;

Procedure TIrcClient.Part (const Channel : String);
var I : Integer;
  Begin
    I := GetChannelIndex (Channel);
    if I >= 0 then
      begin
        if FLoggedIn then
          SendMessage ('PART :' + Channel);
        Remove (ObjectArray (FJoinedChannels), I, 1, True);
      end;
  End;

Procedure TIrcClient.PartAll;
var I : Integer;
  Begin
    if FLoggedIn then
      For I := 0 to Length (FJoinedChannels) - 1 do
        SendMessage ('PART :' + FJoinedChannels [I].Channel);
    FreeObjectArray (FJoinedChannels);
  End;

Procedure TIrcClient.WhoIs (const Nick : String);
  Begin
    if FLoggedIn then
      SendMessage ('WHOIS :' + Nick);
  End;

Procedure TIrcClient.WhoWas (const Nick : String);
  Begin
    if FLoggedIn then
      SendMessage ('WHOWAS :' + Nick);
  End;

{ Channel control                                                              }
Procedure TIrcClient.ChannelMode (const ChannelOrUser, Mode, Parameter : String);
  Begin
    if FLoggedIn then
      SendMessage ('MODE ' + ChannelOrUser + ' ' + Mode + Cond (Parameter <> '', ' ' + Parameter, ''));
  End;

Procedure TIrcClient.SetChannelOperator (const Channel, User : String; const Operator : Boolean);
  Begin
    ChannelMode (Channel, Cond (Operator, '+', '-') + 'o', User);
  End;

Procedure TIrcClient.SetChannelPrivate (const Channel : String; const Priv : Boolean);
  Begin
    ChannelMode (Channel, Cond (Priv, '+', '-') + 'p', '');
  End;

Procedure TIrcClient.SetChannelInviteOnly (const Channel : String; const InviteOnly : Boolean);
  Begin
    ChannelMode (Channel, Cond (InviteOnly, '+', '-') + 'i', '');
  End;

Procedure TIrcClient.SetChannelSecret (const Channel : String; const Secret : Boolean);
  Begin
    ChannelMode (Channel, Cond (Secret, '+', '-') + 's', '');
  End;

Procedure TIrcClient.SetChannelTopicOperatorOnly (const Channel : String; const TopicOperatorOnly : Boolean);
  Begin
    ChannelMode (Channel, Cond (TopicOperatorOnly, '+', '-') + 't', '');
  End;

Procedure TIrcClient.SetChannelNoOutsideMessages (const Channel : String; const NoOutsideMessages : Boolean);
  Begin
    ChannelMode (Channel, Cond (NoOutsideMessages, '+', '-') + 'n', '');
  End;

// Set UserLimit = 0 to remove user limit from channel
Procedure TIrcClient.SetChannelUserLimit (const Channel : String; const UserLimit : Integer);
  Begin
    ChannelMode (Channel, Cond (UserLimit > 0, '+', '-') + 'l', Cond (UserLimit > 0, IntToStr (UserLimit), ''));
  End;

Procedure TIrcClient.SetChannelBan (const Channel : String; const BanMask : String; const Ban : Boolean);
  Begin
    ChannelMode (Channel, Cond (Ban, '+', '-') + 'b', BanMask);
  End;

Procedure TIrcClient.SetChannelModerated (const Channel : String; const Moderated : Boolean);
  Begin
    ChannelMode (Channel, Cond (Moderated, '+', '-') + 'm', '');
  End;

Procedure TIrcClient.SetChannelUserSpeak (const Channel, User : String; const MaySpeak : Boolean);
  Begin
    ChannelMode (Channel, Cond (MaySpeak, '+', '-') + 'v', User);
  End;

// Set Password = '' to remove password
Procedure TIrcClient.SetChannelPassword (const Channel, Password : String);
  Begin
    ChannelMode (Channel, Cond (Password <> '', '+', '-') + 'k', Password);
  End;

Procedure TIrcClient.SetUserInvisible (const Channel : String; const Invisible : Boolean);
  Begin
    ChannelMode (FNick, Cond (Invisible, '+', '-') + 'i', '');
  End;

Procedure TIrcClient.SetUserServerNotices (const Channel : String; const ReceiveServerNotices : Boolean);
  Begin
    ChannelMode (FNick, Cond (ReceiveServerNotices, '+', '-') + 's', '');
  End;

Procedure TIrcClient.SetUserWallops (const Channel : String; const ReceiveWallops : Boolean);
  Begin
    ChannelMode (FNick, Cond (ReceiveWallops, '+', '-') + 'w', '');
  End;

Procedure TIrcClient.Kick (const Channel, Nick : String);
  Begin
    if FLoggedIn then
      SendMessage ('KICK ' + Channel + ' :' + Nick);
  End;

Procedure TIrcClient.UpdateChannelTopic (const Channel : String);
  Begin
    if FLoggedIn then
      SendMessage ('TOPIC :' + Channel);
  End;

Procedure TIrcClient.SetChannelTopic (const Channel : String; const Topic : String);
  Begin
    if FLoggedIn then
      SendMessage ('TOPIC ' + Channel + ' :' + Topic);
  End;

Procedure TIrcClient.InviteUser (const Channel, User : String);
  Begin
    if FLoggedIn then
      SendMessage ('INVITE ' + User + ' :' + Channel);
  End;

Procedure TIrcClient.SetAway (const AwayMsg : String);
  Begin
    if FLoggedIn then
      if AwayMsg <> '' then
        SendMessage ('AWAY :' + AwayMsg) else
        SendMessage ('AWAY');
  End;

{ Buddy checker                                                                }
Procedure TIrcClient.CheckBuddyList;

  Procedure SendISON (var S : String);
    Begin
      if S <> '' then
        begin
          Inc (FIsOnlineCount);
          SendMessage ('ISON ' + S);
          S := '';
        end;
    End;

var F, L : Integer;
    S : String;

  Begin
    SetLength (FBuddiesTmp, 0);
    FIsOnlineCount := 0;
    S := '';
    L := Length (FBuddyList);
    For F := 0 to L - 1 do
      begin
        S := S + Cond (S <> '', ' ', '') + FBuddyList [F];
        if (F < L - 1) and (Length (S) + Length (FBuddyList [F + 1]) + 8 > 128) then
          SendISON (S);
      end;
    SendISON (S);
  End;

Procedure TIrcClient.OnBuddyTimer (Sender : TObject);
  Begin
    CheckBuddyList;
  End;

Procedure TIrcClient.SetBuddyList (const BuddyList : StringArray);
  Begin
    FBuddyList := BuddyList;
  End;

Procedure TIrcClient.AddBuddy (const Nick : String);
  Begin
    Append (FBuddyList, Nick);
  End;

Procedure TIrcClient.RemoveBuddy (const Nick : String);
var I : Integer;
  Begin
    I := PosNext (Nick, FBuddyList);
    if I >= 0 then
      Remove (FBuddyList, I, 1);
  End;


end.

