{$I cHeader.inc}
unit cInternetStandards;

interface

uses
  // System units
  SysUtils,
  Controls,      // TControl
  Classes,       // TComponent
  ScktComp,      // TServerSocket

  // L0 units
  cUtils,
  cStrings,
  cStreams,
  cTypes,
  cDataStructs;



{                                                                              }
{ Internet Standards v0.17 (L0)                                                }
{                                                                              }
{   Implementation of various RFC standards                                    }
{                                                                              }
{ This unit is copyrighted  2000-2001 by David Butler (david@e.co.za)         }
{                                                                              }
{ This unit is part of Delphi Fundamentals, its original name is               }
{ cInternetStandards.                                                          }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send modifications, suggestions and bug        }
{ reports to david@e.co.za                                                     }
{                                                                              }
{                                                                              }
{ 01/02/2000  v0.01  Initial html Functions                                    }
{ 20/02/2000  v0.02  DecomposeFromField, DecodeMIME functions,                 }
{                    DecodeQuotedPrintableText, htmlSafeText                   }
{                    Document code                                             }
{ 04/03/2000  v0.03  RFC Date-Time functions                                   }
{ 05/03/2000  v0.04  HTTP response functions                                   }
{ 06/03/2000  v0.05  Message header decoding functions                         }
{ 08/03/2000  v0.06  Nested MIME messages.                                     }
{ 09/04/2000  v0.07  ThtmlParser.                                              }
{ 11/04/2000  v0.08  ThtmlRenderer.                                            }
{ 12/04/2000  v0.09  Added AhtmlRenderer and moved ThtmlRenderer to            }
{                    cHTMLRenderer.                                            }
{                    710 lines interface. 2090 lines implementation.           }
{ 15/04/2000  v0.10  Rewrote ThtmlParser from HTML 4.0 specifications.         }
{                    769 lines interface. 2400 lines implementation.           }
{ 28/04/2000  v0.11  Moved HTTP code to cHTTPServer.                           }
{ 11/05/2000  v0.12  Moved HTML code to cHTML, moved XML code to cXML.         }
{ 21/07/2000  v0.13  Added TVCard.                                             }
{ 03/10/2000  v0.14  Added AInternetServer.                                    }
{ 17/10/2000  v0.15  Moved MIME functions to unit cMIME.                       }
{                    Added AHeader.                                            }
{ 18/10/2000  v0.16  Added AMessage, AMessagePart.                             }
{ 27/11/2000  v0.17  Moved AInternetServer to cInternetServer.                 }
{                                                                              }



{                                                                              }
{ Miscellaneous                                                                }
{                                                                              }
const
  CRLF  = ASCII_CR + ASCII_LF;
  HTAB  = ASCII_HT;
  CTL   = [#0..#31];
  SPACE = [ASCII_HT, ASCII_SP];

Function  HostFromFullAddress (const S : String) : String;
Function  PortFromFullAddress (const S : String; const Default : Integer) : Integer;
Function  IsIPAddress (const S : String) : Boolean;
Function  HeaderField (const FieldName, FieldValue : String; const Parameters : String = '') : String;
Function  FieldValue (const Header : String; const FieldName : String) : String;
Function  MessageIDFieldBody (Id : String = ''; Host : String = '') : String;
Procedure DecomposeEMailField (const F : String; var EMailAddress, Name : String);


{                                                                              }
{ Date/Time Encoding/Decoding                                                  }
{   From RFC822, RFC850, RFC1123, RFC1036, RFC1945 (See implementation for     }
{   details)                                                                   }
{                                                                              }
Function RFCDateTime (const D : TDateTime) : String;
{ Returns a local DateTime (D) as a RFC DateTime.                              }
Function DateField : String;
Function DateFieldBody : String;
{ Returns a standard header field with the current DateTime.                   }
Function RFCTimeZoneToGMTBias (const Zone : String; const Strict : Boolean = False) : Integer;
{ Returns the bias of the specified Zone relative to GMT (UT) in minutes       }
Function GMTDateTimeToRFC1123DateTime (const D : TDateTime; const IncludeDayOfWeek : Boolean = True) : String;
{ Converts a GMT Date-Time to a RFC1123 Date-Time                              }
{ This is the preferred representation on the Internet for all Date-Times.     }
{ See cDateTime.LocalTimeToGMTTime for converting to GMT time.                 }
Function RFCDateTimeToGMTDateTime (const S : String) : TDateTime;
{ Converts a Date-Time that complies with any of the RFC standards to          }
{   a GMT Date-Time. This is a tolerant implementation.                        }
{ See cDateTime.GMTTimeToLocalTime for converting to local time.               }
Function RFCMonthName (const Month : Integer) : String;
{ Returns the RFC month name, given the month number.                          }
Function RFCMonth (const S : String) : Integer;
{ Returns the RFC month number, given the month name.                          }



{                                                                              }
{ Headers, bodies and messages                                                 }
{                                                                              }
{ From RFC822 (Standard for the format of ARPA Internet Text Messages) and     }
{   RFC2616 (Hypertext Transfer Protocol -- HTTP/1.1). See implementation for  }
{   details.                                                                   }
type
  { AMessagePart                                                               }
  AMessagePart = class (AType)
    public
    Procedure Prepare; virtual;
  end;

  { AHeaderField                                                               }
  AHeaderField = class (AMessagePart)
    protected
    Function  GetName : String; virtual; abstract;
    Procedure SetName (const FieldName : String); virtual; abstract;
    Function  GetBody : String; virtual; abstract;
    Procedure SetBody (const FieldBody : String); virtual; abstract;
    Function  GetAsString : String; override;
    Procedure SetAsString (const F : String); override;
    Function  GetBodyAsInteger : Integer; virtual;
    Procedure SetBodyAsInteger (const F : Integer); virtual;

    public
    Constructor CreateEx (const FieldBody : String); overload;
    Constructor CreateEx (const FieldName, FieldBody : String); overload;

    Property  Name : String read GetName write SetName;
    Property  Body : String read GetBody write SetBody;
    Property  BodyAsInteger : Integer read GetBodyAsInteger write SetBodyAsInteger;
  end;
  CHeaderField = class of AHeaderField;

  { THeaderField                                                               }
  THeaderField = class (AHeaderField)
    protected
    FFieldName : String;
    FFieldBody : String;

    Function  GetName : String; override;
    Procedure SetName (const FieldName : String); override;
    Function  GetBody : String; override;
    Procedure SetBody (const FieldBody : String); override;

    public
    Function  Duplicate : AType; override;
  end;

  { TDateField                                                                 }
  TDateField = class (AHeaderField)
    protected
    Function  GetBody : String; override;
    Procedure SetBody (const FieldBody : String); override;
    Function  GetName : String; override;

    public
    GMTDateTime : TDateTime;

    Constructor CreateNow;
    Constructor CreateEx (const LocalTime : TDateTime);
    Function Duplicate : AType; override;
  end;

  { TEMailField                                                                }
  TEMailField = class (THeaderField)
    protected
    Function  GetBody : String; override;
    Procedure SetBody (const FieldBody : String); override;

    public
    Address  : String;
    UserName : String;
  end;

  { AHeader                                                                    }
  AHeader = class (AMessagePart)
    protected
    Function  GetFieldByIndex (const Idx : Integer) : AHeaderField; virtual; abstract;
    Procedure SetFieldByIndex (const Idx : Integer; const Field : AHeaderField); virtual; abstract;
    Function  GetField (const FieldName : String) : AHeaderField; virtual;
    Procedure SetField (const FieldName : String; const Field : AHeaderField); virtual;
    Function  GetFieldBody (const FieldName : String) : String; virtual;
    Procedure SetFieldBody (const FieldName : String; const FieldBody : String); virtual;
    Function  GetAsString : String; override;
    Procedure SetAsString (const Header : String); override;

    public
    Function  Duplicate : AType; override;
    Function  FieldCount : Integer; virtual; abstract;
    Property  FieldByIndex [const Idx : Integer] : AHeaderField read GetFieldByIndex write SetFieldByIndex;
    Procedure DeleteFieldByIndex (const Idx : Integer); virtual; abstract;
    Procedure DeleteField (const FieldName : String); virtual;
    Function  GetFieldIndex (const FieldName : String) : Integer;
    Procedure AddField (const Field : AHeaderField); overload; virtual; abstract;
    Procedure AddField (const FieldName, FieldBody : String); overload; virtual; abstract;
    Procedure AddField (const FieldLine : String); overload;
    Property  Field [const FieldName : String] : AHeaderField read GetField write SetField; default;
    Function  GetFieldNames : StringArray; virtual;
    Property  FieldBody [const FieldName : String] : String read GetFieldBody write SetFieldBody;
    Function  HasField (const FieldName : String) : Boolean; virtual;
    Procedure Prepare; override;
  end;
  AHeaderArray = Array of AHeader;

  { THeader                                                                    }
  THeader = class (AHeader)
    protected
    HeaderFields : Array of AHeaderField;

    Function  GetFieldByIndex (const Idx : Integer) : AHeaderField; override;
    Procedure SetFieldByIndex (const Idx : Integer; const Field : AHeaderField); override;

    public
    class Function CreateInstance : AType; override;
    Destructor Destroy; override;

    Procedure DeleteFieldByIndex (const Idx : Integer); override;
    Procedure AddField (const FieldName, FieldBody : String); override;
    Procedure AddField (const Field : AHeaderField); override;
    Function  FieldCount : Integer; override;
  end;

  { TEmailMessageHeader                                                        }
  TEmailMessageHeader = class (THeader)
    public
    Constructor CreateEx (const FromField, ToField, SubjectField : String;
                          const ReplyToField : String = '';
                          const DateField : TDateTime = 0.0);
    Procedure AddField (const FieldName, FieldBody : String); override;
    Procedure Prepare; override;
  end;

  { AMessageBodyPart                                                           }
  AMessageBodyPart = class (AMessagePart);

  { AMessage                                                                   }
  AMessage = class (AMessageBodyPart)
    protected
    Function  GetAsString : String; override;
    Function  GetHeader : AHeader; virtual; abstract;
    Procedure SetHeader (const Header : AHeader); virtual; abstract;
    Function  GetBodyPart (const Idx : Integer) : AMessagePart; virtual; abstract;
    Procedure SetBodyPart (const Idx : Integer; const BodyPart : AMessagePart); virtual; abstract;
    Function  GetBodyAsString : String; virtual;
    Procedure SetBodyAsString (const BodyAsString : String); virtual; abstract;

    public
    Procedure Prepare; override;
    Property  Header : AHeader read GetHeader write SetHeader;
    Function  AddBodyPart (const BodyPart : AMessagePart) : Integer; virtual; abstract;
    Function  BodyPartCount : Integer; virtual; abstract;
    Property  BodyPart [const Idx : Integer] : AMessagePart read GetBodyPart write SetBodyPart; default;
    Property  BodyAsString : String read GetBodyAsString write SetBodyAsString;
  end;

  { ATextBodyPart                                                              }
  ATextBodyPart = class (AMessageBodyPart);

  { TTextBodyPart                                                              }
  TTextBodyPart = class (ATextBodyPart)
    protected
    FBody : String;

    Function  GetAsString : String; override;
    Procedure SetAsString (const S : String); override;

    public
    Constructor CreateEx (const Body : String);
  end;

  { ABinaryBodyPart                                                            }
  ABinaryBodyPart = class (AMessageBodyPart);

  { TMessage                                                                   }
  TMessage = class (AMessage)
    protected
    FHeader : AHeader;
    FBody   : Array of AMessagePart;

    Function  GetHeader : AHeader; override;
    Procedure SetHeader (const Header : AHeader); override;
    Function  GetBodyPart (const Idx : Integer) : AMessagePart; override;
    Procedure SetBodyPart (const Idx : Integer; const BodyPart : AMessagePart); override;

    public
    Constructor CreateEx (const Header : THeader; const BodyParts : array of AMessagePart);
    Destructor Destroy; override;
    Function  AddBodyPart (const BodyPart : AMessagePart) : Integer; override;
    Function  BodyPartCount : Integer; override;
  end;

  { TEmailMessage                                                              }
  TEmailMessage = class (TMessage)
    public
    Constructor CreateEx (const Header : THeader; const Body : AMessagePart); overload;
    Constructor CreateEx (const Header : THeader; const Body : String); overload;
    Constructor CreateEx (const HdrFrom, HdrTo, HdrSubject : String;
                          const HdrReplyTo : String = ''; const HdrDate : TDateTime = 0.0;
                          const Body : String = ''); overload;
  end;




{                                                                              }
{ HTML Graphics Rendering                                                      }
{                                                                              }
{ AhtmlRenderer                                                                }
{   Base class for a visual HTML renderer.                                     }
{   HorSpace returns the numbers of pixels left horisontally before the        }
{     Renderer will attempt a Break.                                           }
{   BOL returns True if renderer is at the beginning of a line.                }
{   AddControl gives a visual control to a renderer to be added to the         }
{     display.                                                                 }
{   BreakLeft moves to beginning of the next rendering "line" (newline). An    }
{     additional SkipHeight pixel lines are skipped.                           }
{                                                                              }
type
  ThtmlRenderAttributes = Set of (raBold, raItalic, raUnderline, raSubscript,
                                  raSuperScript, raStrikethrough,
                                  raStrong, raEmphasis, raCode, raKeyboard,
                                  raVariable, raTeletype, raCitation,
                                  raDefinition, raSample, raPreformatted,
                                  raCentered, raSmall, raBig, raIsLink);
type
  ThtmlFontSize = Integer;
  ThtmlFontColour = String;
  ThtmlRendererBreakType = (btMargin,
                            btNextMargin,
                            btLeftMargin);
  ThtmlHorAlign = (haLeft, haRight, haCenter);
  ThtmlVerAlign = (vaBottom, vaTop, vaCenter);
  AhtmlRenderer = class
    protected
    FFontSize       : ThtmlFontSize;
    FFontColour     : ThtmlFontColour;
    FFontBackColour : ThtmlFontColour;
    FAttributes     : ThtmlRenderAttributes;

    FIndent         : Integer;
    FDesiredWidth   : Integer;
    FVAlign         : ThtmlVerAlign;
    FHAlign         : ThtmlHorAlign;

    public
    Constructor Create (const Parent : AhtmlRenderer);

    { HTML attributes }
    Property  FontSize : ThtmlFontSize read FFontSize write FFontSize;
    Property  FontColour : ThtmlFontColour read FFontColour write FFontColour;
    Property  FontBackColour : ThtmlFontColour read FFontBackColour write FFontBackColour;
    Property  Attributes : ThtmlRenderAttributes read FAttributes write FAttributes;

    { Rendering attributes }
    Property  DesiredWidth : Integer read FDesiredWidth write FDesiredWidth;
    Property  Indent : Integer read FIndent write FIndent;
    Function  HorSpace : Integer; virtual; abstract;
    Function  BOL : Boolean; virtual; abstract;
    Function  LeftMargin : Boolean; virtual; abstract;
    Property  VerAlign : ThtmlVerAlign read FVAlign write FVAlign;
    Property  HorAlign : ThtmlHorAlign read FHAlign write FHAlign;

    { Rendering methods }
    Procedure BreakLine (const BreakType : ThtmlRendererBreakType = btMargin; const SkipHeight : Integer = 0); virtual; abstract;
    Procedure AddControl (const Control : TControl; const SetMargin : Boolean = False); virtual; abstract;
  end;





{                                                                              }
{ VCard                                                                        }
{                                                                              }
type
  TVCardName = class
    FamilyName  : String;
    GivenName   : String;
    MiddleNames : String;
    HonPrefix   : String;
    HonPostfix  : String;
  end;
  TVCard = class
    protected
    FFullName  : String;
    FNickNames : TStringArray;
    FEMail     : String;
    FUID       : String;
    FRevision  : TDateTime;
    FSource    : String;
    FNames     : Array of TVCardName;
    FBirthday  : TDateTime;
    FURL       : String;
    FIRCURL    : String;

    Procedure SetAsString (const S : String);
    Function  GetAsString : String;

    public
    Constructor Create;
    Destructor Destroy; override;
    Procedure Clear;
    Procedure GenerateUID;
    Procedure AddName (const FamilyName, GivenName, MiddleNames, HonPrefix, HonPostfix : String);

    Property AsString : String read GetAsString write SetAsString;
    Property FullName : String read FFullName write FFullName;
    Property NickNames : TStringArray read FNickNames;
    Property EMail : String read FEMail write FEMail;
    Property UID : String read FUID write FUID;
    Property Revision : TDateTime read FRevision write FRevision;
    Property Source : String read FSource write FSource;
    Property Birthday : TDateTime read FBirthday write FBirthday;
    Property URL : String read FURL write FURL;
    Property IrcURL : String read FIRCURL write FIRCURL;
  end;



{                                                                              }
{                                                                              }
{                                                                              }
{type
  TOnInternetClientLog = TOnInternetServerLog; }





implementation



uses
  // System units
  Windows,       // GetTickCount

  // L0 units
  cDateTime,     // LocalTimeToGMTTime
  cMaths;        // HexToInt



{                                                                              }
{ Miscellaneous                                                                }
{                                                                              }
Function HostFromFullAddress (const S : String) : String;
  Begin
    Result := CopyBefore (S, ':', True);
  End;

Function PortFromFullAddress (const S : String; const Default : Integer) : Integer;
  Begin
    Result := StrToIntDef (CopyAfter (S, ':', False), Default);
  End;

Function IsIPAddress (const S : String) : Boolean;
  Begin
    try
      Result := Length (StrToByteArray (Remove (cs_WhiteSpace, S), '.')) = 4;
    except
      Result := False;
    end;
  End;

Function HeaderField (const FieldName, FieldValue : String; const Parameters : String = '') : String;
  Begin
    Result := FieldName + ': ' + FieldValue +
              Cond (Parameters <> '', '; ' + Parameters, '') +
              CRLF;
  End;

Function FieldValue (const Header : String; const FieldName : String) : String;
var I, J : Integer;
    S, F : String;
  Begin
    S := LowerCase (Header);
    F := LowerCase (FieldName) + ':';
    I := PosNext (F, S);
    if I = 0 then
      Result := '' else
      begin
        J := PosNext (c_CRLF, S, I);
        if J = 0 then
          J := Length (S) + 1;
        Result := Trim (CopyRange (Header, I + Length (F), J - 1));
      end;
  End;

Function RFCDateTime (const D : TDateTime) : String;
  Begin
    Result := GMTDateTimeToRFC1123DateTime (LocalTimeToGMTTime (D));
  End;

Function DateFieldBody : String;
  Begin
    Result := RFCDateTime (Now);
  End;

Function DateField : String;
  Begin
    Result := HeaderField ('Date', DateFieldBody);
  End;

Function MessageIDFieldBody (Id : String = ''; Host : String = '') : String;
  Begin
    if Id = '' then
      Id := RandomHex (24) + IntToHex (GetTickCount);
    if Host = '' then
      Host := RandomPseudoWord (12);
    Result := '<' + LowerCase (Id) + '@' + LowerCase (Host) + '>';
  End;



{                                                                              }
{ Date/Time Encoding/Decoding                                                  }
{                                                                              }
{ From RFC 822 (Standard for the format of ARPA INTERNET Text Messages):       }
{    "time        =  hour zone                      ; ANSI and Military        }
{     hour        =  2DIGIT ":" 2DIGIT [":" 2DIGIT] ; 00:00:00 - 23:59:59      }
{     zone        =  "UT"  / "GMT"                  ; Universal Time           }
{                                                   ; North American : UT      }
{                 /  "EST" / "EDT"                  ;  Eastern:  - 5/ - 4      }
{                 /  "CST" / "CDT"                  ;  Central:  - 6/ - 5      }
{                 /  "MST" / "MDT"                  ;  Mountain: - 7/ - 6      }
{                 /  "PST" / "PDT"                  ;  Pacific:  - 8/ - 7      }
{                 /  1ALPHA                         ; Military: Z = UT;        }
{                                                   ;  A:-1; (J not used)      }
{                                                   ;  M:-12; N:+1; Y:+12      }
{                 / ( ("+" / "-") 4DIGIT )          ; Local differential       }
{                                                   ;  hours+min. (HHMM)       }
{     date-time   =  [ day "," ] date time          ; dd mm yy                 }
{                                                   ;  hh:mm:ss zzz            }
{     day         =  "Mon"  / "Tue" /  "Wed"  / "Thu"                          }
{                 /  "Fri"  / "Sat" /  "Sun"                                   }
{     date        =  1*2DIGIT month 2DIGIT        ; day month year             }
{                                                 ;  e.g. 20 Jun 82            }
{     month       =  "Jan"  /  "Feb" /  "Mar"  /  "Apr"                        }
{                 /  "May"  /  "Jun" /  "Jul"  /  "Aug"                        }
{                 /  "Sep"  /  "Oct" /  "Nov"  /  "Dec"                    "   }
{                                                                              }
{ Note that even though RFC 822 states hour=2DIGIT":"2DIGIT, none of the       }
{   examples given in the appendix include the ":",                            }
{   for example: "26 Aug 76 1429 EDT"                                          }
{                                                                              }
{                                                                              }
{ From RFC 1036 (Standard for Interchange of USENET Messages):                 }
{                                                                              }
{   "Its format must be acceptable both in RFC-822 and to the getdate(3)       }
{    routine that is provided with the Usenet software.   ...                  }
{    One format that is acceptable to both is:                                 }
{                                                                              }
{                      Wdy, DD Mon YY HH:MM:SS TIMEZONE                        }
{                                                                              }
{    Note in particular that ctime(3) format:                                  }
{                                                                              }
{                          Wdy Mon DD HH:MM:SS YYYY                            }
{                                                                              }
{    is not acceptable because it is not a valid RFC-822 date.  However,       }
{    since older software still generates this format, news                    }
{    implementations are encouraged to accept this format and translate        }
{    it into an acceptable format.                                         "   }
{                                                                              }
{   "Here is an example of a message in the old format (before the             }
{    existence of this standard). It is recommended that                       }
{    implementations also accept messages in this format to ease upward        }
{    conversion.                                                               }
{                                                                              }
{               Posted: Fri Nov 19 16:14:55 1982                           "   }
{                                                                              }
{                                                                              }
{ From RFC 1945 (Hypertext Transfer Protocol -- HTTP/1.0)                      }
{                                                                              }
{  "HTTP/1.0 applications have historically allowed three different            }
{   formats for the representation of date/time stamps:                        }
{                                                                              }
{       Sun, 06 Nov 1994 08:49:37 GMT    ; RFC 822, updated by RFC 1123        }
{       Sunday, 06-Nov-94 08:49:37 GMT   ; RFC 850, obsoleted by RFC 1036      }
{       Sun Nov  6 08:49:37 1994         ; ANSI C's asctime() format           }
{                                                                              }
{   The first format is preferred as an Internet standard and represents       }
{   a fixed-length subset of that defined by RFC 1123 [6] (an update to        }
{   RFC 822 [7]). The second format is in common use, but is based on the      }
{   obsolete RFC 850 [10] date format and lacks a four-digit year.             }
{   HTTP/1.0 clients and servers that parse the date value should accept       }
{   all three formats, though they must never generate the third               }
{   (asctime) format.                                                          }
{                                                                              }
{      Note: Recipients of date values are encouraged to be robust in          }
{      accepting date values that may have been generated by non-HTTP          }
{      applications, as is sometimes the case when retrieving or posting       }
{      messages via proxies/gateways to SMTP or NNTP.                       "  }
{                                                                              }
{  "All HTTP/1.0 date/time stamps must be represented in Universal Time        }
{   (UT), also known as Greenwich Mean Time (GMT), without exception.          }
{                                                                              }
{       HTTP-date      = rfc1123-date | rfc850-date | asctime-date             }
{                                                                              }
{       rfc1123-date   = wkday "," SP date1 SP time SP "GMT"                   }
{       rfc850-date    = weekday "," SP date2 SP time SP "GMT"                 }
{       asctime-date   = wkday SP date3 SP time SP 4DIGIT                      }
{                                                                              }
{       date1          = 2DIGIT SP month SP 4DIGIT                             }
{                        ; day month year (e.g., 02 Jun 1982)                  }
{       date2          = 2DIGIT "-" month "-" 2DIGIT                           }
{                        ; day-month-year (e.g., 02-Jun-82)                    }
{       date3          = month SP ( 2DIGIT | ( SP 1DIGIT ))                    }
{                        ; month day (e.g., Jun  2)                            }
{                                                                              }
{       time           = 2DIGIT ":" 2DIGIT ":" 2DIGIT                          }
{                        ; 00:00:00 - 23:59:59                                 }
{                                                                              }
{       wkday          = "Mon" | "Tue" | "Wed"                                 }
{                      | "Thu" | "Fri" | "Sat" | "Sun"                         }
{                                                                              }
{       weekday        = "Monday" | "Tuesday" | "Wednesday"                    }
{                      | "Thursday" | "Friday" | "Saturday" | "Sunday"         }
{                                                                              }
{       month          = "Jan" | "Feb" | "Mar" | "Apr"                         }
{                      | "May" | "Jun" | "Jul" | "Aug"                         }
{                      | "Sep" | "Oct" | "Nov" | "Dec"                      "  }
{                                                                              }
Function RFCTimeZoneToGMTBias (const Zone : String; const Strict : Boolean = False) : Integer;
type TZoneBias = record
       Zone : String;
       Bias : Integer;
     end;
const TimeZones = 35;
      ZoneBias : Array [1..TimeZones] of TZoneBias =
          ((Zone:'GMT'; Bias:0),                 (Zone:'UT';  Bias:0),
           (Zone:'EST'; Bias:-5*60),             (Zone:'EDT'; Bias:-4*60),
           (Zone:'CST'; Bias:-6*60),             (Zone:'CDT'; Bias:-5*60),
           (Zone:'MST'; Bias:-7*60),             (Zone:'MDT'; Bias:-6*60),
           (Zone:'PST'; Bias:-8*60),             (Zone:'PDT'; Bias:-7*60),
           (Zone:'Z';   Bias:0),                 (Zone:'A';   Bias:-1*60),
           (Zone:'B';   Bias:-2*60),             (Zone:'C';   Bias:-3*60),
           (Zone:'D';   Bias:-4*60),             (Zone:'E';   Bias:-5*60),
           (Zone:'F';   Bias:-6*60),             (Zone:'G';   Bias:-7*60),
           (Zone:'H';   Bias:-8*60),             (Zone:'I';   Bias:-9*60),
           (Zone:'K';   Bias:-10*60),            (Zone:'L';   Bias:-11*60),
           (Zone:'M';   Bias:-12*60),            (Zone:'N';   Bias:1*60),
           (Zone:'O';   Bias:2*60),              (Zone:'P';   Bias:3*60),
           (Zone:'Q';   Bias:4*60),              (Zone:'R';   Bias:3*60),
           (Zone:'S';   Bias:6*60),              (Zone:'T';   Bias:3*60),
           (Zone:'U';   Bias:8*60),              (Zone:'V';   Bias:3*60),
           (Zone:'W';   Bias:10*60),             (Zone:'X';   Bias:3*60),
           (Zone:'Y';   Bias:12*60));
var S : String;
    I : Integer;
  Begin
    if Zone [1] in ['+', '-'] then // +hhmm format
      begin
        if Strict and (Length (Zone) <> 5) then
          raise EConvertError.Create ('Not a RFC1123 time zone: Invalid differential');
        S := Trim (Zone, SPACE);
        try
          I := StrToInt (CopyLeft (S, 3));
        except
          raise EConvertError.Create ('Not a RFC1123 time zone: Invalid hour in differential');
        end;
        if Abs (I) > 12 then
          raise EConvertError.Create ('Not a RFC1123 time zone: Invalid hour in differential');
        Result := I * 60;
        S := CopyFrom (S, 4);
        if S <> '' then
          if Length (S) <> 2 then
            raise EConvertError.Create ('Not a RFC1123 time zone: Invalid minute in differential') else
            try
              I := StrToInt (S);
              Result := Result + I;
            except
              raise EConvertError.Create ('Not a RFC1123 time zone: Invalid minute in differential');
            end;
      end else
      begin // named format
        S := UpperCase (Zone);
        For I := 1 to TimeZones do
          if ZoneBias [I].Zone = S then
            begin
              Result := ZoneBias [I].Bias;
              exit;
            end;
        raise EConvertError.Create ('Not a RFC1123 time zone: Unknown zone name');
      end;
  End;

{ Converts a RFC Time to GMT Time                                              }
Function RFCTimeToGMTTime (const S : String; const StrictlyRFC1123 : Boolean = False) : TDateTime;
var I : Integer;
    T : String;
    HH, MM, SS : Integer;
    U : StringArray;
  Begin
    T := Trim (S, SPACE);

    // Get Zone bias
    I := PosPrev (SPACE, T);
    if I > 0 then
      begin
        Result := Int (RFCTimeZoneToGMTBias (CopyFrom (T, I + 1), StrictlyRFC1123)) / 1440.0;
        T := Trim (CopyLeft (T, I - 1), SPACE);
      end else
      if StrictlyRFC1123 then
        raise EConvertError.Create ('Not a RFC1123 time: No Zone') else
        Result := 0;

    // Get time
    U := Split (T, ':');
    if not StrictlyRFC1123 and (Length (U) = 1) and (Length (U [0]) = 4) then
      try // old hhmm format
        HH := StrToInt (Copy (U [0], 1, 2));
        MM := StrToInt (Copy (U [0], 3, 2));
        SS := 0;
      except
        raise EConvertError.Create ('Not a RFC time: Invalid time');
      end else
      begin // hh:mm[:ss] format (RFC1123)
        if (Length (U) < 2) or (Length (U) > 3) then
          raise EConvertError.Create ('Not a RFC1123 time: Invalid time');
        try
          HH := StrToInt (Trim (U [0], SPACE));
          MM := StrToInt (Trim (U [1], SPACE));
          if Length (U) = 3 then
            SS := StrToInt (Trim (U [2], SPACE)) else
            SS := 0;
        except
          raise EConvertError.Create ('Not a RFC1123 time: Invalid time');
        end;
      end;

    if (HH < 0) or (HH > 23) then
      raise EConvertError.Create ('Not a RFC1123 time: Invalid hour');
    if (MM < 0) or (MM > 59) then
      raise EConvertError.Create ('Not a RFC1123 time: Invalid minute');
    if (SS < 0) or (SS > 59) then
      raise EConvertError.Create ('Not a RFC1123 time: Invalid second');
    Result := Result + EncodeTime (HH, MM, SS, 0);
  End;

{ Converts GMT Time to a RFC1123 Time                                          }
Function GMTTimeToRFC1123Time (const D : TDateTime; const IncludeSeconds : Boolean = True) : String;
var Ho, Mi, Se, Ms : Word;
  Begin
    DecodeTime (D, Ho, Mi, Se, Ms);
    Result := PadLeft (IntToStr (Ho), '0', 2) + ':' +
              PadLeft (IntToStr (Mi), '0', 2) +
              Cond (IncludeSeconds, ':' + PadLeft (IntToStr (Se), '0', 2), '') +
              ' GMT';
  End;


{ Date constants                                                               }
const
  RFC850DayNames : Array [1..7] of String = (
      'Sunday', 'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday');
  RFC1123DayNames : Array [1..7] of String = (
      'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat');
  RFCMonthNames : Array [1..12] of String = (
      'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');

Function RFC850DayOfWeek (const S : String) : Integer;
var I : Integer;
    T : String;
  Begin
    T := UpperCase (S);
    For I := 1 to 7 do
      if UpperCase (RFC850DayNames [I]) = T then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function RFC1123DayOfWeek (const S : String) : Integer;
var I : Integer;
    T : String;
  Begin
    T := UpperCase (S);
    For I := 1 to 7 do
      if UpperCase (RFC1123DayNames [I]) = T then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function RFCMonthName (const Month : Integer) : String;
  Begin
    Assert ((Month >= 1) and (Month <= 12), 'Invalid month');
    Result := RFCMonthNames [Month];
  End;

Function RFCMonth (const S : String) : Integer;
var I : Integer;
    T : String;
  Begin
    T := UpperCase (S);
    For I := 1 to 12 do
      if UpperCase (RFCMonthNames [I]) = T then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function GMTDateTimeToRFC1123DateTime (const D : TDateTime; const IncludeDayOfWeek : Boolean = True) : String;
var Ye, Mo, Da : Word;
  Begin
    DecodeDate (D, Ye, Mo, Da);
    Result := Cond (IncludeDayOfWeek, RFC1123DayNames [DayOfWeek (D)] + ', ', '') +
              PadLeft (IntToStr (Da), '0', 2) + ' ' +
              RFCMonthNames [Mo] + ' ' +
              IntToStr (Ye) + ' ' +
              GMTTimeToRFC1123Time (D);
  End;

Function RFCDateTimeToGMTDateTime (const S : String) : TDateTime;
var T, U : String;
    I : Integer;
    D, M, Y, DOW : Integer;
    V, W : StringArray;
  Begin
    T := Trim (S, SPACE);

    // Extract Day of week
    I := PosNext (SPACE + [','], T);
    if I = 0 then
      DOW := -1 else
      begin
        U := CopyLeft (T, I - 1);
        DOW := RFC850DayOfWeek (U);
        if DOW = -1 then
          DOW := RFC1123DayOfWeek (U);
        if DOW <> -1 then
          T := Trim (CopyFrom (S, I + 1), SPACE);
      end;

    V := Split (T, SPACE);
    if Length (V) < 3 then
      raise EConvertError.Create ('Not a recognised date: Too few elements');

    if PosNext ('-', V [0]) > 0 then // RFC850 date, eg "Sunday, 06-Nov-94 08:49:37 GMT"
      begin
        W := Split (V [0], '-');
        if Length (W) <> 3 then
          raise EConvertError.Create ('Not a recognised date: Not a RFC850 date');
        M := RFCMonth (W [1]);
        if M = -1 then
          raise EConvertError.Create ('Not a recognised date: Not a RFC850 date');
        try
          D := StrToInt (W [0]);
          Y := StrToInt (W [2]);
          if Y < 100 then
            Y := TwoDigitYearToYear (Y);
        except
          raise EConvertError.Create ('Not a recognised date: Not a RFC850 date');
        end;
        Result := EncodeDate (Y, M, D) + RFCTimeToGMTTime (V [1]  + V [2]);
        exit;
      end;

    M := RFCMonth (V [1]);
    if M >= 1 then // RFC822 date, eg Sun, 06 Nov 1994 08:49:37 GMT
      begin
        try
          D := StrToInt (V [0]);
          Y := StrToInt (V [2]);
        except
          raise EConvertError.Create ('Not a recognised date: Not a RFC822 date');
        end;
        Result := EncodeDate (Y, M, D);
        if Length (V) = 4 then
          Result := Result + RFCTimeToGMTTime (V [3]) else
          if Length (V) >= 5 then
            Result := Result + RFCTimeToGMTTime (V [3] + ' ' + V [4]);
        exit;
      end;

    M := RFCMonth (V [0]);
    if M >= 1 then // ANSI C asctime() format, eg "Sun Nov  6 08:49:37 1994"
      begin
        try
          D := StrToInt (V [1]);
          Y := StrToInt (V [3]);
        except
          raise EConvertError.Create ('Not a recognised date: Not a asctime() date');
        end;
        Result := EncodeDate (Y, M, D) + RFCTimeToGMTTime (V [2]);
      end else
      raise EConvertError.Create ('Not a recognised date.');
  End;



{ Allowed Usenet News formats (RFC850):                                        }
{     mark@cbosgd.UUCP                                                         }
{     mark@cbosgd.UUCP (Mark Horton)                                           }
{     Mark Horton <mark@cbosgd.UUCP>                                           }
{ Additional formats supported:                                                }
{     <david@e.co.za> David Butler                                             }
{     "David Butler" david@e.co.za                                             }
{     david@e.co.za "David Butler"                                             }
Procedure DecomposeEMailField (const F : String; var EMailAddress, Name : String);
  Begin
    Name := CopyBetween (F, '"', '"');
    if Name = '' then
      begin
        Name := CopyBetween (F, '(', ')');
        if Name = '' then
          begin
            Name := Trim (CopyBefore (F, '<', False));
            if Name = '' then
              Name := Trim (CopyAfter (F, '>'));
          end;
      end;

    EmailAddress := CopyBetween (F, '<', '>');
    if EmailAddress = '' then
      begin
        EmailAddress := Trim (CopyBefore (F, [',', ':', ';', ' ', '"', '('], False));
        if EmailAddress = '' then
          begin
            EmailAddress := Trim (CopyAfter (F, '"'));
            if EmailAddress = '' then
              EmailAddress := F;
          end;
      end;
  End;

{                                                                              }
{ Header Decoding/Encoding                                                     }
{                                                                              }
{ From RFC822 (Standard for the format of ARPA Internet Text Messages):        }
{     field       =  field-name ":" [ field-body ] CRLF                        }
{     field-name  =  1*<any CHAR, excluding CTLs, SPACE, and ":">              }
{     field-body  =  field-body-contents                                       }
{                    [CRLF LWSP-char field-body]                               }
{     LWSP-char   =  SPACE / HTAB                 ; semantics = SPACE          }
{                                                                              }
{ From RFC2616 (Hypertext Transfer Protocol -- HTTP/1.1):                      }
{   HTTP/1.1 header field values can be folded onto multiple lines if the      }
{   continuation line begins with a space or horizontal tab. All linear        }
{   white space, including folding, has the same semantics as SP. A            }
{   recipient MAY replace any linear white space with a single SP before       }
{   interpreting the field value or forwarding the message downstream.         }
{       LWS            = [CRLF] 1*( SP | HT )                                  }

{ AHeaderField                                                                 }
Constructor AHeaderField.CreateEx (const FieldBody : String);
  Begin
    inherited Create;
    SetBody (FieldBody);
  End;

Constructor AHeaderField.CreateEx (const FieldName, FieldBody : String);
  Begin
    inherited Create;
    SetName (FieldName);
    SetBody (FieldBody);
  End;

Function AHeaderField.GetAsString : String;
var N, B : String;
  Begin
    N := Name;
    B := Body;
    if (N = '') and (B = '') then
      Result := '' else
      Result := N + ': ' + B + CRLF;
  End;

Procedure AHeaderField.SetAsString (const F : String);
var I : Integer;
    S : String;
  Begin
    S := Trim (F, SPACE);
    I := PosNext (':', F);
    if I <= 1 then
      raise EConvertError.Create ('Not a valid header field');
    Name := CopyLeft (S, I - 1);
    Body := Trim (CopyFrom (S, I + 1), SPACE);
  End;

Function AHeaderField.GetBodyAsInteger : Integer;
  Begin
    Result := StrToIntDef (GetBody, -1);
  End;

Procedure AHeaderField.SetBodyAsInteger (const F : Integer);
  Begin
    SetBody (IntToStr (F));
  End;

{ THeaderField                                                                 }
Function THeaderField.GetName : String;
  Begin
    Result := FFieldName;
  End;

Procedure THeaderField.SetName (const FieldName : String);
  Begin
    FFieldName := FieldName;
  End;

Function THeaderField.GetBody : String;
  Begin
    Result := FFieldBody;
  End;

Procedure THeaderField.SetBody (const FieldBody : String);
  Begin
    FFieldBody := FieldBody;
  End;

Function THeaderField.Duplicate : AType;
  Begin
    Result := THeaderField.CreateEx (FFieldName, FFieldBody);
  End;

{ AMessagePart                                                                 }
Procedure AMessagePart.Prepare;
  Begin
  End;

{ AHeader                                                                      }
Function AHeader.Duplicate : AType;
var I : Integer;
    R : AHeader;
  Begin
    R := CreateInstance as AHeader;
    Result := R;
    For I := 0 to FieldCount - 1 do
      R.AddField (FieldByIndex [I].Duplicate as AHeaderField);
  End;

Function AHeader.GetFieldNames : StringArray;
var I, L : Integer;
  Begin
    L := FieldCount;
    SetLength (Result, L);
    For I := 0 to L - 1 do
      Result [I] := FieldByIndex [I].Name;
  End;

Function AHeader.GetFieldIndex (const FieldName : String) : Integer;
var I : Integer;
    S : String;
    F : AHeaderField;
  Begin
    S := UpperCase (Trim (FieldName, SPACE));
    For I := 0 to FieldCount - 1 do
      begin
        F := GetFieldByIndex (I);
        if S = UpperCase (F.Name) then
          begin
            Result := I;
            exit;
          end;
      end;
    Result := -1;
  End;

Procedure AHeader.DeleteField (const FieldName : String);
var I : Integer;
  Begin
    I := GetFieldIndex (FieldName);
    if I >= 0 then
      DeleteFieldByIndex (I);
  End;

Function AHeader.GetField (const FieldName : String) : AHeaderField;
var I : Integer;
  Begin
    I := GetFieldIndex (FieldName);
    if I = -1 then
      Result := nil else
      Result := GetFieldByIndex (I);
  End;

Procedure AHeader.SetField (const FieldName : String; const Field : AHeaderField);
var I : Integer;
  Begin
    if Assigned (Field) and (UpperCase (Field.Name) <> UpperCase (FieldName)) then
      Field.Name := FieldName;
    I := GetFieldIndex (FieldName);
    if I = -1 then
      begin
        if Assigned (Field) then
          AddField (Field);
      end else
      if not Assigned (Field) then
        DeleteFieldByIndex (I) else
        FieldByIndex [I] := Field;
  End;

Function AHeader.GetFieldBody (const FieldName : String) : String;
var F : AHeaderField;
  Begin
    F := GetField (FieldName);
    if not Assigned (F) then
      Result := '' else
      Result := F.Body;
  End;

Procedure AHeader.SetFieldBody (const FieldName : String; const FieldBody : String);
var F : AHeaderField;
  Begin
    F := GetField (FieldName);
    if not Assigned (F) then
      AddField (FieldName, FieldBody) else
      F.Body := FieldBody;
  End;

Function AHeader.HasField (const FieldName : String) : Boolean;
  Begin
    Result := Assigned (GetField (FieldName));
  End;

Procedure AHeader.AddField (const FieldLine : String);
var I : Integer;
  Begin
    I := PosNext (':', FieldLine);
    if I = 0 then
      raise EConvertError.Create ('Not a valid field line');
    AddField (Trim (CopyLeft (FieldLine, I - 1)), Trim (CopyFrom (FieldLine, I + 1)));
  End;

Function AHeader.GetAsString : String;
var I : Integer;
  Begin
    Result := '';
    For I := 0 to FieldCount - 1 do
      Result := Result + FieldByIndex [I].AsString;
    Result := Result + CRLF;
  End;

Procedure AHeader.SetAsString (const Header : String);
var S     : StringArray;
    I, J  : Integer;
    Name,
    Body  : String;
  Begin
    S := Split (Header, CRLF);
    Name := '';
    For I := 0 to Length (S) - 1 do
      begin
        J := PosNext (':', S [I]);
        if J > 0 then // header-field
          begin
            if Name <> '' then
              AddField (Name, Body);
            Name := Trim (CopyLeft (S [I], J - 1), SPACE);
            Body := Trim (CopyFrom (S [I], J + 1), SPACE);
          end else
          if (Name <> '') and (S [I] <> '') and (S [I][1] in SPACE) then // header-line continuation
            Body := Body + S [I] else
            begin
              if Name <> '' then
                AddField (Name, Body);
              Name := '';
            end;
      end;
    if Name <> '' then
      AddField (Name, Body);
  End;

Procedure AHeader.Prepare;
var I : Integer;
  Begin
    For I := 0 to FieldCount - 1 do
      FieldByIndex [I].Prepare;
  End;

{ THeader                                                                      }
class Function THeader.CreateInstance : AType;
  Begin
    Result := THeader.Create;
  End;

Destructor THeader.Destroy;
var I : Integer;
  Begin
    For I := Length (HeaderFields) - 1 downto 0 do
      FreeAndNil (HeaderFields [I]);
    SetLength (HeaderFields, 0);
    inherited Destroy;
  End;

Procedure THeader.AddField (const Field : AHeaderField);
  Begin
    Append (ObjectArray (HeaderFields), Field);
  End;

Procedure THeader.AddField (const FieldName, FieldBody : String);
var F : THeaderField;
  Begin
    F := THeaderField.Create;
    F.Name := FieldName;
    F.Body := FieldBody;
    AddField (F);
  End;

Function THeader.FieldCount : Integer;
  Begin
    Result := Length (HeaderFields);
  End;

Function THeader.GetFieldByIndex (const Idx : Integer) : AHeaderField;
  Begin
    Result := HeaderFields [Idx];
  End;

Procedure THeader.SetFieldByIndex (const Idx : Integer; const Field : AHeaderField);
  Begin
    FreeAndNil (HeaderFields [Idx]);
    HeaderFields [Idx] := Field;
  End;

Procedure THeader.DeleteFieldByIndex (const Idx : Integer);
  Begin
    Remove (ObjectArray (HeaderFields), Idx, 1, True);
  End;


{ TEmailMessageHeader                                                          }
Constructor TEmailMessageHeader.CreateEx (const FromField, ToField, SubjectField : String; const ReplyToField : String; const DateField : TDateTime);
  Begin
    inherited Create;
    AddField ('From', FromField);
    AddField ('To', ToField);
    AddField ('Subject', SubjectField);
    if ReplyToField <> '' then
      AddField ('Reply-To', ReplyToField);
    if DateField <> 0.0 then
      AddField (TDateField.CreateEx (DateField));
  End;

Procedure TEMailMessageHeader.AddField (const FieldName, FieldBody : String);
var S : String;
    H : CHeaderField;
  Begin
    S := LowerCase (FieldName);
    if (S = 'from') or (S = 'reply-to') then   // Note, to: cc: bcc: fields can contain multiple addresses (not supported by TEMailField)
      H := TEMailField else
    if S = 'date' then
      H := TDateField else
      H := nil;
    if Assigned (H) then
      AddField (H.CreateEx (FieldName, FieldBody)) else
      inherited AddField (FieldName, FieldBody);
  End;

Procedure TEMailMessageHeader.Prepare;
var S, T : String;
  Begin
    inherited Prepare;
    if not HasField ('Date') then
      AddField (TDateField.CreateNow);
    if not HasField ('Message-ID') then
      begin
        DecomposeEMailField (FieldBody ['From'], S, T);
        AddField ('Message-ID', MessageIDFieldBody ('', CopyAfter (S, '@')));
      end;
  End;

{                                                                              }
{ Miscellaneous header fields                                                  }
{                                                                              }

{ TDateField                                                                   }
Constructor TDateField.CreateNow;
  Begin
    inherited Create;
    GMTDateTime := LocalTimeToGMTTime (Now);
  End;

Constructor TDateField.CreateEx (const LocalTime : TDateTime);
  Begin
    inherited Create;
    GMTDateTime := LocalTimeToGMTTime (LocalTime);
  End;

Function TDateField.Duplicate : AType;
  Begin
    Result := TDateField.Create;
    TDateField (Result).GMTDateTime := GMTDateTime;
  End;

Function TDateField.GetBody : String;
  Begin
    Result := GMTDateTimeToRFC1123DateTime (GMTDateTime);
  End;

Procedure TDateField.SetBody (const FieldBody : String);
  Begin
    GMTDateTime := RFCDateTimeToGMTDateTime (FieldBody);
  End;

Function TDateField.GetName : String;
  Begin
    Result := 'Date';
  End;

{ TEMailField                                                                  }
Function TEMailField.GetBody : String;
  Begin
    if UserName <> '' then
      Result := '<' + Address + '> (' + UserName + ')' else
      Result := Address;
  End;

Procedure TEMailField.SetBody (const FieldBody : String);
  Begin
    DecomposeEMailField (FieldBody, Address, UserName);
  End;

Function UUHeaderLine (const S : String; var FileName : String) : Boolean;
var T : String;
    R : Boolean;
  Begin
    T := LowerCase (S);
    Result := (Copy (T, 1, 6) = 'begin ') or (PosNext ('--- cut here ---', T) > 0);
    if Result then
      begin
        if Copy (T, 1, 6) = 'begin ' then
          T := TrimLeft (TrimLeft (CopyFrom (T, 7), [' ']), ['0'..'9']);
        Repeat
          T := TrimLeft (T, ['-', ' ']);
          R := Copy (T, 1, 8) = 'cut here';
          if R then
            T := CopyFrom (T, 9);
        Until not R;
        FileName := T;
      end else
      FileName := '';
  End;

{                                                                              }
{ TMailMessageDecoder                                                          }
{                                                                              }
{ Mail Encoding                                                                }

{ See RFC2045                                                                  }
Function DecodeQuotedPrintableText (const T : String) : String;
var I : Integer;
  Begin
    Result := Remove ([#0..#8, #11..#12, #14..#31], T);

    if Result <> '' then
      begin
        // Trim trailing soft line break (=)
        if Result [Length (Result)] = '=' then
          SetLength (Result, Length (Result) - 1);

        // =HH hex representation of char
        I := 0;
        Repeat
          I := PosNext ('=', Result, I);
          if (I > 0) and (I <= Length (Result) - 2)
             and (Result [I + 1] in cs_HexDigit)
             and (Result [I + 2] in cs_HexDigit) then
            begin
              Result [I] := Char (HexToInt (Result [I + 1] + Result [I + 2]));
              System.Delete (Result, I + 1, 2);
            end;
        Until I = 0;

        // Trim trailing spaces
        Result := TrimRight (Result, cs_WhiteSpace);
      end;
  End;

{                                                                              }
{ TTextBodyPart                                                                }
{                                                                              }
Constructor TTextBodyPart.CreateEx (const Body : String);
  Begin
    inherited Create;
    FBody := Body;
  End;

Function TTextBodyPart.GetAsString : String;
  Begin
    Result := FBody;
  End;

Procedure TTextBodyPart.SetAsString (const S : String);
  Begin
    FBody := S;
  End;



{                                                                              }
{ AMessage                                                                     }
{                                                                              }
Function AMessage.GetBodyAsString : String;
var I : Integer;
  Begin
    Result := '';
    For I := 0 to BodyPartCount - 1 do
      Result := Result + GetBodyPart (I).AsString;
  End;

Function AMessage.GetAsString : String;
  Begin
    Result := GetHeader.AsString +
              GetBodyAsString;
  End;

Procedure AMessage.Prepare;
var I : Integer;
  Begin
    Header.Prepare;
    For I := 0 to BodyPartCount - 1 do
      BodyPart [I].Prepare;
  End;



{                                                                              }
{ TMessage                                                                     }
{                                                                              }
Constructor TMessage.CreateEx (const Header : THeader; const BodyParts : array of AMessagePart);
var I : Integer;
  Begin
    inherited Create;
    FHeader := Header;
    SetLength (FBody, High (BodyParts) + 1);
    For I := 0 to High (BodyParts) do
      FBody [I] := BodyParts [I];
  End;

Destructor TMessage.Destroy;
  Begin
    FreeAndNil (FHeader);
    FreeObjectArray (FBody);
    inherited Destroy;
  End;

Function TMessage.GetHeader : AHeader;
  Begin
    if not Assigned (FHeader) then
      FHeader := THeader.Create;
    Result := FHeader;
  End;

Procedure TMessage.SetHeader (const Header : AHeader);
  Begin
    FreeAndNil (FHeader);
    FHeader := Header;
  End;

Function TMessage.GetBodyPart (const Idx : Integer) : AMessagePart;
  Begin
    Result := FBody [Idx];
  End;

Procedure TMessage.SetBodyPart (const Idx : Integer; const BodyPart : AMessagePart);
  Begin
    FBody [Idx] := BodyPart;
  End;

Function TMessage.AddBodyPart (const BodyPart : AMessagePart) : Integer;
  Begin
    Result := Append (ObjectArray (FBody), BodyPart);
  End;

Function TMessage.BodyPartCount : Integer;
  Begin
    Result := Length (FBody);
  End;


  
{                                                                              }
{ TEmailMessage                                                                }
{                                                                              }
Constructor TEmailMessage.CreateEx (const Header : THeader; const Body : AMessagePart);
  Begin
    if Assigned (Body) then
      inherited CreateEx (Header, [Body]) else
      inherited CreateEx (Header, []);
  End;

Constructor TEmailMessage.CreateEx (const Header : THeader; const Body : String);
  Begin
    inherited CreateEx (Header, [TTextBodyPart.CreateEx (Body)]);
  End;

Constructor TEmailMessage.CreateEx (const HdrFrom, HdrTo, HdrSubject : String; const HdrReplyTo : String; const HdrDate : TDateTime; const Body : String);
  Begin
    inherited CreateEx (TEmailMessageHeader.CreateEx (HdrFrom, HdrTo, HdrSubject, HdrReplyTo, HdrDate),
                        [TTextBodyPart.CreateEx (Body)]);
  End;



{                                                                              }
{ HTML Representations                                                         }
{                                                                              }

Constructor AhtmlRenderer.Create (const Parent : AhtmlRenderer);
  Begin
    inherited Create;
    if Assigned (Parent) then
      begin
        FontSize := Parent.FontSize;
        FontColour := Parent.FontColour;
        FontBackColour := Parent.FontBackColour;
        Attributes := Parent.Attributes;
      end else
      begin
        FontSize := 3;
      end;
  End;



{                                                                              }
{ VCard                                                                        }
{                                                                              }
Constructor TVCard.Create;
  Begin
    inherited Create;
    FNickNames := TStringArray.Create;
  End;

Destructor TVCard.Destroy;
  Begin
    FreeAndNil (FNickNames);
    FreeObjectArray (FNames);
    inherited Destroy;
  End;

Procedure TVCard.GenerateUID;
  Begin
    FUID := IntToHex (CalcXOR32 (FFullName) xor RandomUniform) + '-' +
            IntToHex (CalcXOR32 (FEMail) xor RandomUniform) + '-' +
            IntToHex (RandomUniform);
  End;

Function VCardField (const FieldName : String; const FieldBody : Array of String) : String;
var I : Integer;
  Begin
    if (Length (FieldBody) = 1) and (FieldBody [0] = '') then
      Result := '' else
      begin
        Result := FieldName + ':';
        For I := 0 to High (FieldBody) do
          Result := Result +
                  Replace (';', '\;',
                  Replace (',', '\,',
                  Replace (#13#10, '\n',
                  Replace ('\', '\\', FieldBody [I])))) +
                  Cond (I < High (FieldBody), ';', '');
        Result := Result + c_CRLF;
      end;
  End;

Procedure TVCard.AddName (const FamilyName, GivenName, MiddleNames, HonPrefix, HonPostfix : String);
var N : TVCardName;
  Begin
    N := TVCardName.Create;
    N.FamilyName := FamilyName;
    N.GivenName := GivenName;
    N.MiddleNames := MiddleNames;
    N.HonPrefix := HonPrefix;
    N.HonPostfix := HonPostfix;
    Append (ObjectArray (FNames), N);
  End;

Function TVCard.GetAsString : String;
var I : Integer;
    D : TDateTime;
  Begin
    Result := 'BEGIN:VCARD' + c_CRLF +
              VCardField ('VERSION', ['3.0']) +
              VCardField ('FN', FFullName) +
              VCardField ('UID', FUID);
    if FRevision <> 0.0 then
      begin
        D := LocalTimeToGMTTime (FRevision);
        Result := Result + VCardField ('REV', [FormatDateTime ('yyyy-mm-dd', D) +
                           'T' + FormatDateTime ('hh:nn:ss', D) + 'Z']);
      end;
    For I := 0 to Length (FNames) - 1 do
      With FNames [I] do
        Result := Result + VCardField ('N', [FamilyName, GivenName, MiddleNames, HonPrefix, HonPostfix]);
    Result := Result + VCardField ('EMAIL;TYPE=internet', FEMail);
    if FBirthday <> 0.0 then
      Result := Result + VCardField ('BDAY', [FormatDateTime ('yyyy-mm-dd', FBirthday)]);
    For I := 0 to FNickNames.Count - 1 do
      Result := Result + VCardField ('NICKNAME', [FNickNames [I]]);
    Result := Result + VCardField ('SOURCE', FSource) +
              VCardField ('URL', FURL) +
              VCardField ('URL', FIrcURL) +
              'END:VCARD' + c_CRLF;
  End;

Procedure TVCard.Clear;
  Begin
    FFullName := '';
    FUID := '';
    FRevision := 0.0;
    FreeObjectArray (FNames);
    FEMail := '';
    FBirthday := 0.0;
    FNickNames.Clear;
    FSource := '';
    FURL := '';
  End;

Function VCardDateTimeToDateTime (const S : String) : TDateTime;
  Begin
  End;

Function DecodeVCardParam (const S : String) : String;
  Begin
    Result := Replace ('\\', '\',
              Replace ('\;', ';',
              Replace ('\,', ',',
              Replace ('\n', c_CRLF, S))));
  End;

Procedure TVCard.SetAsString (const S : String);
var T, P : StringArray;
    I, J : Integer;
    FN : String;
  Begin
    Clear;
    T := Split (S, c_CRLF);
    For I := 0 to Length (T) - 1 do
      begin
        FN := UpperCase (Trim (CopyBefore (T [I], ':', False)));
        if FN <> '' then
          begin
            P := Split (Trim (CopyAfter (T [I], ':', False)), ';');
            For J := 0 to Length (P) - 1 do
              P [J] := DecodeVCardParam (P [J]);
            if Length (P) > 0 then
              if FN = 'FN' then
                FFullName := P [0] else
              if FN = 'UID' then
                FUID := P [0] else
              if FN = 'REV' then
                FRevision := GMTTimeToLocalTime (VCardDateTimeToDateTime (P [0])) else
              if (FN = 'N') and (Length (P) >= 5)  then
                AddName (P [0], P [1], P [2], P [3], P [4]) else
              if FN = 'BDAY' then
                FBirthday := GMTTimeToLocalTime (VCardDateTimeToDateTime (P [0])) else
              if FN = 'EMAIL;TYPE=INTERNET' then
                FEMail := P [0] else
              if FN = 'NICKNAME' then
                FNickNames.Add (P [0]) else
              if FN = 'SOURCE' then
                FSource := P [0] else
              if FN = 'URL' then
                FURL := P [0];
          end;
      end;
  End;


end.

{                                                                              }
{                                                                              }
{                                                                              }
{  TOnTextMessageLine = Function (var Line : String) : Boolean of object;
  TMailMessageDecoder = class
    protected
    FInlineUUDecode    : Boolean;
    FHeader            : THeader;
    FBody              : TMessagePart;
    FOnTextMessageLine : TOnTextMessageLine;

    public
    Constructor Create;

    Property  InlineUUDecode : Boolean read FInlineUUDecode write FInlineUUDecode;
    Procedure Decode (const Header : String; const Body : TExStream);
    Property  OnTextMessageLine : TOnTextMessageLine read FOnTextMessageLine write FOnTextMessageLine;

    Property  Header : THeader read FHeader write FHeader;
    Property  Body : TMessagePart read FBody write FBody;
  end; }

{Function DecodeTransferEncoding (const T : String; const TransferEncoding : TTransferEncoding) : String;
  Begin
    Case TransferEncoding of
      te_quoted_printable : Result := DecodeQuotedPrintableText (T);
      te_base64           : Result := MIMEBase64Decode (T);
      te_x_uuencode       : Result := UUDecode (T);
      else Result := T;
    end;
  End; }

(*
Constructor TMailMessageDecoder.Create;
  Begin
    inherited Create;
    FInlineUUDecode := True;
  End;

Procedure TMultiPartMessagePart.AddPart (const BodyPart : TMessagePart);
var L : Integer;
  Begin
    L := Length (Body);
    SetLength (Body, L + 1);
    Body [L] := BodyPart;
  end;

Procedure TMailMessageDecoder.Decode (const Header : String; const Body : TExStream);
var Line : String;
    TransferEncoding : TTransferEncoding;
    BodyP : TTextParser;

  Function ExtractMimeHeader : TMimeMessageHeader;
  var H, S : String;
    Begin
      H := '';
      Repeat
        S := BodyP.Readln;
        if S <> '' then
          H := H + S + CRLF;
      Until S = '';
      Result := TMimeMessageHeader.Create;
      Result.Decode (H);
    End;

  Function ExtractRest : String;
    Begin
      Result := '';
      While not Body.EOF do
        begin
          Line := BodyP.Readln;
          Result := Result + DecodeTransferEncoding (Line, TransferEncoding);
        end;
    End;

  Function ExtractToBoundary (const Boundary : String; var OpenBoundary : Boolean) : String;
    Begin
      Result := '';
      OpenBoundary := False;
      While not BodyP.EOF do
        begin
          Line := BodyP.Readln;
          OpenBoundary := Line = '--' + Boundary;
          if OpenBoundary or (Line = '--' + Boundary + '--') then 
            exit;
          Result := Result + DecodeTransferEncoding (Line, TransferEncoding);
        end;
    End;

  Function ExtractMimeBodyPart (const ParentHdr : TMimeMessageHeader) : TMessagePart;
  var Hdr  : TMimeMessageHeader;
      CT   : TMimeContentTypeField;
      Open : Boolean;
    Begin
      if ParentHdr.FBoundary <> '' then
        begin
          CT := TMimeContentTypeField (ParentHdr.GetField ('Content-Type'));
          if CT.MajorType = 'multipart' then
            Result := TMultipartMessagePart.Create else
            Result := TMessagePart.Create;
          ExtractToBoundary (ParentHdr.FBoundary, Open);
          While Open do
            begin
              Hdr := ExtractMimeHeader;
              if CT.MajorType = 'multipart' then
                begin
                  TMultipartMessagePart (Result).AddPart (ExtractMimeBodyPart (Hdr));
                  ExtractToBoundary (ParentHdr.FBoundary, Open);
                end else
                Result.Body := Result.Body + ExtractToBoundary (ParentHdr.FBoundary, Open);
            end;
        end else
        begin
          Result := TMessagePart.Create;
          Result.Body := ExtractRest;
        end;
    End;

  Function NextLine : String;
    Begin
      Result := BodyP.Readln;
    End;

  Function ExtractInlineUUEncoding (var Name : String; var DataBad : Boolean) : String;
  var S, T : String;
      R    : Boolean;
      F    : Integer;
    Begin
      // UU header
      Repeat
        S := NextLine;
        R := UUHeaderLine (S, T);
        if R and (Name = '') then
          Name := T;
        R := R or (Length (T) <= 1);
      Until not R or BodyP.EOF;

      // UU data
      Result := '';
      DataBad := False;
      if not BodyP.EOF then
        Repeat
          if Length (T) > 1 then
            if LowerCase (Copy (T, 1, 3)) = 'end' then
              break else
              begin
                System.Delete (T, 1, 1);
                try
                  Result := Result + UUDecode (T);
                except
                  DataBad := True;
                end;
              end;
          R := BodyP.EOF;
          if not R then
            T := NextLine;
        Until R;

      // UU footer
      F := BodyP.Stream.Position;
      T := LowerCase (NextLine);
      if PosNext ('cut here', T) = 0 then
        BodyP.Stream.Position := F;
    End;

  Function ExtractRFC822BodyPart : TMessagePart;
  var S, N    : String;
      Data    : String;
      DataBad : Boolean;
      MB      : TMultipartMessagePart;
      B       : TMessagePart;
      R       : Boolean;
    Begin
      MB := TMultipartMessagePart.Create;
      While not BodyP.EOF do
        begin
          Data := '';
          R := False;
          While not BodyP.EOF and not R do
            begin
              S := NextLine;
              R := FInlineUUDecode and UUHeaderLine (S, N);
              if not R then
                if not Assigned (FOnTextMessageLine) then
                  Data := Data + S + CRLF else
                  if FOnTextMessageLine (S) then
                    Data := Data + S + CRLF;
            end;

          if Data <> '' then
            begin
              B := TTextMessagePart.Create;
              B.Body := Data;
              MB.AddPart (B);
            end;

          if R then
            begin
              B := TAttachmentMessagePart.Create;
              B.Body := ExtractInlineUUEncoding (N, DataBad);
              TAttachmentMessagePart (B).Name := N;
            end;
        end;
      Result := MB;
    End;

var MHdr : TMimeMessageHeader;
    TE   : TMimeContentTransferEncodingField;
  Begin
    //                                                             Decode header
    MHdr := TMimeMessageHeader.Create;
    MHdr.Decode (Header);
    FHeader := MHdr;

    //                                                               Decode body
    BodyP := TTextParser.Create (Body);
    if MHdr.IsMimeMessage then
      begin
        TE := TMimeContentTransferEncodingField (MHdr.GetField ('Content-Transfer-Encoding'));
        if Assigned (TE) then
          TransferEncoding := TE.TransferEncoding else
          TransferEncoding := te_None;
        FBody := ExtractMimeBodyPart (MHdr);
      end else
      FBody := ExtractRFC822BodyPart;
    FreeAndNil (BodyP);
  End;

*)






























