{ Conditional defines: DEBUGLOG - Log debugging information                    }
{                                                                              }
{ SMTP v0.07 (L0)                                                              }
{                                                                              }
{   Implementation of a SMTP client and server                                 }
{    See RFC 821  - SMTP                                                       }
{        RFC 1869 - ESMTP                                                      }
{        RFC 2554 - Authentication Extention                                   }
{        RFC 2920 - Pipelining Extention                                       }
{        RFC 1845 - Checkpoint/Restart Extention                               }
{        RFC 1870 - Size declaration Extention                                 }
{        RFC 1891 - Delivery Status Notification                               }
{                                                                              }
{ This unit is copyrighted  2000 by David Butler (david@e.co.za)              }
{                                                                              }
{ This unit is part of Delphi Fundamentals, its original name is cSMTP.pas     }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send modifications, suggestions and bug        }
{ reports to david@e.co.za                                                     }
{                                                                              }
{                                                                              }
{ 20/08/2000  v0.01  Inital version. TSMTPClient. TSMTPClientThread.           }
{ 03/10/2000  v0.02  Some refactoring.                                         }
{ 04/10/2000  v0.03  Added TSMTPServer and TSMTPServerClientThread.            }
{ 05/10/2000  v0.04  Added ESMTP support to the server.                        }
{                    Added Authentication (PLAIN and ANONYMOUS)                }
{                    319 lines interface. 848 lines implementation.            }
{ 10/10/2000  v0.05  Support for maximum message sizes (RFC 1870).             }
{ 18/10/2000  v0.06  Support for Delivery Status Notifications (RFC 1891).     }
{ 19/10/2000  v0.07  Refactored TsmtpClient.                                   }
{                                                                              }
unit cSMTP;

interface

uses
  // System units
  SysUtils,
  Classes,

  // Delphi Fundamentals
  cUtils,
  cStreams,
  cInternetStandards,

  // Internet Component Suite (ICS)
  DNSQuery;



{                                                                              }
const
  UnitVersion     = '0.07';
  DefaultSMTPPort = 25;
  DefaultTimeOut  = 5 * 60 * 1000;  // 5 minutes



{ Response classes                                                             }
const
  rc_PositivePreliminary         = 100;
  rc_PositiveCompletion          = 200;
  rc_PositiveIntermediate        = 300;
  rc_TransientNegativeCompletion = 400;
  rc_PermanentNegativeCompletion = 500;

Function IsResponseClass (const Code, ResponseClass : Integer) : Boolean;


{ Status codes                                                                 }
const
  SMTP_Error                     = 500;
  SMTP_ParamSyntaxError          = 501;
  SMTP_CommandNotImplemented     = 502;
  SMTP_BadSequenceOfCommands     = 503;
  SMTP_ParamNotImplemented       = 504;
  SMTP_UnknownParameter          = 555;

  SMTP_SystemStatus              = 211;
  SMTP_HelpMessage               = 214;
  SMTP_ServiceReady              = 220;
  SMTP_ServiceClosingChannel     = 221;
  SMTP_ServiceNotAvailable       = 421;

  SMTP_OK                        = 250;
  SMTP_WillForward               = 251;
  SMTP_TemporaryFailure          = 450;
  SMTP_PermanentFailure          = 550;
  SMTP_ActionAborted             = 451;
  SMTP_UserNotLocal              = 551;
  SMTP_InsufficientStorage       = 452;
  SMTP_AbortInsufficientStorage  = 552;
  SMTP_InvalidName               = 553;
  SMTP_StartMail                 = 354;
  SMTP_TransactionFailed         = 554;

  { VRFY }
  SMTP_VerifyResponse            = 252;

  { AUTH }
  SMTP_PasswordTransitionNeeded  = 432;
  SMTP_MechanismTooWeak          = 534;
  SMTP_EncryptionRequired        = 538;
  SMTP_TemporaryAuthFailure      = 454;
  SMTP_AuthenticationRequired    = 530;
  SMTP_AuthenticationSuccessful  = 235;
  SMTP_AuthenticationFailed      = 535;
  SMTP_AuthenticationNotRequired = 533;
  SMTP_AuthenticationChallenge   = 334;

Function smtpMessage (const StatusCode : Integer) : String;



{                                                                              }
const
  // Delivery verbs
  smtp_MAIL = 'MAIL';
  smtp_SEND = 'SEND';
  smtp_SOML = 'SOML';
  smtp_SAML = 'SAML';

  // DSN return types
  smtp_dsnret_FULL = 'FULL';
  smtp_dsnret_HDRS = 'HDRS';

type
  TsmtpAuthMechanism = (auth_Plain,
                        auth_Anonymous,
                        auth_CRAM_MD5);
  TsmtpDSNNotification = (dsn_success,
                          dsn_failure,
                          dsn_delay,
                          dsn_never);
  TsmtpDSNNotify = set of TsmtpDSNNotification;
  TsmtpRecipient = class
    Recipient   : String;
    dsnOrigRcpt : String;
    dsnNotify   : TsmtpDSNNotify;

    Constructor Create;
    Constructor CreateEx (const TheRecipient : String; const TheDSNOrigRcpt : String = '';
                const TheDSNNotify : TsmtpDSNNotify = [dsn_failure, dsn_delay]);
  end;
  TsmtpRecipientArray = array of TsmtpRecipient;
  TsmtpRecipientList = class
    protected
    FItems : TsmtpRecipientArray;

    Function  GetRecipients : StringArray;
    Procedure SetRecipients (const Recipients : StringArray);

    public
    Destructor Destroy; override;
    Procedure Clear;
    Function  HasRecipients : Boolean;
    Function  Count : Integer;
    Property  Items : TsmtpRecipientArray read FItems;
    Property  Recipients : StringArray read GetRecipients write SetRecipients;
    Procedure Add (const Recipient : TsmtpRecipient);
    Function  GetToFieldBody : String;
  end;

Function dsnNotifyAsString (const N : TsmtpDSNNotify) : String;
Function StringAsDSNNotify (S : String) : TsmtpDSNNotify;


{                                                                              }
{ TSMTPClient                                                                  }
{                                                                              }
type
  EsmtpClient = class (Exception)
    protected
    FsmtpErrorCode : Integer;
    FsmtpErrorMsg  : String;

    public
    Constructor Create (const smtpErrorCode : Integer; const smtpErrorMsg : String; const Msg : String);

    Property smtpErrorCode : Integer read FsmtpErrorCode;
    Property smtpErrorMsg : String read FsmtpErrorMsg;
  end;
  TsmtpProgressEvent = Procedure (const Sender : TObject; const PercentComplete : Integer; var Abort : Boolean) of object;
  TsmtpClient = class
    private
    MsgHdr             : String;
    UsingESMTP         : Boolean;
    UsingSizeExt       : Boolean;
    UsingPipeliningExt : Boolean;
    UsingDSNExt        : Boolean;

    protected
    FHost      : String;
    FPort      : Integer;
    FTimeOut   : Integer;
    FOnLog     : TOnInternetClientLog;

    FEnableESMTP  : Boolean;
    FEnableDSNExt : Boolean;

    FMailer      : String;
    FSenderHost  : String;
    FReturnPath  : String;

    FDeliveryVerb   : String;
    FRecipients     : TsmtpRecipientList;
    FRcptStatusCode : IntegerArray;
    FRcptStatusMsg  : StringArray;
    FMsgHeader      : AHeader;
    FMsgBody        : String;
    FDSNReturn      : String;
    FDSNEnvID       : String;

    FOnProgress : TsmtpProgressEvent;
    FTerminated : Boolean;

    SockStream : TSocketStream;
    Conn : TTextParser;

    Function  SendEHLO (var Msg : String) : Integer;
    Function  SendHELO (var Msg : String) : Integer;
    Function  SendMAIL (var Msg : String) : Integer;
    Function  SendRCPT (var Msg : String; const Recipient : TsmtpRecipient) : Integer;
    Procedure SendQUIT;
    Procedure SendRSET;
    Procedure SendDATA;
    Procedure SendMessageBody;
    Procedure SendRequest (const S : String);
    Function  GetResponse (var Msg : String) : Integer;
    Procedure ProgressCallback (const Complete : Int64; var Abort : Boolean);
    Procedure SetMsgHeader (const MsgHeader : AHeader);
    Procedure Log (const LogType : Char; const LogMsg : String);
    Function  GetLocalSMTPHost : String;
    Procedure Init; virtual;

    public
    Constructor Create;
    Constructor CreateEx (const Host : String;
                          const Port : Integer = DefaultSMTPPort;
                          const TimeOut : Integer = DefaultTimeOut;
                          const SenderHost : String = '';
                          const ReturnPath : String = '';
                          const Recipients : StringArray = nil;
                          const MsgHeader : AHeader = nil;
                          const MsgBody : String = '');
    Destructor Destroy; override;

    Property  Host : String read FHost write FHost;
    Property  Port : Integer read FPort write FPort;
    Property  TimeOut : Integer read FTimeOut write FTimeOut;
    Property  OnLog : TOnInternetClientLog read FOnLog write FOnLog;

    Property  EnableESMTP : Boolean read FEnableESMTP write FEnableESMTP;
    Property  EnableDSNExtention : Boolean read FEnableDSNExt write FEnableDSNExt;

    Property  Mailer : String read FMailer write FMailer;
    Property  SenderHost : String read FSenderHost write FSenderHost;
    Property  ReturnPath : String read FReturnPath write FReturnPath;

    Property  DeliveryVerb : String read FDeliveryVerb write FDeliveryVerb;
    Property  MsgHeader : AHeader read FMsgHeader write SetMsgHeader;
    Property  MsgBody : String read FMsgBody write FMsgBody;
    Property  Recipients : TsmtpRecipientList read FRecipients;
    Property  DSNReturn : String read FDSNReturn write FDSNReturn;
    Property  DSNEnvID : String read FDSNEnvID write FDSNEnvID;

    Procedure Prepare;
    Procedure SendMessage;
    Property  OnProgress : TsmtpProgressEvent read FOnProgress write FOnProgress;
    Property  RecipientStatusCode : IntegerArray read FRcptStatusCode;
    Property  RecipientStatusMesssage : StringArray read FRcptStatusMsg;
    Property  Terminated : Boolean read FTerminated write FTerminated;
  end;



{                                                                              }
{ TsmtpClientThread                                                            }
{                                                                              }
type
  EsmtpClientThread = class (Exception);
  TsmtpClientThread = class (TThread)
    private
    FMXLookupDone   : Boolean;
    FMXLookupResult : String;
    FLogType        : Char;
    FLogMsg         : String;

    protected
    FOnLog           : TOnInternetClientLog;
    FUseHostMX       : Boolean;
    FDNSServer       : String;
    FMXLookup        : TDnsQuery;
    FClient          : TsmtpClient;
    FOnProgress      : TsmtpProgressEvent;
    FAbort           : Boolean;
    FPercentComplete : Integer;
    FSuccess         : Boolean;
    FErrorCode       : Integer;
    FErrorMsg        : String;
    FPermFailure     : Boolean;

    Procedure OnSMTPClientLog (const Sender : TObject; const LogType : Char; const LogMsg : String);
    Procedure OnSMTPClientProgress (const Sender : TObject; const PercentComplete : Integer; var Abort : Boolean);
    Procedure OnMXLookupDone (Sender : TObject; Error : Word);
    Procedure NotifyProgress;
    Procedure NotifyLog;
    Procedure Init;

    public
    Constructor Create;
    Constructor CreateEx (const Host : String; const UseHostMX : Boolean); overload;
    Constructor CreateEx (const Client : TsmtpClient); overload;
    Destructor Destroy; override;

    Property  OnLog : TOnInternetClientLog read FOnLog write FOnLog;
    Property  Client : TsmtpClient read FClient;
    Property  UseHostMXAddress : Boolean read FUseHostMX write FUseHostMX;
    Property  DNSServer : String read FDNSServer write FDNSServer;

    Procedure Execute; override;

    Property  OnProgress : TsmtpProgressEvent read FOnProgress write FOnProgress;
    Property  PercentComplete : Integer read FPercentComplete;

    Property  Success : Boolean read FSuccess;
    Property  PermanentFailure : Boolean read FPermFailure;
    Property  ErrorCode : Integer read FErrorCode;
    Property  ErrorMsg : String read FErrorMsg;
    Property  Terminated;
  end;



{                                                                              }
{ TsmtpServer                                                                  }
{                                                                              }
type
  TsmtpServerAuthenticateAction = (auth_Failed, auth_Successful, auth_TooWeak);
  TsmtpServerClientThread = class (AInternetServerClientThread)
    private
    FResponseCode    : Integer;
    FResponseMsg     : String;
    FResponseList    : StringArray;
    FVerifyName      : String;
    FRecipient       : String;
    FRecipientNotify : TsmtpDSNNotify;
    FAuthRequired    : Boolean;
    FAllowAccess     : Boolean;
    FAuthenticated   : Boolean;
    FAuthMech        : TsmtpAuthMechanism;
    FUserName        : String;
    FPassword        : String;
    FAuthAction      : TsmtpServerAuthenticateAction;

    protected
    FLocalSMTPHost  : String;
    FRemoteSMTPHost : String;
    FDeliveryVerb   : String;
    FReversePath    : String;
    FRecipients     : TsmtpRecipientList;
    FRawMsg         : String;
    FMsgHeader      : THeader;
    FMsgBody        : String;
    FDSNReturn      : String;
    FDSNEnvID       : String;
    FDSNOrigRcpt    : String;

    Procedure ClientServe; override;
    Procedure NotifyHello;
    Procedure NotifyRecipient;
    Procedure NotifyReversePath;
    Procedure NotifyVerify;
    Procedure NotifyExpand;
    Procedure NotifyMessage;
    Procedure NotifyAuthenticate;

    public
    Destructor Destroy; override;
    Property LocalSMTPHost : String read FLocalSMTPHost;
    Property RemoteSMTPHost : String read FRemoteSMTPHost;
    Property DeliveryVerb : String read FDeliveryVerb;
    Property ReversePath : String read FReversePath;
    Property Recipients : TsmtpRecipientList read FRecipients;
    Property RawMessage : String read FRawMsg;
    Property MessageHeader : THeader read FMsgHeader;
    Property MessageBody : String read FMsgBody;
    Function RemoteHostMatch (const Mask : String) : Boolean;
    Property DSNReturn : String read FDSNReturn;
    Property DSNEnvID : String read FDSNEnvID;
    Property DSNOrigRcpt : String read FDSNOrigRcpt;
  end;
  TsmtpServerClientEvent = Procedure (const Sender : TsmtpServerClientThread;
      var ResponseCode : Integer; var ResponseMsg : String) of object;
  TsmtpServerHelloEvent = Procedure (const Sender : TsmtpServerClientThread;
      var AuthRequired : Boolean; var AllowAccess : Boolean) of object;
  TsmtpServerRecipientEvent = Procedure (const Sender : TsmtpServerClientThread;
      const Recipient : String; var ResponseCode : Integer; var ResponseMsg : String) of object;
  TsmtpServerVerifyEvent = Procedure (const Sender : TsmtpServerClientThread;
      const Name : String; var ResponseCode : Integer; var ResponseMsg : String) of object;
  TsmtpServerExpandEvent = Procedure (const Sender : TsmtpServerClientThread;
      const Name : String; var ResponseList : StringArray;
      var ResponseCode : Integer; var ResponseMsg : String) of object;
  TsmtpServerAuthenticateEvent = Procedure (const Sender : TsmtpServerClientThread;
      const AuthMechanism : TsmtpAuthMechanism; const UserName : String; var Password : String;
      var Action : TsmtpServerAuthenticateAction) of object;
  ESMTPServer = class (EInternetServer);
  TsmtpServer = class (AInternetServer)
    protected
    FHostName               : String;
    FServerReady            : Boolean;
    FServerBusyMsg          : String;
    FClientIdentityRequired : Boolean;
    FRecipientRequired      : Boolean;
    FMaxMessageSize         : Cardinal;
    FOnHello                : TsmtpServerHelloEvent;
    FOnRecipient            : TsmtpServerRecipientEvent;
    FOnReversePath          : TsmtpServerClientEvent;
    FOnVerify               : TsmtpServerVerifyEvent;
    FOnExpand               : TsmtpServerExpandEvent;
    FOnMessage              : TsmtpServerClientEvent;
    FOnAuthenticate         : TsmtpServerAuthenticateEvent;
    FEnableAuthExt          : Boolean;
    FEnablePipeliningExt    : Boolean;
    FEnableDSNExt           : Boolean;

    Procedure Init; override;

    public
    Constructor CreateEx (const AOwner : TComponent; const TimeOut : Integer);

    Property HostName : String read FHostName write FHostName;
    Property ServerReady : Boolean read FServerReady write FServerReady;
    Property ServerBusyMsg : String read FServerBusyMsg write FServerBusyMsg;
    Property ClientIdentityRequired : Boolean read FClientIdentityRequired write FClientIdentityRequired;
    Property RecipientRequired : Boolean read FRecipientRequired write FRecipientRequired;
    Property MaxMessageSize : Cardinal read FMaxMessageSize write FMaxMessageSize;
    Property EnableAuthenticationExtention : Boolean read FEnableAuthExt write FEnableAuthExt;
    Property EnablePipeliningExtention : Boolean read FEnablePipeliningExt write FEnablePipeliningExt;
    Property EnableDSNExtention : Boolean read FEnableDSNExt write FEnableDSNExt;

    Property OnHello : TsmtpServerHelloEvent read FOnHello write FOnHello;
    Property OnRecipient : TsmtpServerRecipientEvent read FOnRecipient write FOnRecipient;
    Property OnReversePath : TsmtpServerClientEvent read FOnReversePath write FOnReversePath;
    Property OnVerify : TsmtpServerVerifyEvent read FOnVerify write FOnVerify;
    Property OnExpand : TsmtpServerExpandEvent read FOnExpand write FOnExpand;
    Property OnMessage : TsmtpServerClientEvent read FOnMessage write FOnMessage;
    Property OnAuthenticate : TsmtpServerAuthenticateEvent read FOnAuthenticate write FOnAuthenticate;
  end;

{ Server mail messages                                                         }
Function DeliveryNotificationStatusMessage (const NotificationType : TsmtpDSNNotification;
         const Hdr : AHeader; const ServerDomain, FailureReason : String;
         const Recipient : TsmtpRecipient) : AMessage;



{ Forward-path functions                                                       }
Function MailboxFromPath (const Path : String) : String;
Function ForwardPathFromPath (const Path : String) : String;
Function ForwardHostFromPath (const Path : String) : String;



{ Reverse-path functions                                                       }
Function AddHostToReversePath (const ReversePath, Host : String) : String;



implementation

uses
  Forms,
  Windows,

  // Delphi Fundamentals
  cStrings,
  cDateTime,
  cMaths,
  cMIME;



{                                                                              }
Function IsResponseClass (const Code, ResponseClass : Integer) : Boolean;
  Begin
    Result := (Code div 100) = (ResponseClass div 100);
  End;

Function SMTPMessage (const StatusCode : Integer) : String;
  Begin
    Case StatusCode of
      SMTP_Error                     : Result := 'Syntax error, command unrecognized';
      SMTP_ParamSyntaxError          : Result := 'Syntax error in parameters or arguments';
      SMTP_CommandNotImplemented     : Result := 'Command not implemented';
      SMTP_BadSequenceOfCommands     : Result := 'Bad sequence of commands';
      SMTP_ParamNotImplemented       : Result := 'Command parameter not implemented';
      SMTP_SystemStatus              : Result := 'System status, or system help reply';
      SMTP_HelpMessage               : Result := 'Help message';
      SMTP_ServiceReady              : Result := 'Service ready';
      SMTP_ServiceClosingChannel     : Result := 'Service closing transmission channel';
      SMTP_ServiceNotAvailable       : Result := 'Service not available, closing transmission channel';
      SMTP_OK                        : Result := 'OK';
      SMTP_WillForward               : Result := 'User not local; will forward';
      SMTP_TemporaryFailure          : Result := 'Temporary failure';
      SMTP_PermanentFailure          : Result := 'Permanent failure';
      SMTP_ActionAborted             : Result := 'Requested action aborted: error in processing';
      SMTP_UserNotLocal              : Result := 'User not local';
      SMTP_InsufficientStorage       : Result := 'Requested action not taken: insufficient system storage';
      SMTP_AbortInsufficientStorage  : Result := 'Requested mail action aborted: exceeded storage allocation';
      SMTP_InvalidName               : Result := 'Invalid name';
      SMTP_StartMail                 : Result := 'Start mail input; end with <CRLF>.<CRLF>';
      SMTP_TransactionFailed         : Result := 'Transaction failed';
      SMTP_UnknownParameter          : Result := 'Unknown parameter';
      SMTP_PasswordTransitionNeeded  : Result := 'Password transition needed';
      SMTP_MechanismTooWeak          : Result := 'Authentication mechanism too weak';
      SMTP_EncryptionRequired        : Result := 'Encryption required for requested authentication mechanism';
      SMTP_TemporaryAuthFailure      : Result := 'Temporary authentication failure';
      SMTP_AuthenticationRequired    : Result := 'Authentication required';
      SMTP_AuthenticationSuccessful  : Result := 'Authentication successful';
      SMTP_AuthenticationFailed      : Result := 'Authentication failed';
      SMTP_AuthenticationNotRequired : Result := 'Authentication not required';
      else
        Case (StatusCode div 100) * 100 of
          rc_PositivePreliminary         : Result := 'Positive preliminary response code ' + IntToStr (StatusCode);
          rc_PositiveCompletion          : Result := 'Positive completion code ' + IntToStr (StatusCode);
          rc_PositiveIntermediate        : Result := 'Positive intermediate response code ' + IntToStr (StatusCode);
          rc_TransientNegativeCompletion : Result := 'Transient negative completion code ' + IntToStr (StatusCode);
          rc_PermanentNegativeCompletion : Result := 'Permanent negative completion code ' + IntToStr (StatusCode);
          else Result := 'Status code ' + IntToStr (StatusCode);
        end;
    end;
  End;


{ See RFC 1891 - DSN                                                           }
{     xtext = *( xchar / hexchar )                                             }
{     xchar = any ASCII CHAR between "!" (33) and "~" (126) inclusive,         }
{             except for "+" and "=".                                          }
{     hexchar = ASCII "+" immediately followed by two upper case               }
{               hexadecimal digits                                             }
const
  XChar = ['!'..'~'] - ['+', '='];
  EscXChar = [#$00..#$FF] - XChar;

Function EncodeXText (const S : String) : String;
  Begin
    Result := HexEscapeText (S, EscXChar, '+', True);
  End;

Function DecodeXText (const S : String) : String;
  Begin
    Result := HexUnescapeText (S, '+');
  End;




{ TsmtpRecipient                                                               }
Constructor TsmtpRecipient.Create;
  Begin
    inherited Create;
    DSNNotify := [dsn_Failure, dsn_Delay];
  End;

Constructor TsmtpRecipient.CreateEx (const TheRecipient : String; const TheDSNOrigRcpt : String; const TheDSNNotify : TsmtpDSNNotify);
  Begin
    inherited Create;
    Recipient := TheRecipient;
    DSNOrigRcpt := TheDSNOrigRcpt;
    DSNNotify := TheDSNNotify;
  End;

{ TsmtpRecipientList                                                           }
Procedure TsmtpRecipientList.Clear;
  Begin
    FreeAndNilObjectArray (ObjectArray (FItems));
  End;

Destructor TsmtpRecipientList.Destroy;
  Begin
    Clear;
    inherited Destroy;
  End;

Function TsmtpRecipientList.HasRecipients : Boolean;
  Begin
    Result := Length (FItems) > 0;
  End;

Function TsmtpRecipientList.Count : Integer;
  Begin
    Result := Length (FItems);
  End;

Function TsmtpRecipientList.GetToFieldBody : String;
  Begin
    Result := Join (GetRecipients, ',');
  End;

Function TsmtpRecipientList.GetRecipients : StringArray;
var I, L : Integer;
  Begin
    L := Length (FItems);
    SetLength (Result, L);
    For I := 0 to L - 1 do
      Result [I] := FItems [I].Recipient;
  End;

Procedure TsmtpRecipientList.SetRecipients (const Recipients : StringArray);
var I, L : Integer;
  Begin
    Clear;
    L := Length (Recipients);
    SetLength (FItems, L);
    For I := 0 to L - 1 do
      FItems [I] := TsmtpRecipient.CreateEx (Recipients [I]);
  End;

Procedure TsmtpRecipientList.Add (const Recipient : TsmtpRecipient);
  Begin
    Append (ObjectArray (FItems), Recipient);
  End;


{                                                                              }
{ TsmtpClient                                                                  }
{                                                                              }
const
  SMTPClientVersion = 'Fundamentals ESMTP v' + UnitVersion;

Constructor ESMTPClient.Create (const SMTPErrorCode : Integer; const SMTPErrorMsg : String; const Msg : String);
var S : String;
  Begin
    FsmtpErrorCode := SMTPErrorCode;
    FsmtpErrorMsg := SMTPErrorMsg;
    S := Msg;
    if SMTPErrorCode > 0 then
      begin
        S := S + #13 + 'SMTP Error ' + IntToStr (SMTPErrorCode) + ' (';
        if SMTPErrorMsg <> '' then
          S := S + SMTPErrorMsg else
          S := S + SMTPMessage (SMTPErrorCode);
        S := S + ')';
      end;
    inherited Create (S);
  End;

Procedure TsmtpClient.Init;
  Begin
    FPort := DefaultSMTPPort;
    FTimeOut := DefaultTimeOut;
    FMailer := '';
    FRecipients := TsmtpRecipientList.Create;
    FEnableESMTP := True;
    FEnableDSNExt := True;
    FDeliveryVerb := smtp_MAIL;
    FDSNReturn := '';
  End;

Constructor TsmtpClient.Create;
  Begin
    inherited Create;
    Init;
  End;

Constructor TsmtpClient.CreateEx (const Host : String; const Port : Integer; const TimeOut : Integer; const SenderHost : String; const ReturnPath : String; const Recipients : StringArray; const MsgHeader : AHeader; const MsgBody : String);
  Begin
    inherited Create;
    Init;
    FHost := Host;
    FPort := Port;
    FTimeOut := TimeOut;
    FSenderHost := SenderHost;
    FReturnPath := ReturnPath;
    FRecipients.Recipients := Recipients;
    FMsgHeader := MsgHeader;
    FMsgBody := MsgBody;
  End;

Destructor TsmtpClient.Destroy;
  Begin
    FreeAndNil (FRecipients);
    FreeAndNil (FMsgHeader);
    inherited Destroy;
  End;

Procedure TsmtpClient.Log (const LogType : Char; const LogMsg : String);
  Begin
    if Assigned (FOnLog) then
      FOnLog (self, LogType, LogMsg);
  End;

Procedure TsmtpClient.SetMsgHeader (const MsgHeader : AHeader);
  Begin
    FreeAndNil (FMsgHeader);
    FMsgHeader := MsgHeader;
  End;

Function TsmtpClient.GetResponse (var Msg : String) : Integer;
var S, T : String;
    I : Integer;
  Begin
    if FTerminated then
      raise EsmtpClient.Create (-1, '', 'Session terminated.');
    S := Conn.Readln;
    {$IFDEF DEBUGLOG}
    Log (lt_Debug, 'S: ' + S);
    {$ENDIF}
    I := StrToIntDef (CopyBefore (S, '-', False), -1);
    if I > 0 then // multi-line response
      begin
        Result := I;
        Msg := '';
        Repeat
          Msg := Msg + CopyAfter (S, '-', False) + c_CRLF;
          S := Conn.Readln;
        Until not Match (IntToStr (I) + '-', S);
        if Match (IntToStr (I) + ' ', S) then
          Msg := Msg + CopyFrom (S, 4);
      end else
      begin
        Result := StrToIntDef (CopyLeft (S, 3), -1);
        Msg := Trim (CopyFrom (S, 4));
      end;
  End;

Procedure TsmtpClient.SendRequest (const S : String);
  Begin
    if FTerminated then
      raise EsmtpClient.Create (-1, '', 'Session terminated.');
    Conn.Writeln (S);
    {$IFDEF DEBUGLOG}
    Log (lt_Debug, 'C: ' + S);
    {$ENDIF}
  End;

Function TsmtpClient.GetLocalSMTPHost : String;
  Begin
    if FSenderHost <> '' then
      Result := FSenderHost else
      Result := SockStream.LocalHost;
  End;

Function TsmtpClient.SendEHLO (var Msg : String) : Integer;
  Begin
    FSenderHost := GetLocalSMTPHost;
    SendRequest ('EHLO ' + FSenderHost);
    Result := GetResponse (Msg);
  End;

Function TsmtpClient.SendHELO (var Msg : String) : Integer;
  Begin
    FSenderHost := GetLocalSMTPHost;
    SendRequest ('HELO ' + FSenderHost);
    Result := GetResponse (Msg);
  End;
  
Function ValidPath (const A : String) : String;
  Begin
    Result := '<' + TrimLeft (TrimRight (A, [#0..' ', '>']), [#0..' ', '<']) + '>';
  End;

Function TsmtpClient.SendMAIL (var Msg : String) : Integer;
var S, V : String;
    TryAgain : Boolean;
    TryingSizeExt  : Boolean;
    TryingDSNRet   : Boolean;
    TryingDSNEnvID : Boolean;
    TryingVerbExt  : Boolean;
  Begin
    TryingSizeExt := UsingESMTP and UsingSizeExt;
    TryingDSNRet := UsingESMTP and UsingDSNExt and (FDSNReturn <> '');
    TryingDSNEnvID := UsingESMTP and UsingDSNExt and (FDSNEnvID <> '');
    V := UpperCase (DeliveryVerb);
    TryingVerbExt := V <> smtp_MAIL;
    Repeat
      if TryingVerbExt then
        S := V else
        S := smtp_MAIL;
      S := S + ' FROM:' + ValidPath (FReturnPath);
      if TryingSizeExt then
        S := S + ' SIZE:' + IntToStr (Length (MsgHdr) + Length (MsgBody));
      if TryingDSNRet then
        S := S + ' RET=' + FDSNReturn;
      if TryingDSNEnvID then
        S := S + ' ENVID=' + FDSNEnvID;
      SendRequest (S);
      Result := GetResponse (Msg);
      if (TryingDSNRet or TryingDSNEnvID or TryingSizeExt) and
         ((Result = SMTP_Error) or (Result = SMTP_ParamSyntaxError) or
          (Result = SMTP_ParamNotImplemented) or (Result = SMTP_UnknownParameter)) then
        begin
          TryAgain := True;
          if TryingSizeExt then
            TryingSizeExt := False else
            if TryingDSNRet then
              TryingDSNRet := False else
              TryingDSNEnvID := False;
        end else
        if TryingVerbExt and ((Result = SMTP_Error) or (Result = SMTP_CommandNotImplemented)) then
          begin
            TryingVerbExt := False;
            TryAgain := (V = smtp_SOML) or (V = smtp_SAML);
          end else
          TryAgain := False;
    Until not TryAgain;
  End;

Function dsnNotifyAsString (const N : TsmtpDSNNotify) : String;
var T : StringArray;
  Begin
    T := nil;
    if dsn_never in N then
      Append (T, 'NEVER') else
      begin
        if dsn_success in N then
          Append (T, 'SUCCESS');
        if dsn_failure in N then
          Append (T, 'FAILURE');
        if dsn_delay in N then
          Append (T, 'DELAY');
      end;
    if Length (T) > 0 then
      Result := Join (T, ',') else
      Result := '';
  End;

Function StringAsDSNNotify (S : String) : TsmtpDSNNotify;
var M : StringArray;
    F : Integer;
  Begin
    Result := [];
    S := UpperCase (Trim (S));
    if S = '' then
      exit;
    if S = 'NEVER' then
      begin
        Result := [dsn_never];
        exit;
      end;
    M := Split (S, ',');
    For F := 0 to Length (M) - 1 do
      M [F] := Trim (M [F]);
    For F := 0 to Length (M) - 1 do
      if M [F] = 'SUCCESS' then
        Result := Result + [dsn_Success] else
      if M [F] = 'FAILURE' then
        Result := Result + [dsn_Failure] else
      if M [F] = 'DELAY' then
        Result := Result + [dsn_Delay] else
        raise EConvertError.Create ('Invalid value for NOTIFY parameter.');
  End;

Function TsmtpClient.SendRCPT (var Msg : String; const Recipient : TsmtpRecipient) : Integer;
var S, T : String;
    TryAgain : Boolean;
    TryingDSNExt : Boolean;
  Begin
    TryingDSNExt := UsingESMTP and UsingDSNExt;
    Repeat
      S := 'RCPT TO:' + ValidPath (Recipient.Recipient);
      if TryingDSNExt then
        begin
          T := dsnNotifyAsString (Recipient.dsnNotify);
          if T <> '' then
            S := S + ' NOTIFY=' + T;
          if Recipient.dsnOrigRcpt <> '' then
            S := S + ' ORCPT=' + Recipient.dsnOrigRcpt;
          TryingDSNExt := (T <> '') or (Recipient.dsnOrigRcpt <> '');
        end;
      SendRequest (S);
      Result := GetResponse (Msg);
      TryAgain := (TryingDSNExt) and
                  ((Result = SMTP_Error) or (Result = SMTP_ParamSyntaxError) or
                  (Result = SMTP_ParamNotImplemented) or (Result = SMTP_UnknownParameter));
      if TryAgain then
        TryingDSNExt := False;
    Until not TryAgain;
  End;

Procedure TsmtpClient.SendQUIT;
  Begin
    SendRequest ('QUIT');
  End;

Procedure TsmtpClient.SendRSET;
  Begin
    SendRequest ('RSET');
  End;

Procedure TsmtpClient.SendDATA;
  Begin
    SendRequest ('DATA');
  End;

Procedure TsmtpClient.ProgressCallback (const Complete : Int64; var Abort : Boolean);
  Begin
    if Assigned (FOnProgress) then
      FOnProgress (self, Trunc (Complete / Length (MsgBody) * 100), Abort);
  End;

Procedure TsmtpClient.Prepare;
var S : String;                    
  Begin
    FMsgHeader.Prepare;
    if not FRecipients.HasRecipients then
      raise ESMTPClient.Create (-1, '', 'No recipients.');
    FMsgHeader.FieldBody ['To'] := FRecipients.GetToFieldBody;
    if not FMsgHeader.HasField ('Date') then
      FMsgHeader.AddField (TDateField.CreateNow);
    if not FMsgHeader.HasField ('Message-id') then
      begin
        if SenderHost = '' then
          S := CopyAfter (MailboxFromPath (FReturnPath), '@', False) else
          S := SenderHost;
        FMsgHeader.AddField ('Message-id', MessageIDFieldBody ('', S));
      end;
    if not FMsgHeader.HasField ('X-Mailer') then
      if FMailer <> '' then
        FMsgHeader.AddField ('X-Mailer', FMailer + ' (' + SMTPClientVersion + ')') else
        FMsgHeader.AddField ('X-Mailer', SMTPClientVersion);
  End;

Procedure TsmtpClient.SendMessageBody;
  Begin
    Conn.Stream.Write (MsgHdr);
    StreamDotLineTerminated (MsgBody, Conn.Stream, ProgressCallback);
  End;

Procedure TsmtpClient.SendMessage;
var I, J, L : Integer;
    S : String;
    One, All : Boolean;
    V : StringArray;
  Begin
    if not FRecipients.HasRecipients then
      raise ESMTPClient.Create (-1, '', 'No recipients.');
    if FHost = '' then
      raise ESMTPClient.Create (-1, '', 'No SMTP host.');
    Prepare;
    MsgHdr := FMsgHeader.AsString;

    UsingESMTP := False;
    UsingSizeExt := False;
    UsingPipeliningExt := False;
    UsingDSNExt := False;
    try
      SockStream := TSocketStream.Create (FHost, FPort, FTimeOut, True);
    except
      on E : Exception do
        raise ESMTPClient.Create (-1, '', 'Error connecting to SMTP server.'#13 + E.Message);
    end;
    Conn := TTextParser.Create (SockStream);
    try try
      Conn.EOLType := [eol_CRLF, eol_EOF];
      I := GetResponse (S);
      if not IsResponseClass (I, rc_PositiveCompletion) then
        raise ESMTPClient.Create (I, S, 'Server not ready');

      if FEnableESMTP then
        begin
          I := SendEHLO (S);
          if not IsResponseClass (I, rc_PositiveCompletion) then
            I := SendHELO (S) else
            begin
              UsingESMTP := True;
              V := Split (S, CRLF);
              For J := 1 to Length (V) - 1 do
                begin
                  S := UpperCase (CopyBefore (V [J], ' ', True));
                  if S = 'SIZE' then
                    UsingSizeExt := True else
                  if S = 'PIPELINING' then
                    UsingPipeliningExt := True else
                  if S = 'DSN' then
                    UsingDSNExt := True;
                end;
            end;
        end else
        I := SendHELO (S);
      if not IsResponseClass (I, rc_PositiveCompletion) then
        raise ESMTPClient.Create (I, S, 'Server not ready for this host');

      I := SendMAIL (S);
      if not IsResponseClass (I, rc_PositiveCompletion) then
        raise ESMTPClient.Create (I, S, 'Server won''t initiate mail session');

      L := FRecipients.Count;
      SetLength (FRcptStatusCode, L);
      SetLength (FRcptStatusMsg, L);
      One := False;
      All := True;
      For J := 0 to L - 1 do
        begin
          I := SendRCPT (S, FRecipients.Items [J]);
          FRcptStatusCode [J] := I;
          FRcptStatusMsg [J] := S;
          if IsResponseClass (I, rc_PositiveCompletion) then
            One := True else
            All := False;
        end;
      if not One then
        raise ESMTPClient.Create (I, S, 'Recipient not accepted');

      SendDATA;
      I := GetResponse (S);
      if not IsResponseClass (I, rc_PositiveCompletion) and not IsResponseClass (I, rc_PositiveIntermediate) then
        raise ESMTPClient.Create (I, S, 'Server won''t accept message');

      SendMessageBody;
      I := GetResponse (S);
      if not IsResponseClass (I, rc_PositiveCompletion) then
        raise ESMTPClient.Create (I, S, 'Message not sent');
    except
      if not FTerminated then
        try
          SendRSET;
          I := GetResponse (S);
        except end;
      raise;
    end;
    finally
      if not FTerminated then
        try
          SendQUIT;
          I := GetResponse (S);
        except end;
      FreeAndNil (Conn);
    end;
  End;



{                                                                              }
{ TsmtpClientThread                                                            }
{                                                                              }
Procedure TsmtpClientThread.Init;
  Begin
    FClient.OnProgress := OnSMTPClientProgress;
    FClient.OnLog := OnSMTPClientLog;
    FreeOnTerminate := True;
  End;

Constructor TsmtpClientThread.Create;
  Begin
    inherited Create (True);
    FClient := TsmtpClient.Create;
    Init;
  End;

Constructor TsmtpClientThread.CreateEx (const Client : TsmtpClient);
  Begin
    inherited Create (True);
    FClient := Client;
    Init;
  End;

Constructor TsmtpClientThread.CreateEx (const Host : String; const UseHostMX : Boolean);
  Begin
    inherited Create (True);
    FClient := TsmtpClient.CreateEx (Host);
    FUseHostMX := UseHostMX;
    Init;
  End;

Destructor TsmtpClientThread.Destroy;
  Begin
    FreeAndNil (FMXLookup);
    FreeAndNil (FClient);
    inherited Destroy;
  End;

Procedure TsmtpClientThread.NotifyLog;
  Begin
    if Assigned (FOnLog) then
      FOnLog (self, FLogType, FLogMsg);
  End;

Procedure TsmtpClientThread.OnSMTPClientLog (const Sender : TObject; const LogType : Char; const LogMsg : String);
  Begin
    if Assigned (FOnLog) then
      begin
        FLogType := LogType;
        FLogMsg := LogMsg;
        Synchronize (NotifyLog);
      end;
  End;

Procedure TsmtpClientThread.OnMXLookupDone (Sender : TObject; Error : Word);
  Begin
    FMXLookupDone := True;
    if Error <> 0 then
      FMXLookupResult := '' else
      FMXLookupResult := FMXLookup.MXExchange [0];
  End;

Procedure TsmtpClientThread.Execute;
  Begin
    try
      if FUseHostMX then
        begin
          FMXLookupDone := False;
          FMXLookup := TDnsQuery.Create (nil);
          FMXLookup.WSocket.MultiThreaded := True;
          FMXLookup.OnRequestDone := OnMXLookupDone;
          FMXLookup.Addr := FDNSServer;
          FMXLookup.MXLookup (FClient.Host);
          While not Terminated and not FMXLookupDone do
            FMXLookup.WSocket.ProcessMessages;
          if FMXLookupResult = '' then
            raise EsmtpClientThread.Create ('MX Address lookup failed.');
          FClient.Host := FMXLookupResult;
        end;

      FClient.Prepare;
      FClient.SendMessage;
      FSuccess := not Terminated;
    except
      on E : ESMTPClient do
        begin
          FSuccess := False;
          FErrorCode := E.FSMTPErrorCode;
          FErrorMsg := E.Message;
          FPermFailure := IsResponseClass (FErrorCode, rc_PermanentNegativeCompletion);
        end;
      on E : Exception do
        begin
          FSuccess := False;
          FErrorCode := -1;
          FErrorMsg := E.Message;
          FPermFailure := False;
        end;
    end;
  End;

Procedure TsmtpClientThread.NotifyProgress;
  Begin
    FAbort := False;
    if Assigned (FOnProgress) then
      FOnProgress (self, FPercentComplete, FAbort);
  End;

Procedure TsmtpClientThread.OnSMTPClientProgress (const Sender : TObject; const PercentComplete : Integer; var Abort : Boolean);
  Begin
    FPercentComplete := PercentComplete;
    if Terminated then
      FAbort := True else
      Synchronize (NotifyProgress);
    Abort := FAbort;
    if Abort then
      FClient.Terminated := True;
  End;



{                                                                              }
{ TsmtpServer                                                                  }
{                                                                              }
const
  SMTPServerVersion = 'Fundamentals ESMTP v' + UnitVersion;

Function MailboxFromPath (const Path : String) : String;
  Begin
    Result := Trim (CopyAfter (CopyBetween (Path, '<', '>', True, True), ':', True));
    if (Result <> '') and (PosNext ('@', Result) = 0) then
      Result := Result + '@';
  End;

Function ForwardPathFromPath (const Path : String) : String;
  Begin
    Result := Trim (CopyBefore (CopyBetween (Path, '<', '>', True, True), ':', False));
  End;

Function ForwardHostFromPath (const Path : String) : String;
  Begin
    Result := Trim (CopyBefore (ForwardPathFromPath (Path), ',', True));
  End;

Function AddHostToReversePath (const ReversePath, Host : String) : String;
var R : String;
  Begin
    R := Trim (CopyBetween (ReversePath, '<', '>', True, True));
    if PosNext (':', R) = 0 then
      Result := Host + ':' + ReversePath else
      Result := Host + Cond (R <> '', ',' + R, '');
    Result := '<' + Result + '>';
  End;

Procedure TsmtpServer.Init;
  Begin
    inherited Init;
    FServerReady := True;
    FClientIdentityRequired := True;
    FRecipientRequired := True;
    FMaxMessageSize := 0;
    FEnableAuthExt := True;
    FEnablePipeliningExt := True;
    FEnableDSNExt := True;
  End;

Constructor TsmtpServer.CreateEx (const AOwner : TComponent; const TimeOut : Integer);
  Begin
    inherited CreateEx (AOwner, DefaultSMTPPort, TimeOut, TsmtpServerClientThread);
  End;

Function SMTPServerResponse (const StatusCode : Integer; const Msg : String) : String;
  Begin
    Assert ((StatusCode >= 100) and (StatusCode <= 999), 'Invalid SMTP StatusCode');
    Result := IntToStr (StatusCode) +
              Cond ((Msg <> '') and (Msg [1] = '-'), '', ' ') +
              Cond (Msg <> '', Msg, SMTPMessage (StatusCode));
  End;

Destructor TsmtpServerClientThread.Destroy;
  Begin
    FreeAndNil (FRecipients);
    inherited Destroy;
  End;

Function TsmtpServerClientThread.RemoteHostMatch (const Mask : String) : Boolean;
  Begin
    Result := MatchFileMask (Mask, RemoteHost) or
       MatchFileMask (Mask, RemoteAddress) or
       MatchFileMask (Mask, RemoteSMTPHost);
  End;

Procedure TsmtpServerClientThread.NotifyHello;
  Begin
    if Assigned (TsmtpServer (FServer).FOnHello) then
      TsmtpServer (FServer).FOnHello (self, FAuthRequired, FAllowAccess);
  End;

Procedure TsmtpServerClientThread.NotifyRecipient;
  Begin
    if Assigned (TsmtpServer (FServer).FOnRecipient) then
      TsmtpServer (FServer).FOnRecipient (self, FRecipient, FResponseCode, FResponseMsg);
  End;

Procedure TsmtpServerClientThread.NotifyReversePath;
  Begin
    if Assigned (TsmtpServer (FServer).FOnReversePath) then
      TsmtpServer (FServer).FOnReversePath (self, FResponseCode, FResponseMsg);
  End;

Procedure TsmtpServerClientThread.NotifyVerify;
  Begin
    if Assigned (TsmtpServer (FServer).FOnVerify) then
      TsmtpServer (FServer).FOnVerify (self, FVerifyName, FResponseCode, FResponseMsg);
  End;

Procedure TsmtpServerClientThread.NotifyExpand;
  Begin
    if Assigned (TsmtpServer (FServer).FOnExpand) then
      TsmtpServer (FServer).FOnExpand (self, FVerifyName, FResponseList, FResponseCode, FResponseMsg);
  End;

Procedure TsmtpServerClientThread.NotifyMessage;
  Begin
    if Assigned (TsmtpServer (FServer).FOnMessage) then
      TsmtpServer (FServer).FOnMessage (self, FResponseCode, FResponseMsg);
  End;

Procedure TsmtpServerClientThread.NotifyAuthenticate;
  Begin
    if Assigned (TsmtpServer (FServer).FOnAuthenticate) then
      TsmtpServer (FServer).FOnAuthenticate (self, FAuthMech, FUserName, FPassword, FAuthAction);
  End;

const
  HelpTextLines = 24;
  HelpText : Array [1..HelpTextLines] of String = (
      '-The following are the SMTP commands:',
      '- ',
      '-  HELO <SP> <domain> <CRLF>',
      '-  MAIL <SP> FROM:<reverse-path> <CRLF>',
      '-  RCPT <SP> TO:<forward-path> <CRLF>',
      '-  DATA <CRLF>',
      '-  RSET <CRLF>',
      '-  SEND <SP> FROM:<reverse-path> <CRLF>',
      '-  SOML <SP> FROM:<reverse-path> <CRLF>',
      '-  SAML <SP> FROM:<reverse-path> <CRLF>',
      '-  VRFY <SP> <string> <CRLF>',
      '-  EXPN <SP> <string> <CRLF>',
      '-  HELP [<SP> <string>] <CRLF>',
      '-  NOOP <CRLF>',
      '-  QUIT <CRLF>',
      '-  TURN <CRLF>',
      '- ',
      '-The following extentions may be enabled:',
      '- ',
      '-  DSN         - Delivery Status Notification',
      '-  SIZE        - Size Declaration',
      '-  PIPELINING  - Pipelining',
      '-  AUTH        - Authentication',
      ' ');

Procedure TsmtpServerClientThread.ClientServe;
var SMTPServer : TsmtpServer;

  Function ServiceNotReadyMsg (const BusyMsg : String) : String;
    Begin
      Result := FLocalSMTPHost +
                ' Service Not Ready' +
                Cond (BusyMsg <> '', ' (' + BusyMsg + ')', '');
    End;

  Procedure SendResponse (const StatusCode : Integer; const Msg : String = '');
  var S : String;
    Begin
      if Terminated or not SMTPServer.Active or not SMTPServer.ServerReady then
        begin
          try
            S := SMTPServerResponse (SMTP_ServiceNotAvailable,
                             ServiceNotReadyMsg (SMTPServer.ServerBusyMsg));
            {$IFDEF DEBUGLOG}
            Log (lt_Debug, 'S: ' + S);
            {$ENDIF}
            FTClient.Writeln (S);
          except end;
          raise ESMTPServer.Create ('Session terminated.');
        end;

      S := SMTPServerResponse (StatusCode, Msg);
      {$IFDEF DEBUGLOG}
      Log (lt_Debug, 'S: ' + S);
      {$ENDIF}
      FTClient.Writeln (S);
    End;

  Procedure SendGreeting;
    Begin
      SendResponse (SMTP_ServiceReady,
                FLocalSMTPHost + ' Simple Mail Transfer Service Ready (' + SMTPServerVersion + ')');
    End;

  Procedure SendRSETResponse;
    Begin
      SendResponse (SMTP_OK, 'OK');
    End;

  Procedure SendUnknownCommandResponse;
    Begin
      SendResponse (SMTP_Error, 'Unknown command specified.');
    End;

  Function HELOInfo : String;
    Begin
      Result := FLocalSMTPHost + ' OK ' + FRemoteSMTPHost + ' [' + RemoteAddress + '].';
    End;

  Procedure SendHELOResponse;
    Begin
      SendResponse (SMTP_OK, HELOInfo);
    End;

  Procedure SendEHLOResponse;
    Begin
      SendResponse (SMTP_OK, '-' + HELOInfo);
      SendResponse (SMTP_OK, '-EXPN');
      SendResponse (SMTP_OK, '-SEND');
      SendResponse (SMTP_OK, '-SOML');
      SendResponse (SMTP_OK, '-SAML');
      if SMTPServer.EnableDSNExtention then
        SendResponse (SMTP_OK, '-DSN');
      if SMTPServer.EnableAuthenticationExtention then
        SendResponse (SMTP_OK, '-AUTH PLAIN ANONYMOUS CRAM-MD5');
      if SMTPServer.MaxMessageSize > 0 then
        SendResponse (SMTP_OK, '-SIZE ' + IntToStr (SMTPServer.MaxMessageSize));
      if SMTPServer.EnablePipeliningExtention then
        SendResponse (SMTP_OK, '-PIPELINING');
      SendResponse (SMTP_OK, 'HELP');
    End;

  Function CheckAuthenticated : Boolean;
    Begin
      Result := not FAuthRequired or FAuthenticated;
      if not Result then
        SendResponse (SMTP_AuthenticationRequired);
    End;

  Procedure SplitParameter (const Param : String; var Key, Value : String);
    Begin
      Key := UpperCase (Trim (CopyBefore (Param, [':', '='], True)));
      Value := Trim (CopyAfter (Param, [':', '='], False));
    End;

  Function DecodeAuthParam (const Param : String; var Mechanism, InitialResponse : String) : Boolean;
    Begin
      Mechanism := UpperCase (CopyBefore (Param, ' ', True));
      try
        InitialResponse := MIMEBase64Decode (CopyAfter (Param, ' ', False));
        Result := True;
      except
        Result := False;
      end;
    End;

  Procedure GetAuthentication;
    Begin
      FAuthAction := auth_Failed;
      if Assigned (SMTPServer.OnAuthenticate) then
        Synchronize (NotifyAuthenticate);
    End;

  Procedure SplitCommandParam (const Param : String; var List : StringArray);
    Begin
      List := Split (RemoveDup (' ', Replace (': ', ':', Param)), ' ');
    End;

  Function GetCommand (var Cmd, Param : String) : Boolean;
  var R : Boolean;
      I : Integer;
    Begin
      Cmd := '';
      Param := '';
      Repeat
        Repeat
          Result := ReadLine (Cmd);
          if not Result then
            exit;
        Until Cmd <> '';
        {$IFDEF DEBUGLOG}
        Log (lt_Debug, 'C: ' + Cmd);
        {$ENDIF}
        Param := Trim (CopyAfter (Cmd, ' '));
        Cmd := UpperCase (Trim (CopyBefore (Cmd, ' ')));
        R := False;
        if Cmd = 'NOOP' then
          SendResponse (SMTP_OK, 'Did Nothing.') else
        if Cmd = 'HELP' then
          begin
            For I := 1 to HelpTextLines do
              SendResponse (SMTP_HelpMessage, HelpText [I]);
          end else
        if Cmd = 'VRFY' then
          begin
            if not Assigned (SMTPServer.FOnVerify) then
              SendResponse (SMTP_PermanentFailure, 'VRFY command has been disabled.') else
              begin
                FVerifyName := Param;
                FResponseCode := SMTP_PermanentFailure;
                FResponseMsg := 'String does not match anything.';
                Synchronize (NotifyVerify);
                SendResponse (FResponseCode, FResponseMsg);
              end;
          end else
        if Cmd = 'EXPN' then
          begin
            if not Assigned (SMTPServer.FOnExpand) then
              SendResponse (SMTP_PermanentFailure, 'EXPN command has been disabled.') else
              begin
                FVerifyName := Param;
                FResponseCode := SMTP_PermanentFailure;
                FResponseMsg := 'String does not match anything.';
                FResponseList := nil;
                Synchronize (NotifyExpand);
                if IsResponseClass (FResponseCode, rc_PositiveCompletion) and
                   Assigned (FResponseList) then
                  begin
                    For I := 0 to Length (FResponseList) - 1 do
                      SendResponse (FResponseCode, Cond (I < Length (FResponseList) - 1, '-', '') +
                                  FResponseList [I]);
                  end else
                  SendResponse (FResponseCode, FResponseMsg);
              end;
          end else
        if (Cmd = 'TURN') or (Cmd = 'ETRN') or (Cmd = 'BPUT') then
          SendResponse (SMTP_CommandNotImplemented) else
          R := True;
      Until R;

      if Cmd = 'QUIT' then
        begin
          Log (lt_ClientState, Cond (FRemoteHost <> '', FRemoteSMTPHost + ' ', '') +
                               'Session closed.');
          SendResponse (SMTP_ServiceClosingChannel, FLocalSMTPHost + ' Bye');
          Result := False;
        end;
    End;

var Cmd, Param    : String;
    I, F          : Integer;
    R             : Boolean;
    S, T, U       : String;
    Hdr           : String;
    MsgFin        : Boolean;
    MsgOK         : Boolean;
    N             : TDateTime;
    L, M          : StringArray;
    UsingESMTP    : Boolean;
    J             : Int64;
    Rcpt          : TsmtpRecipient;
    HF            : AHeaderField;
  Begin
    // Initialize
    UsingESMTP := False;
    FTClient.EOLType := [eol_CRLF, eol_EOF];
    FTClient.MaxLineLength := 4096;
    SMTPServer := FServer as TsmtpServer;

    if SMTPServer.HostName = '' then
      FLocalSMTPHost := SMTPServer.LocalHost else
      FLocalSMTPHost := SMTPServer.HostName;
    FDeliveryVerb := '';
    FReversePath  := '';
    FMsgBody := '';
    FRawMsg := '';
    FAuthenticated := False;
    FRecipients := TsmtpRecipientList.Create;

    // System Status
    SendGreeting;

    try
      // Get HELO
      Repeat
        if not GetCommand (Cmd, Param) then
          exit;
        R := (Cmd = 'HELO') or (Cmd = 'EHLO');
        if not R then
          begin
            if (Cmd = 'RCPT') or (Cmd = 'MAIL') or (Cmd = 'SEND') or (Cmd = 'SOML') or
               (Cmd = 'SAML') or (Cmd = 'DATA') or (Cmd = 'AUTH') then
              SendResponse (SMTP_BadSequenceOfCommands, 'No HELO command has been given.') else
            if Cmd = 'RSET' then
              SendRSETResponse else
              SendUnknownCommandResponse;
          end else
          begin
            UsingESMTP := Cmd = 'EHLO';
            R := not SMTPServer.ClientIdentityRequired or (Param <> '');
            if not R then
              SendResponse (SMTP_Error, 'Host name required.');
            FRemoteSMTPHost := Param;
            if FRemoteSMTPHost = '' then
              FRemoteSMTPHost := 'unknown';

            if R then
              begin
                FAuthRequired := False;
                FAllowAccess := True;
                if Assigned (SMTPServer.OnHello) then
                  Synchronize (NotifyHello);
                R := FAllowAccess;
                if not R then
                  SendResponse (SMTP_ServiceNotAvailable, ServiceNotReadyMsg ('Access denied'));
              end;
          end;
      Until R;

      Log (lt_ClientState, FRemoteSMTPHost +
                      Cond (FRemoteSMTPHost <> FRemoteHost, ' (' + FRemoteHost + ')', '') +
                      ' Connected.');

      if UsingESMTP then
        SendEHLOResponse else
        SendHELOResponse;

      Repeat
        // Message
        FDeliveryVerb := '';
        FReversePath := '';
        FRecipients.Clear;
        MsgFin := False;
        MsgOK := False;
        Repeat
          if not GetCommand (Cmd, Param) then
            exit;

          if (Cmd = 'MAIL') or (Cmd = 'SEND') or (Cmd = 'SOML') or (Cmd = 'SAML') then
            begin
              FDeliveryVerb := Cmd;
              FReversePath := '';
              FDSNReturn := '';
              FDSNEnvID := '';
              SplitCommandParam (Param, L);
              R := True;
              For I := 0 to Length (L) - 1 do
                begin
                  SplitParameter (L [I], S, T);
                  if S = 'FROM' then
                    FReversePath := T else
                  if S = 'SIZE' then
                    begin
                      J := StrToInt64Def (T, -1);
                      if J = -1 then
                        begin
                          R := False;
                          SendResponse (SMTP_ParamSyntaxError, 'Invalid value to SIZE parameter in MAIL command.');
                          break;
                        end;
                      if (SMTPServer.MaxMessageSize > 0) and
                         (J > SMTPServer.MaxMessageSize) then
                        begin
                          R := False;
                          SendResponse (SMTP_AbortInsufficientStorage, 'Declared size exceeds fixed maximum message size.');
                          break;
                        end;
                    end else
                  if S = 'RET' then
                    begin
                      T := UpperCase (T);
                      if (T <> smtp_dsnret_HDRS) and (T <> smtp_dsnret_FULL) then
                        begin
                          R := False;
                          SendResponse (SMTP_ParamSyntaxError, 'Unknown value for RET parameter in MAIL command.');
                          break;
                        end;
                      FDSNReturn := T;
                    end else
                  if S = 'ENVID' then
                    begin
                      if T = '' then
                        begin
                          R := False;
                          SendResponse (SMTP_ParamSyntaxError, 'Bad value for ENVID parameter in MAIL command.');
                          break;
                        end;
                      FDSNEnvID := DecodeXText (T);
                    end else
                  if S <> 'AUTH' then
                    begin
                      R := False;
                      SendResponse (SMTP_UnknownParameter);
                      break;
                    end;
                end;
              if R and CheckAuthenticated then
                begin
                  FResponseCode := SMTP_OK;
                  FResponseMsg := 'OK' +
                                  Cond (FReversePath = '', ' (No reverse path specified)', '');
                  if Assigned (SMTPServer.FOnReversePath) then
                    Synchronize (NotifyReversePath);
                  SendResponse (FResponseCode, FResponseMsg);
                  S := 'Mail from ' + FReversePath;
                  if not IsResponseClass (FResponseCode, rc_PositiveCompletion) then
                    begin
                      FDeliveryVerb := '';
                      FReversePath := '';
                      S := S + ' (rejected)';
                    end;
                  Log (lt_ClientState, FRemoteSMTPHost + ' ' + S);
                end;
            end else
          if Cmd = 'RCPT' then
            begin
              FRecipient := '';
              FRecipientNotify := [];
              FDSNOrigRcpt := '';
              SplitCommandParam (Param, L);
              R := True;
              For I := 0 to Length (L) - 1 do
                begin
                  SplitParameter (L [I], S, T);
                  if S = 'TO' then
                    FRecipient := T else
                  if S = 'NOTIFY' then
                    try
                      FRecipientNotify := StringAsDSNNotify (T);
                    except
                      R := False;
                      SendResponse (SMTP_ParamSyntaxError, 'Bad value for NOTIFY parameter in RCPT command.');
                      break;
                    end else
                  if S = 'ORCPT' then
                    FDSNOrigRcpt := DecodeXText (T) else
                    begin
                      R := False;
                      SendResponse (SMTP_UnknownParameter);
                      break;
                    end;
                end;

              if R then
                if FRecipient = '' then
                  SendResponse (SMTP_ParamSyntaxError, 'Bad argument in RCPT command.') else
                  if CheckAuthenticated then
                    begin
                      FResponseCode := SMTP_OK;
                      FResponseMsg := MailboxFromPath (FRecipient) + ' OK';
                      if Assigned (SMTPServer.FOnRecipient) then
                        Synchronize (NotifyRecipient);
                      S := 'Recipient: ' + FRecipient;
                      SendResponse (FResponseCode, FResponseMsg);
                      if IsResponseClass (FResponseCode, rc_PositiveCompletion) then
                        begin
                          Rcpt := TsmtpRecipient.CreateEx (FRecipient, FDSNOrigRcpt, FRecipientNotify);
                          FRecipients.Add (Rcpt);
                        end else
                        S := S + ' (rejected)';
                      Log (lt_ClientState, FRemoteSMTPHost + ' ' + S);
                    end;
            end else
          if Cmd = 'DATA' then
            begin
              if FDeliveryVerb = '' then
                SendResponse (SMTP_BadSequenceOfCommands, 'MAIL command required before DATA command.') else
              if SMTPServer.RecipientRequired and (FRecipients.Count = 0) then
                SendResponse (SMTP_Error, 'No recipient for message.') else
                if CheckAuthenticated then
                  begin
                    SendResponse (SMTP_StartMail, '');
                    FRawMsg := ExtractDotLineTerminated (FClient, [eol_CRLF, eol_EOF]);
                    if (SMTPServer.MaxMessageSize > 0) and
                       (Length (FRawMsg) > SMTPServer.MaxMessageSize + 16384) then
                      begin
                        SendResponse (SMTP_AbortInsufficientStorage, 'Actual size exceeds fixed maximum message size.');
                        Log (lt_ClientState, FRemoteSMTPHost + ' Rejected large message (' + StorageSize (Length (FRawMsg)) + ')');
                      end else
                      begin
                        MsgFin := True;
                        MsgOK := True;
                      end;
                  end;
            end else
          if Cmd = 'RSET' then
            begin
              MsgFin := True;
              SendRSETResponse;
            end else
          if (Cmd = 'HELO') or (Cmd = 'EHLO') then
            SendResponse (SMTP_BadSequenceOfCommands, 'HELO command already issued.') else
          if Cmd = 'AUTH' then
            begin
              if FAuthenticated then
                SendResponse (SMTP_BadSequenceOfCommands, 'Already authenticated.') else
              if not FAuthRequired then
                SendResponse (SMTP_AuthenticationNotRequired, 'Authentication not required.') else
              if not DecodeAuthParam (Param, S, T) then
                SendResponse (SMTP_ParamSyntaxError, 'Bad argument to AUTH command.') else
              if (S <> 'PLAIN') and (S <> 'ANONYMOUS') and (S <> 'CRAM-MD5') then
                SendResponse (SMTP_ParamNotImplemented, 'Authentication method unknown.') else
                begin
                  R := True;
                  if S = 'PLAIN' then
                    begin
                      T := Trim (T, [#0..#32]);
                      FUserName := CopyBefore (T, #0, False);
                      T := CopyAfter (T, #0, True);
                      FAuthMech := auth_Plain;
                      FPassword := T;
                      GetAuthentication;
                    end else
                  if S = 'ANONYMOUS' then
                    begin
                      FUserName := 'anonymous';
                      FAuthMech := auth_Anonymous;
                      FPassword := T;
                      GetAuthentication;
                    end else
                  if S = 'CRAM-MD5' then
                    begin
                      U := MIMEBase64Encode (MessageIDFieldBody);
                      SendResponse (SMTP_AuthenticationChallenge, U);
                      if not ReadLine (T) then
                        exit;
                      if T = '*' then
                        begin
                          SendResponse (SMTP_ParamSyntaxError, 'Authentication aborted.');
                          R := False;
                        end else
                        begin
                          try
                            T := MIMEBase64Decode (T);
                          except
                            SendResponse (SMTP_ParamSyntaxError, 'Authentication response invalid.');
                            R := False;
                          end;
                          if R then
                            begin
                              FUserName := CopyBefore (T, ' ', True);
                              T := CopyAfter (T, ' ', False);
                              FPassword := '';
                              GetAuthentication;
                              if FAuthAction = auth_Successful then
                                begin
                                  S := MD5KeyAsString (CalcKeyedMD5 (FPassword, S));
                                  if S <> T then
                                    FAuthAction := auth_Failed;
                                end;
                            end;
                        end;
                    end;

                  if R then
                    Case FAuthAction of
                      auth_TooWeak    : SendResponse (SMTP_MechanismTooWeak);
                      auth_Failed     : SendResponse (SMTP_AuthenticationFailed);
                      auth_Successful : SendResponse (SMTP_AuthenticationSuccessful);
                    end;
                  FAuthenticated := FAuthAction = auth_Successful;
                end;
            end else
            SendUnknownCommandResponse;
        Until MsgFin;

        if MsgOK then
          begin
            Log (lt_ClientState, FRemoteSMTPHost + ' Message received. (' + IntToStr (Length (FRawMsg)) + ' bytes)');

            FMsgBody := CopyAfter (FRawMsg, c_CRLF + c_CRLF, True);
            Hdr := CopyBefore (FRawMsg, c_CRLF + c_CRLF, False);
            FMsgHeader := THeader.Create;
            try
              FMsgHeader.AsString := Hdr;
              N := LocalTimeToGMTTime (Now);

              // Check for loops
              For I := 0 to FMsgHeader.FieldCount - 1 do
                begin
                  HF := FMsgHeader.FieldByIndex [I];
                  if LowerCase (HF.Name) = 'received' then
                    begin
                      S := LowerCase (HF.Body);
                      if Trim (CopyBetween (S, 'by ', [' ', ';', '('], False, True)) = LowerCase (LocalSMTPHost) then
                        begin
                          FResponseCode := SMTP_PermanentFailure;
                          FResponseMsg := 'This message has been relayed by this server before, possible mail loop.';
                          MsgOK := False;
                          break;
                        end;
                    end;
                end;

              if MsgOK then
                if not Assigned (SMTPServer.FOnMessage) then
                  begin
                    FResponseCode := SMTP_PermanentFailure;
                    FResponseMsg := 'No message handler.';
                    MsgOK := False;
                  end;

              if MsgOK then
                begin
                  // Update message header
                  if FReversePath <> '' then
                    FMsgHeader.FieldBody ['Return-Path'] := FReversePath;
                  S := MailboxFromPath (FReversePath);
                  if not FMsgHeader.HasField ('From') then
                    FMsgHeader.AddField ('From', S);
                  if (FRecipients.Count > 0) and not FMsgHeader.HasField ('To') then
                    FMsgHeader.AddField ('To', FRecipients.GetToFieldBody);
                  U := Trim (FMsgHeader.FieldBody ['Message-id']);
                  if U = '' then
                    FMsgHeader.AddField ('Message-id', MessageIDFieldBody ('', FLocalSMTPHost));
                  if not FMsgHeader.HasField ('Date') then
                    begin
                      FMsgHeader.AddField ('Date', RFCDateTime (N));
                      FMsgHeader.AddField ('Date-warning', 'Date header was inserted by ' + FLocalSMTPHost);
                    end;
                  S := FRemoteSMTPHost + ' (';
                  if RemoteHost <> RemoteAddress then
                    S := S + RemoteHost + ' ';
                  S := S + '[' + RemoteAddress + '])';
                  if FRecipients.Count <> 1 then
                    T := '' else
                    T := MailboxFromPath (FRecipients.Items [0].Recipient);
                  FMsgHeader.AddField ('Received',
                      'from ' + S + ' ' +
                      'by ' + FLocalSMTPHost + ' (' + SMTPServerVersion + ') ' +
                      Cond (T <> '', 'for ' + T + ' ', '') +
                      'with ' + Cond (UsingESMTP, 'E', '') + 'SMTP' + Cond (U <> '', ' id ' + U, '') +
                      '; ' + FormatDateTime ('d mmm yyyy hh:nn:ss', N) + ' -0000');
                end;

              if MsgOK then
                begin
                  FResponseCode := SMTP_OK;
                  FResponseMsg := 'OK, ' + IntToStr (Length (FRawMsg)) + ' bytes received.';
                  Synchronize (NotifyMessage);
                end;

              // Respond
              SendResponse (FResponseCode, FResponseMsg);
            finally
              FreeAndNil (FMsgHeader);
            end;
          end;
      Until False;
    except
      on E : Exception do
        begin
          SendResponse (SMTP_ServiceNotAvailable, 'Internal server error: ' + ExceptionToStr (E));
          raise;
        end;
    end;
  End;

Function DeliveryNotificationStatusMessage (const NotificationType : TsmtpDSNNotification;
         const Hdr : AHeader; const ServerDomain, FailureReason : String;
         const Recipient : TsmtpRecipient) : AMessage;
var S, M, T : String;
    R       : TMimeEmailMessage;
    H       : TMimeEmailMessageHeader;
    MM, SM  : TMimeMessage;
    MH      : TMimeMessageHeader;
  Begin
    S := Hdr.FieldBody ['Return-Path'];
    if S = '' then
      DecomposeEMailField (Hdr.FieldBody ['From'], S, T);
    M := MailboxFromPath (S);
    if M = '' then
      begin
        Result := nil;
        exit;
      end;
    R := TMimeEmailMessage.Create;
    H := R.Header as TMimeEmailMessageHeader;
    H.AddField ('Return-Path', '<>');
    H.AddField ('Message-ID', MessageIDFieldBody ('', ServerDomain));
    H.AddField ('To', '<' + M + '>');
    H.AddField ('From', '<postmaster@' + ServerDomain + '>');
    Case NotificationType of
      dsn_success : S := 'Success';
      dsn_failure : S := 'Failure';
      dsn_delay   : S := 'Delay';
    end;
    H.AddField ('Subject', 'Delivery Notification (' + S + ') for ' + M);
    H.ContentTypeField.ContentType := 'multipart/report';
    H.ContentTypeField.Parameter ['report-type'] := 'delivery-status';
    H.ContentTypeField.GenerateBoundary;

    MM := TMimeMessage.Create;
    R.AddBodyPart (MM);
    MH := MM.Header as TMimeMessageHeader;
    MH.ContentTypeField.ContentType := 'text/plain';
    MH.ContentTypeField.Parameter ['charset'] := 'us-ascii';
    T := Recipient.Recipient;
    S := 'Your message, addressed to ' + T + ', ';
    Case NotificationType of
      dsn_Success : S := S + 'was successfully delivered.';
      dsn_failure : S := S + 'could not be delivered, for the following reason:' + CRLF +
                             FailureReason;
      dsn_delay   : S := S + 'have been delayed.' + CRLF +
                             FailureReason;
    end;
    S := S + CRLF + CRLF;
    MM.AddBodyPart (TTextBodyPart.CreateEx (S));

    MM := TMimeMessage.Create;
    R.AddBodyPart (MM);
    MH := MM.Header as TMimeMessageHeader;
    MH.ContentTypeField.ContentType := 'message/delivery-status';
    SM := TMimeMessage.Create;
    MM.AddBodyPart (SM);
    MH := TMimeMessageHeader.Create;

    SM.AddBodyPart (MH);
    if PosNext ('.', ServerDomain) = 0 then
      SM.Header.FieldBody ['Reporting-MTA'] := 'x-local-hostname; ' + ServerDomain else
      SM.Header.FieldBody ['Reporting-MTA'] := 'dns; ' + ServerDomain;
    if Recipient.dsnOrigRcpt <> '' then
      MH.FieldBody ['Original-Recipient'] := 'rfc822;' + Recipient.dsnOrigRcpt;
    MH.FieldBody ['Final-Recipient'] := 'rfc822;' + T;
    MH.FieldBody ['Action'] := Cond (NotificationType = dsn_Success, 'relayed', 'failed');
    Case NotificationType of
      dsn_success : S := '?.?.?';
      dsn_failure : S := '5.0.0';
      dsn_delay   : S := '4.0.0';
    end;
    MH.FieldBody ['Status'] := S;

    MM := TMimeMessage.Create;
    R.AddBodyPart (MM);
    MH := MM.Header as TMimeMessageHeader;
    MH.ContentTypeField.ContentType := 'message/rfc822';
    SM := TMimeMessage.CreateEx (Hdr.Duplicate as THeader, []);
    MM.AddBodyPart (SM);

    R.Prepare;
    Result := R;
  End;

end.

