{$INCLUDE cHeader.inc}
unit cXML;

{                                                                              }
{                            XML Parser v0.01 (L0)                             }
{                             (Under development)                              }
{                                                                              }
{        This unit is copyright  2000 by David Butler (david@e.co.za)         }
{                                                                              }
{                  This unit is part of Delphi Fundamentals.                   }
{                      Its original file name is cXML.pas                      }
{                                                                              }
{                I invite you to use this unit, free of charge.                }
{        I invite you to distibute this unit, but it must be for free.         }
{             I also invite you to contribute to its development,              }
{             but do not distribute a modified copy of this file.              }
{       Send modifications, suggestions and bug reports to david@e.co.za       }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   11/05/2000  0.01  Moved from cInternetStandards                            }
{                                                                              }

interface

uses
  SysUtils,

  // Delphi Fundamentals (L0)
  cStreams;




{                                                                              }
{ XML Representations                                                          }
{                                                                              }
type
  { AxmlComponent                                                              }
  AxmlComponent = class
    Function GetAsString : String; virtual; abstract;
  end;

  { TxmlComponentList                                                          }
  TxmlComponentList = class (AxmlComponent)
    protected
    FComponents : Array of AxmlComponent;

    Function GetComponent (const Idx : Integer) : AxmlComponent;

    public
    Constructor Create (const First, Second : AxmlComponent); reintroduce; overload;
    Destructor Destroy; override;

    Function  GetAsString : String; override;

    Procedure Add (const C : AxmlComponent);
    Function  Count : Integer;
    Property  Component [const Idx : Integer] : AxmlComponent read GetComponent;
  end;

  { AxmlTag                                                                    }
  AxmlTag = class (AxmlComponent)
    protected
    Function GetName : String; virtual; abstract;

    public
    Property Name : String read GetName;
    Function AttrAsString (const Attr : String; const DefaultValue : String = '') : String; virtual; abstract;
    Function AttrAsInteger (const Attr : String; const DefaultValue : Integer = -1) : Integer; virtual; abstract;
  end;
  TxmlTagAttribute = record
    Attr  : String;
    Value : String;
  end;
  TxmlTagAttributes = class
    protected
    FAttr : Array of TxmlTagAttribute;

    Function IndexOf (const Attr : String) : Integer;

    public
    Procedure AddAttribute (const Attr : String; const Value : String = '');
    Function  AsString : String;
    Function  AttrAsString (const Attr : String; const DefaultValue : String = '') : String;
    Function  AttrAsInteger (const Attr : String; const DefaultValue : Integer = -1) : Integer;
    Function  Duplicate : TxmlTagAttributes;
  end;
  TxmlTag = class (AxmlTag)
    protected
    FName : String;
    FAttr : TxmlTagAttributes;

    Function  GetName : String; override;

    public
    Constructor Create (const Name : String; const Attributes : TxmlTagAttributes);
    Destructor Destroy; override;

    Function AttrAsString (const Attr : String; const DefaultValue : String = '') : String; override;
    Function AttrAsInteger (const Attr : String; const DefaultValue : Integer = -1) : Integer; override;
  end;
  TxmlOpenTag = class (TxmlTag)
    Function GetAsString : String; override;
  end;
  TxmlCloseTag = class (TxmlTag)
    Function GetAsString : String; override;
  end;
  TxmlEmptyTag = class (TxmlTag)
    Function GetAsString : String; override;
  end;

  { AxmlElement                                                                }
  AxmlElement = class (AxmlComponent);
  AxmlTagElement = class (AxmlElement)
    protected
    Function GetContent : AxmlComponent; virtual; abstract;

    public
    Property Content : AxmlComponent read GetContent;
  end;
  TxmlTagElement = class (AxmlTagElement)
    protected
    FOpenTag,
    FCloseTag : AxmlTag;
    FContent  : AxmlComponent;

    Function GetContent : AxmlComponent; override;

    public
    Constructor Create (const OpenTag, CloseTag : AxmlTag;
                        const Content : AxmlComponent);
    Destructor Destroy; override;

    Function  GetAsString : String; override;
  end;
{  TxmlEmptyElementTag = class (AxmlElement)
    protected
    FEmptyTag : AxmlTag;

    public
    Constructor Create (const EmptyTag : AxmlTag);
    Destructor Destroy; override;

    Function GetAsString : String; override;
  end; }

  { TxmlCharacterEntity                                                        }
  TxmlCharacterEntity = class (AxmlComponent)
    protected
    FEntity : String;

    public
    Constructor Create (const Entity : String);
    Function GetAsString : String; override;
  end;

  { TxmlComment                                                                }
  TxmlComment = class (AxmlComponent)
    protected
    FComment : String;

    public
    Constructor Create (const Comment : String);
    Function GetAsString : String; override;
  end;

  { TxmlDTDTag (Document Type Declaration)                                     }
  TxmlDTDTag = class (AxmlComponent)
    protected
    FText : String;

    public
    Constructor Create (const Text : String);
    Function GetAsString : String; override;
  end;

  { TxmlXMLTag                                                                 }
  TxmlXMLTag = class (AxmlComponent)
    protected
    FTag : String;

    public
    Constructor Create (const Tag : String);
    Function GetAsString : String; override;
  end;

  { TxmlContent                                                                }
  TxmlContent = class (AxmlComponent)
    protected
    FText : String;

    public
    Constructor Create (const Text : String);
    Function GetAsString : String; override;
  end;



{                                                                              }
{ XML Decoding                                                                 }
{                                                                              }
type
  TxmlDocument = class
    FContent : TxmlComponentList;

    Constructor Create;
    Destructor Destroy; override;
  end;

  ExmlParser = class (Exception);
  TxmlParser = class (TTextParser)
    protected
    FToken  : AxmlComponent;

    Function InterpretText (const Text : String) : AxmlComponent; virtual;
    Function InterpretCharacterEntity (const CharacterEntity : String) : AxmlComponent; virtual;
    Function InterpretTag (const Tag : AxmlTag) : AxmlComponent; overload; virtual;
    Function InterpretTag (const OpenTag, CloseTag : AxmlTag; const Content : AxmlComponent) : AxmlComponent; overload; virtual;

    Function GetNextToken : AxmlComponent; virtual;

    public
    Constructor Create (const S : TExStream);

    Function ExtractFlatXML : AxmlComponent;
    Function ExtractXMLDocument : TxmlDocument;
  end;


Function xmlSafeText (const Txt : String) : String;
Function xmlSafeQuotedText (const Txt : String) : String;


implementation

uses
  cUtils,
  cStrings;

const
  { XML "escape" encodings }
  xmlAmp            = '&amp;';        // &
  xmlLT             = '&lt;';         // <
  xmlGT             = '&gt;';         // >
  xmlQuote          = '&quot;';       // "

{ Ch > 127? }
Function xmlSafeText (const Txt : String) : String;
  Begin
    Result := Replace ('&', xmlAmp, Txt);
    Result := Replace ('<', xmlLT, Result);
    Result := Replace ('>', xmlGT, Result);
    Result := Replace ('"', xmlQuote, Result);
  End;

Function xmlSafeQuotedText (const Txt : String) : String;
  Begin
    Result := '"' + xmlSafeText (Txt) + '"';
  End;

{                                                                              }
{ XML Representations                                                          }
{                                                                              }

{ TxmlComponentList                                                           }
Constructor TxmlComponentList.Create (const First, Second : AxmlComponent);
  Begin
    inherited Create;
    SetLength (FComponents, 2);
    FComponents [0] := First;
    FComponents [1] := Second;
  End;

Destructor TxmlComponentList.Destroy;
  Begin
    FreeObjectArray (FComponents);
    inherited Destroy;
  End;

Procedure TxmlComponentList.Add (const C : AxmlComponent);
  Begin
    Append (ObjectArray (FComponents), C);
  End;

Function TxmlComponentList.Count : Integer;
  Begin
    Result := Length (FComponents);
  End;

Function TxmlComponentList.GetComponent (const Idx : Integer) : AxmlComponent;
  Begin
    Result := FComponents [Idx];
  End;

Function TxmlComponentList.GetAsString : String;
var I : Integer;
  Begin
    Result := '';
    For I := 0 to Length (FComponents) - 1 do
      Result := Result + FComponents [I].GetAsString;
  End;

Procedure AppendComponents (var R : AxmlComponent; const C : AxmlComponent);
  Begin
    if Assigned (C) then
      if not Assigned (R) then
        R := C else
        if not (R is TxmlComponentList) then
          R := TxmlComponentList.Create (R, C) else
          TxmlComponentList (R).Add (C);
  End;

{ TxmlContent                                                                  }
Constructor TxmlContent.Create (const Text : String);
  Begin
    inherited Create;
    FText := Text;
  End;

Function TxmlContent.GetAsString : String;
  Begin
    Result := FText;
  End;

{ TxmlCharacterEntity                                                          }
Constructor TxmlCharacterEntity.Create (const Entity : String);
  Begin
    inherited Create;
    FEntity := Entity;
  End;

Function TxmlCharacterEntity.GetAsString : String;
  Begin
    Result := '&' + FEntity;
  End;

{ TxmlComment                                                                  }
Constructor TxmlComment.Create (const Comment : String);
  Begin
    inherited Create;
    FComment := Comment;
  End;

Function TxmlComment.GetAsString : String;
  Begin
    Result := '<-- ' + FComment + ' -->';
  End;

{ TxmlDTDTag                                                                   }
Constructor TxmlDTDTag.Create (const Text : String);
  Begin
    inherited Create;
    FText := Text;
  End;

Function TxmlDTDTag.GetAsString : String;
  Begin
    Result := '<!' + FText + '>';
  End;

{ TxmlXMLTag                                                                   }
Constructor TxmlXMLTag.Create (const Tag : String);
  Begin
    inherited Create;
    FTag := Tag;
  End;

Function TxmlXMLTag.GetAsString : String;
  Begin
    Result := '<?' + FTag + '>';
  End;

{ TxmlTagAttributes                                                            }
Procedure TxmlTagAttributes.AddAttribute (const Attr, Value : String);
var L : Integer;
  Begin
    L := Length (FAttr);
    SetLength (FAttr, L + 1);
    FAttr [L].Attr := Attr;
    FAttr [L].Value := Value;
  End;

Function TxmlTagAttributes.AsString : String;
var I : Integer;
  Begin
    Result := '';
    For I := 0 to Length (FAttr) - 1 do
      Result := Result + Cond (Result <> '', ' ', '') +
                FAttr [I].Attr + '=' +
                xmlSafeQuotedText (FAttr [I].Value);
  End;

Function TxmlTagAttributes.IndexOf (const Attr : String) : Integer;
var I : Integer;
  Begin
    For I := 0 to Length (FAttr) - 1 do
      if FAttr [I].Attr = Attr then
        begin
          Result := I;
          exit;
        end;
    Result := -1;
  End;

Function TxmlTagAttributes.AttrAsString (const Attr : String; const DefaultValue : String) : String;
var I : Integer;
  Begin
    I := IndexOf (Attr);
    if I = -1 then
      Result := DefaultValue else
      Result := FAttr [I].Value;
  End;

Function TxmlTagAttributes.AttrAsInteger (const Attr : String; const DefaultValue : Integer) : Integer;
var I : Integer;
  Begin
    I := IndexOf (Attr);
    if I = -1 then
      Result := DefaultValue else
      try
        Result := StrToInt (FAttr [I].Value);
      except
        Result := DefaultValue;
      end;
  End;

Function TxmlTagAttributes.Duplicate : TxmlTagAttributes;
  Begin
    Result := TxmlTagAttributes.Create;
    Result.FAttr := FAttr;
  End;

{ TxmlTag                                                                      }
Constructor TxmlTag.Create (const Name : String; const Attributes : TxmlTagAttributes);
  Begin
    inherited Create;
    FName := Name;
    FAttr := Attributes;
  End;

Destructor TxmlTag.Destroy;
  Begin
    FreeAndNil (FAttr);
    inherited Destroy;
  End;

Function TxmlTag.GetName : String;
  Begin
    Result := FName;
  End;

Function TxmlTag.AttrAsString (const Attr : String; const DefaultValue : String) : String;
  Begin
    if Assigned (FAttr) then
      Result := FAttr.AttrAsString (Attr, DefaultValue) else
      Result := DefaultValue;
  End;

Function TxmlTag.AttrAsInteger (const Attr : String; const DefaultValue : Integer) : Integer;
  Begin
    if Assigned (FAttr) then
      Result := FAttr.AttrAsInteger (Attr, DefaultValue) else
      Result := DefaultValue;
  End;

Function TxmlOpenTag.GetAsString : String;
var A : String;
  Begin
    A := FAttr.AsString;
    Result := '<' + FName + Cond (A <> '', ' ' + A, '') + '>';
  End;

Function TxmlCloseTag.GetAsString : String;
var A : String;
  Begin
    A := FAttr.AsString;
    Result := '</' + FName + Cond (A <> '', ' ' + A, '') + '>';
  End;

Function TxmlEmptyTag.GetAsString : String;
var A : String;
  Begin
    A := FAttr.AsString;
    Result := '<' + FName + Cond (A <> '', ' ' + A + ' ', '') + '/>';
  End;

{ TxmlTagElement                                                               }
Constructor TxmlTagElement.Create (const OpenTag, CloseTag : AxmlTag; const Content : AxmlComponent);
  Begin
    inherited Create;
    FOpenTag := OpenTag;
    FContent := Content;
    FCloseTag := CloseTag;
  End;

Function TxmlTagElement.GetAsString : String;
  Begin
    Result := FOpenTag.GetAsString;
    if Assigned (FContent) then
      Result := Result + FContent.GetAsString;
    if Assigned (FCloseTag) then
      Result := Result + FCloseTag.GetAsString;
  End;

Function TxmlTagElement.GetContent : AxmlComponent;
  Begin
    Result := FContent;
  End;

Destructor TxmlTagElement.Destroy;
  Begin
    FreeAndNil (FCloseTag);
    FreeAndNil (FContent);
    FreeAndNil (FOpenTag);
    inherited Destroy;
  End;



{                                                                              }
{ TxmlParser                                                                   }
{                                                                              }
Constructor TxmlParser.Create (const S : TExStream);
  Begin
    inherited Create (S);
  End;

Function TxmlParser.InterpretTag (const Tag : AxmlTag) : AxmlComponent;
  Begin
    Result := Tag;
  End;

Function TxmlParser.InterpretTag (const OpenTag, CloseTag : AxmlTag; const Content : AxmlComponent) : AxmlComponent;
  Begin
    Result := TxmlTagElement.Create (OpenTag, CloseTag, Content);
  End;

Function TxmlParser.InterpretText (const Text : String) : AxmlComponent;
  Begin
    Result := TxmlContent.Create (Text);
  End;

Function TxmlParser.InterpretCharacterEntity (const CharacterEntity : String) : AxmlComponent;
  Begin
    Result := TxmlCharacterEntity.Create (CharacterEntity);
  End;

{ TxmlParser.GetNextToken                                                      }
{   Lexical tokeniser                                                          }
Function TxmlParser.GetNextToken : AxmlComponent;

  { "<!-- "<comment>" -->"                                                     }
  Function ExtractComment : AxmlComponent;
    Begin
      Result := TxmlComment.Create (ExtractTo (FStream, '-->', True));
    End;

  { "<!"<dtd tag>">"                                                           }
  Function ExtractDTD : AxmlComponent;
    Begin
      Result := TxmlDTDTag.Create (ExtractTo (FStream, '>', True));
    End;

  { "<?"<processing instructions>">"                                           }
  { "<?"<xml tag>"?>"                                                          }
  Function ExtractXMLTag : AxmlComponent;
    Begin
      Result := TxmlXMLTag.Create (ExtractTo (FStream, '>', True));
    End;

  // Pre:  S <> ''
  // Post: Returns '' if entity not recognised.
  Function DecodeEntity (const S : String) : String;
  var U : String;
    Begin
      if S [1] = '#' then
        try
          Result := Char (StrToInt (Replace ('x', '$', CopyFrom (S, 2))));
        except
          Result := '';
        end else
        begin
          U := LowerCase (S);
          if S = 'amp' then
            Result := '&' else
          if S = 'lt' then
            Result := '<' else
          if S = 'gt' then
            Result := '>' else
          if S = 'quote' then
            Result := '"' else
            Result := '';
        end;
    End;

  const
    cs_Entity = cs_Alpha;

  { "&"<entity>[";"]                                                           }
  { "&#"<entity code>[";"]                                                     }
  Function ExtractEntity : AxmlComponent;
  var T, R : String;
    Begin
      T := ExtractChar (FStream, ['#']);
      if T <> '' then
        T := T + ExtractChars (FStream, cs_Numeric) else
        T := ExtractChars (FStream, cs_Alpha);
      if T <> '' then
        begin
          ExtractChar (FStream, [';']);
          R := DecodeEntity (T);
          if R <> '' then
            begin
              Result := InterpretText (R);
              exit;
            end;
          Result := InterpretCharacterEntity (T);
        end else
        Result := InterpretText ('&');
    End;

  Function DecodeAttrValue (const S : String) : String;
  var I    : Integer;
      T, R : String;
      U    : Boolean;
    Begin
      I := 0;
      Result := S;
      Repeat
        I := PosNext ('&', Result, I);
        if I > 0 then
          begin
            T := CopyBefore (Result, cs_Everything - cs_AlphaNumeric - ['#'], False, I + 1);
            if Length (T) >= 2 then
              begin
                R := DecodeEntity (T);
                if R <> '' then
                  begin
                    U := (Length (Result) >= I + 1 + Length (T)) and
                         (Result [I + 1 + Length (T)] = ';');
                    Delete (Result, I, Length (T) + 1 + Cond (U, 1, 0));
                    Insert (R, Result, I);
                  end;
              end;
          end;
      Until I = 0;
    End;

  { "<"<name>[SP*<attrname>["="(<attrvalue>|""""<attrvalue>"""")]]*">"         }
  Function ExtractTag : AxmlComponent;
  const
    cs_TagNameStart  = cs_Alpha;
    cs_TagName       = cs_AlphaNumeric + ['-', '.'];
    cs_AttrNameStart = cs_Alpha;
    cs_AttrName      = cs_AlphaNumeric + ['-', '.'];
  var Pre,
      Name   : String;
      AName,
      AVal   : String;
      Attr   : TxmlTagAttributes;
    Begin
      Pre := ExtractChar (FStream, ['/']);
      Name := ExtractChar (FStream, cs_TagNameStart);
      if Name = '' then
        begin
          Result := InterpretText ('<' + Pre + Name);
          exit;
        end;
      Name := Pre + UpperCase (Name + ExtractChars (FStream, cs_TagName));
      SkipWhiteSpace (FStream);
      Attr := nil;
      While not EOF and not Match ('>') do
        begin
          AName := UpperCase (ExtractTo (FStream, cs_WhiteSpace + ['>', '='], False));
          SkipWhiteSpace (FStream);
          if Extract ('=') <> '' then
            begin
              SkipWhiteSpace (FStream);
              Pre := ExtractChar (FStream, ['"', '''']);
              if Pre <> '' then
                begin
                  AVal := ExtractTo (FStream, [Pre [1]] + ['>'], False);
                  if Match (Pre [1]) then
                    FStream.Skip;
                end else
                AVal := ExtractTo (FStream, cs_WhiteSpace + ['>'], False);
              SkipWhiteSpace (FStream);
            end else
            begin
              AVal := AName;
              AName := '';
            end;
          if not Assigned (Attr) then
            Attr := TxmlTagAttributes.Create;
          Attr.AddAttribute (AName, DecodeAttrValue (AVal));
        end;
      FStream.Skip;
      Result := TxmlTag.Create (Name, Attr);
    End;

var Ch : Char;
    T  : String;
  Begin
    if not EOF then
      begin
        T := ExtractTo (FStream, ['<', '&']);
        if T <> '' then
          Result := InterpretText (T) else
          begin
            Ch := Char (FStream.Read);
            if Ch = '&' then
              Result := ExtractEntity else
            if Ch = '<' then
              if Extract ('?') <> '' then
                Result := ExtractXMLTag else
              if Extract ('!--') <> '' then
                Result := ExtractComment else
              if Extract ('!') <> '' then
                Result := ExtractDTD else
                Result := ExtractTag;
          end;
      end else
      Result := nil;

    FToken := Result;
  End;

Function TxmlParser.ExtractFlatXML : AxmlComponent;
  Begin
    Result := nil;
    GetNextToken;
    While Assigned (FToken) do
      begin
        AppendComponents (Result, FToken);
        GetNextToken;
      end;
  End;

Constructor TxmlDocument.Create;
  Begin
    inherited Create;
    FContent := TxmlComponentList.Create;
  End;

Destructor TxmlDocument.Destroy;
  Begin
    FreeAndNil (FContent);
    inherited Destroy;
  End;

Function TxmlParser.ExtractXMLDocument : TxmlDocument;

{  Function ExtractElement : AxmlComponent; forward;

  Function ExtractTag : AxmlComponent;
  var EOpen  : AxmlTag;
      EClose : AxmlTag;
    Begin
      EOpen := AxmlTag (FToken);
      GetNextToken;
      Result := nil;
      if Assigned (FToken) then
        begin
          Result := ExtractElement;
          if (FToken is TxmlCloseTag) and (TxmlCloseTag (FToken).Name = EOpen.Name) then
            begin
              EClose := AxmlTag (FToken);
              GetNextToken;
            end else
            EClose := nil;
        end else
        EClose := nil;

      Result := InterpretTag (EOpen, EClose, Result);
    End;

  Function ExtractElement : AxmlComponent;
    Begin
      if FToken is TxmlOpenTag then
        Result := ExtractTag else
      if FToken is TxmlEmptyTag then
        Result := InterpretTag (TxmlEmptyTag (FToken)) else
        Result := FToken;
    End; }

  Begin
{    Result := TxmlDocument.Create;
    GetNextToken;
    While Assigned (FToken) do
      begin
        AppendComponents (Result.FContent, ExtractElement);

        if Assigned (FToken) then
          begin
            AppendComponents (Result.FContent, FToken);
            GetNextToken;
          end;
      end; }
  End;


end.



