unit cMemManager;

{                                                                              }
{                    Delphi Memory Manager Patch v0.01 (L0)                    }
{                                                                              }
{ Using this unit in your application will install a patch for the memory      }
{ manager that improves the performance for "growing" memory blocks.           }
{                                                                              }
{ Some operations in Delphi, like the concatenation of two strings, might      }
{ attempt a call to "ReallocMem" instead of "AllocMem" in the hope that the    }
{ allocation of a new memory block can be avoided. When ReallocMem is called   }
{ and the existing memory block can not be extended, a new block is allocated  }
{ using "AllocMem" and the existing data copied to the new location. These     }
{ memory move operations are very expensive and becomes noticable when you     }
{ have large blocks constantly growing in small chunks.                        }
{ This patch works around the problem by allocating some extra space whenever  }
{ ReallocMem is called, in anticipation of the block growing further. The      }
{ penalty paid in the extra memory allocated is small.                         }
{ In some situations the total memory usage of the application actually        }
{ decreases. This is because the memory is less fragmented and there might be  }
{ less unused memory blocks allocated by the application.                      }
{                                                                              }
{        This unit is copyright  2000 by David Butler (david@e.co.za)         }
{                                                                              }
{                  This unit is part of Delphi Fundamentals.                   }
{                  Its original file name is cMemManager.pas                   }
{                                                                              }
{                I invite you to use this unit, free of charge.                }
{        I invite you to distibute this unit, but it must be for free.         }
{             I also invite you to contribute to its development,              }
{             but do not distribute a modified copy of this file.              }
{       Send modifications, suggestions and bug reports to david@e.co.za       }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   2000/11/08  v0.01  Included in Delphi Fundamentals.                        }
{                                                                              }

interface



{ Performance monitor                                                          }
{ mmReallocCount count the number of calls to ReallocMem.                      }
{ mmReallocMisses count the number of times ReallocMem had to call the         }
{     old ReallocMem function.                                                 }
var
  mmReallocCount,
  mmReallocMisses : Integer;



implementation



var
  OldMemoryManager,
  NewMemoryManager  : TMemoryManager;

Function ReallocMem (P : Pointer; Size : Integer) : Pointer;
type PInteger = ^Integer;
var OldSize, NewSize : Integer;
  Begin
    OldSize := PInteger (PChar (P) - 4)^ and not 3;
    Inc (mmReallocCount);

    if OldSize >= (Size + 7) and not 3 then
      Result := P else
      begin
        if Size <= 16 then
          NewSize := 32 else
        if Size <= 128 then
          NewSize := Size * 2 else
          NewSize := Size + 112 + (Size shr 3);
        Result := OldMemoryManager.ReallocMem (P, NewSize);
        Inc (mmReallocMisses);
      end;
  End;

initialization
  GetMemoryManager (OldMemoryManager);
  NewMemoryManager := OldMemorymanager;
  NewMemoryManager.ReallocMem := ReallocMem;
  SetMemoryManager (NewMemoryManager);
finalization
  SetMemoryManager (OldMemoryManager);
end.

