{$INCLUDE cHeader.inc}
unit cSysUtils;

{                                                                              }
{                    System utility functions v0.01 (L0)                       }
{                                                                              }
{      This unit is copyright  2000-2001 by David Butler (david@e.co.za)      }
{                                                                              }
{                  This unit is part of Delphi Fundamentals.                   }
{                   Its original file name is cSysUtils.pas                    }
{                                                                              }
{                I invite you to use this unit, free of charge.                }
{        I invite you to distibute this unit, but it must be for free.         }
{             I also invite you to contribute to its development,              }
{             but do not distribute a modified copy of this file.              }
{       Send modifications, suggestions and bug reports to david@e.co.za       }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   2000/10/01  v0.01  Initial version spawned from cUtils.                    }
{                                                                              }

interface

uses
  Windows,
  cUtils;



{                                                                              }
{ Path functions                                                               }
{                                                                              }
const
  PathSeparator = {$IFDEF UNIX}'/'{$ELSE}'\'{$ENDIF};

Function UnixPathToWindowsPath (const Path : String) : String;
Function WindowsPathToUnixPath (const Path : String) : String;
Function PathWithSlashAtEnd (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathWithoutSlashAtEnd (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathLeftElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathRightElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathWithoutLeftElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function PathWithoutRightElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
Function MergeFilePath (const Path : String; const FileName : String; const SlashCh : Char = PathSeparator) : String;



{                                                                              }
{ FPU Precision                                                                }
{                                                                              }
type
  TFPUPrecision = (pSingle, pDouble, pExtended);

Procedure SetFPUPrecision (const Precision : TFPUPrecision);



{                                                                              }
{ Locked integer manipulation (simple synchronization)                         }
{                                                                              }
Function LockedAdd (var Target : Integer; const Value : Integer) : Integer; assembler; register;
Function LockedCompareExchange (var Target : Integer; const Exch, Comp : Integer) : Integer; assembler; register;
Function LockedDec (var Target : Integer) : Integer; assembler; register;
Function LockedExchange (var Target : Integer; Value : Integer) : Integer; assembler; register;
Function LockedExchangeAdd (var Target : Integer; Value : Integer) : Integer; assembler; register;
Function LockedExchangeDec (var Target : Integer) : Integer; assembler; register;
Function LockedExchangeInc (var Target : Integer) : Integer; assembler; register;
Function LockedExchangeSub (var Target : Integer; Value : Integer) : Integer; assembler; register;
Function LockedInc (var Target : Integer) : Integer; assembler; register;
Function LockedSub (var Target : Integer; const Value : Integer) : Integer; assembler; register;



{                                                                              }
{ Simple locks                                                                 }
{                                                                              }
type
  SimpleLock = Integer;
  PSimpleLock = ^SimpleLock;
  SimpleLockList = PointerArray;

Function  AcquireLock (var X : SimpleLock; const Block : Boolean) : Boolean; overload;
Function  AcquireLock (var X : SimpleLockList; const Block : Boolean) : Boolean; overload;
Procedure ReleaseLock (var X : SimpleLock); overload;
Procedure ReleaseLock (var X : SimpleLockList); overload;




{                                                                              }
{ Miscellaneous                                                                }
{                                                                              }
Function GetUserName : String;
Function GetEnvironmentStrings : StringArray;
Function ContentTypeFromExtention (Extention : String) : String;

{ Windows API                                                                  }
Function SetEnvVariable (szName : PChar; szValue : PChar; cmo : Integer) : LongBool; stdcall;

const
  // ContentType
  ct_HTML        = 'text/html';
  ct_ASCIIText   = 'text/plain; charset=US-ASCII';
  ct_JPG         = 'image/jpeg';
  ct_GIF         = 'image/gif';
  ct_BMP         = 'image/x-xbitmap';
  ct_PNG         = 'image/png';
  ct_MPG         = 'video/mpeg';
  ct_AVI         = 'video/avi';
  ct_QT          = 'video/quicktime';
  ct_Binary      = 'application/binary';
  ct_PDF         = 'application/pdf';
  ct_Postscript  = 'application/postscript';
  ct_BasicAudio  = 'audio/basic';
  ct_URLEncoded  = 'application/x-www-form-urlencoded';



implementation

uses
  Registry,
  SysUtils,
  cStrings;



{                                                                              }
{ Path functions                                                               }
{                                                                              }
Function PathWithSlashAtEnd (const Path : String; const SlashCh : Char = '\') : String;
  Begin
    if (Path <> '') and (Path [Length (Path)] <> SlashCh) then
      Result := Path + SlashCh else
      Result := Path;
  End;

Function PathWithoutSlashAtEnd (const Path : String; const SlashCh : Char = '\') : String;
  Begin
    if (Path <> '') and (Path [Length (Path)] = SlashCh) then
      Result := Copy (Path, 1, Length (Path) - 1) else
      Result := Path;
  End;

Function UnixPathToWindowsPath (const Path : String) : String;
  Begin
    Result := Replace ('/', '\', Replace (['\', ':', '"', '<', '>', '|'], '_', Path));
  End;

{ Network path:  \\      --->   /                                              }
{ Device path:   \\.\    --->   /                                              }
{ Drive letter:  X:      --->   X/                                             }
Function WindowsPathToUnixPath (const Path : String) : String;
  Begin
    Result := Path;
    if Length (Result) >= 2 then
      if Result [2] = ':' then // drive letter
        Result [2] := '/' else
      if (Result [1] = '\') and (Result [2] = '\') then
        if (Length (Result) >= 3) and (Result [3] = '.') then
          Delete (Result, 1, 3) else    // device path
          Delete (Result, 1, 1);        // network path
    Result := Replace ('\', '/', Replace (['/', ':', '"', '<', '>', '|'], '_', Path));
  End;

Function PathLeftElement (const Path : String; const SlashCh : Char = '\') : String;
  Begin
    Result := CopyBefore (Path, SlashCh, True);
  End;

Function PathRightElement (const Path : String; const SlashCh : Char = '\') : String;
  Begin
    Result := CopyFrom (Path, PosPrev (SlashCh, Path) + 1);
  End;

Function PathWithoutLeftElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
  Begin
    Result := CopyAfter (Path, SlashCh, False);
  End;

Function PathWithoutRightElement (const Path : String; const SlashCh : Char = PathSeparator) : String;
  Begin
    Result := CopyBefore (Path, SlashCh, False);
  End;

Function MergeFilePath (const Path : String; const FileName : String; const SlashCh : Char = PathSeparator) : String;
  Begin
    Result := PathWithSlashAtEnd (Path, SlashCh) + FileName;
  End;



{ You can modify the level of accuracy by changing the FPU's control word.     }
{ The default accuracy, as initialized by the Delphi runtime library,          }
{ is the slowest, but most precise one (i.e. extended). Note that changing     }
{ this only changes the execution time of division and, in the case of         }
{ Pentium II and Pentium III processors, square roots.                         }
{ See http://econos.com/optimize/float.htm#ctrlwrd                             }
Procedure SetFPUPrecision (const Precision : TFPUPrecision);
  Begin
    Case Precision of
      pSingle   : Set8087CW (Default8087CW and $FCFF);
      pDouble   : Set8087CW ((Default8087CW and $FCFF) or $0200);
      pExtended : Set8087CW (Default8087CW or $0300);
    end;
  End;



{                                                                              }
{ Locked integer manipulation                                                  }
{ Code by Azret Botash.                                                        }
{   LockedAdd, LockedDec, LockedSub, LockedInc does operation with Value on    }
{     Target and returns the result.                                           }
{   LockedCompareExchange compares Target with Comp and if equal sets Target   }
{     to Exch.                                                                 }
{   LockedExchange set current value of Target to Value and returns old value  }
{     of Target.                                                               }
{                                                                              }
Function LockedAdd (var Target : Integer; const Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, EDX
      LOCK
      XADD    [ECX], EAX
      ADD     EAX, EDX
  End;

Function LockedSub (var Target : Integer; const Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      NEG     EDX
      MOV     EAX, EDX
      LOCK
      XADD    [ECX], EAX
      ADD     EAX, EDX
  End;

Function LockedInc (var Target : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, 1
      LOCK
      XADD    [ECX], EAX
      INC     EAX
  End;

Function LockedDec (var Target : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, -1
      LOCK
      XADD    [ECX], EAX
      DEC     EAX
  End;

Function LockedCompareExchange (var Target : Integer; const Exch, Comp : Integer) : Integer; assembler; register;
  Asm
      XCHG    EAX, ECX
      LOCK
      CMPXCHG [ECX], EDX
  End;

Function LockedExchange (var Target : Integer; Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, EDX
      LOCK
      XCHG    [ECX], EAX
  End;

Function LockedExchangeAdd (var Target : Integer; Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, EDX
      LOCK
      XADD    [ECX], EAX
  End;

Function LockedExchangeDec (var Target : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, -1
      LOCK
      XADD    [ECX], EAX
  End;

Function LockedExchangeInc (var Target : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      MOV     EAX, 1
      LOCK
      XADD    [ECX], EAX
  End;

Function LockedExchangeSub (var Target : Integer; Value : Integer) : Integer; assembler; register;
  Asm
      MOV     ECX, EAX
      NEG     EDX
      MOV     EAX, EDX
      LOCK
      XADD    [ECX], EAX
  End;



{                                                                              }
{ Simple locks                                                                 }
{                                                                              }
Function AcquireLock (var X : SimpleLock; const Block : Boolean) : Boolean;
var R : Boolean;
  Begin
    Repeat
      Result := LockedInc (X) = 1;
      if not Result then
        begin
          LockedDec (X);
          R := Block;
        end else
        R := False;
      if R then
        Sleep (0);
    Until not R;
  End;

Function AcquireLock (var X : SimpleLockList; const Block : Boolean) : Boolean;
var I, J, L : Integer;
  Begin
    L := Length (X);
    if L = 0 then
      begin
        Result := True;
        exit;
      end;
    Repeat
      Result := AcquireLock (PSimpleLock (X [0])^, Block);
      if Result then
        For I := 1 to L - 1 do
          if not AcquireLock (PSimpleLock (X [I])^, False) then
            begin
              For J := I - 1 downto 0 do
                ReleaseLock (PSimpleLock (X [I])^);
              Swap (X [0], X [I]);
              Result := False;
              break;
            end;
    Until Result or not Block;
  End;

Procedure ReleaseLock (var X : SimpleLock);
  Begin
    LockedDec (X);
  End;

Procedure ReleaseLock (var X : SimpleLockList);
var I : Integer;
  Begin
    For I := Length (X) - 1 downto 0 do
      ReleaseLock (PSimpleLock (X [I])^);
  End;

{                                                                              }
{ Windows API calls                                                            }
{                                                                              }
Function SetEnvVariable; external kernel32 name 'SetEnvVariable';

Function GetUserName : String;
var L : LongWord;
  Begin
    L := 128;
    SetLength (Result, L);
    if Windows.GetUserName (PChar (Result), L) and (L > 0) then
      SetLength (Result, L - 1) else
      Result := '';
  End;

Function GetEnvironmentStrings : StringArray;
type ByteArray = Array [0..1048575] of Byte;
var P : ^ByteArray;
    I : Integer;
    S : String;
  Begin
    P := Pointer (Windows.GetEnvironmentStrings);
    try
      I := 0;
      While (P^ [I] <> 0) and (P^ [I + 1] <> 0) do
        Inc (I);
      Inc (I);
      SetLength (S, I);
      Move (P^, S [1], I);
    finally
      FreeEnvironmentStrings (Pointer (P));
    end;
    Result := Split (S, #0);
  End;

Function GetRegistryString (const RootKey : HKey; const Key : String; const Name : String) : String;
var R : TRegistry;
  Begin
    R := TRegistry.Create;
    try
      R.RootKey := RootKey;
      if R.OpenKeyReadOnly (Key) then
        try
          Result := R.ReadString (Name);
        finally
          R.CloseKey;
        end else
        Result := '';
    finally
      FreeAndNil (R);
    end;
  End;

Function ContentTypeFromExtention (Extention : String) : String;
type
  TContentTypeEntry = record
    Ext : String;
    CT  : String;
  end;
const
  BuiltInContentTypes = 18;
  BuiltInContentType : Array [1..BuiltInContentTypes] of TContentTypeEntry = (
      (Ext: '.htm';  CT: ct_HTML),       (Ext: '.html'; CT: ct_HTML),
      (Ext: '.jpg';  CT: ct_JPG),        (Ext: '.jpeg'; CT: ct_JPG),
      (Ext: '.gif';  CT: ct_GIF),        (Ext: '.png';  CT: ct_PNG),
      (Ext: '.bmp';  CT: ct_BMP),        (Ext: '.txt';  CT: ct_ASCIIText),
      (Ext: '.mpeg'; CT: ct_MPG),        (Ext: '.mpg';  CT: ct_MPG),
      (Ext: '.pdf';  CT: ct_PDF),        (Ext: '.exe';  CT: ct_Binary),
      (Ext: '.ps';   CT: ct_PostScript), (Ext: '.avi';  CT: ct_AVI),
      (Ext: '.qt';   CT: ct_QT),         (Ext: '.mov';  CT: ct_QT),
      (Ext: '.au';   CT: ct_BasicAudio), (Ext: '.wav';  CT: ct_BasicAudio));
var I : Integer;
  Begin
    // pre-defined types
    Extention := LowerCase (Extention);
    For I := 1 to BuiltInContentTypes do
      if Extention = BuiltInContentType [I].Ext then
        begin
          Result := BuiltInContentType [I].CT;
          exit;
        end;

    // check registry
    Result := GetRegistryString (HKEY_CLASSES_ROOT, '\' + Extention, 'Content Type');
  End;



end.

